////////////////////////////////////////////////////////////
// Copyright (C) Roman Ryltsov, 2008-2011
// Created by Roman Ryltsov roman@alax.info
// 
// $Id: NotifyIconWindow.h 93 2012-08-16 21:54:03Z roman $

#pragma once

#include <vfwmsgs.h>
#include <asferr.h> // FACILITY_NS
#include <mferror.h> // FACILITY_MF
#include "rowinhttp.h"
#include "AboutDialog.h"

////////////////////////////////////////////////////////////
// CNotifyIconWindow

class CNotifyIconWindow : 
	public CMessageOnlyWindowImpl<CNotifyIconWindow>
{
public:

DECLARE_WND_CLASS(_T("CNotifyIconWindow"))

BEGIN_MSG_MAP_EX(CNotifyIconWindow)
	_Z4(atlTraceUI, 4, _T("uMsg 0x%04X, wParam 0x%08X, lParam 0x%08X\n"), uMsg, wParam, lParam);
	//CHAIN_MSG_MAP(CMessageOnlyWindowImpl<CNotifyIconWindow>)
	MSG_WM_CREATE(OnCreate)
	MSG_WM_DESTROY(OnDestroy)
	MSG_WM_CHANGECBCHAIN(OnChangeCbChain)
	MSG_WM_DRAWCLIPBOARD(OnDrawClipboard)
	MSG_WM_MOUSEMOVE(OnMouseMove)
	MSG_WM_LBUTTONDBLCLK(OnLButtonDblClk)
	MSG_WM_RBUTTONUP(OnRButtonUp)
	MESSAGE_HANDLER_EX(NIN_BALLOONUSERCLICK, OnNotifyIconBalloonUserClick)
	MESSAGE_HANDLER_EX(WM_NOTIFYICON, OnNotifyIcon)
	COMMAND_ID_HANDLER_EX(ID_APP_ABOUT, OnApplicationAbout)
	COMMAND_ID_HANDLER_EX(ID_APP_EXIT, OnApplicationExit)
	REFLECT_NOTIFICATIONS()
END_MSG_MAP()

public:

	//////////////////////////////////////////////////////////
	// Window Message Identifiers

	enum
	{
		WM_FIRST = WM_APP,
		WM_NOTIFYICON,
	};

private:
	ULONGLONG m_nCommonControlsVersion;
	ULONGLONG m_nShellVersion;
	NOTIFYICONDATA m_NotifyIconData;
	CWindow m_NextClipboardViewerWindow;
	CString m_sQuery;

	static BOOL IsQuartzResult(HRESULT nResult, CString* psMessage = NULL)
	{
		if(HRESULT_FACILITY(nResult) != FACILITY_ITF)
			return FALSE;
		const SCODE nCode = HRESULT_CODE(nResult);
		if(nCode < 0x0200)// && nCode <= 0xFFFF)
			return FALSE;
		const CString sMessage = AtlFormatSystemMessage(CDataLibraryMap::LoadLibrary(_T("quartz.dll")), nResult);
		if(sMessage.IsEmpty())
			return FALSE;
		if(psMessage)
			*psMessage = sMessage;
		return TRUE;
	}
	static CString LookupQuartzIdentifier(HRESULT nValue)
	{
		static const struct { HRESULT nValue; LPCSTR pszName; } g_pMap[] = 
		{
			#define A(x) { x, #x },
			A(VFW_E_INVALIDMEDIATYPE)
			A(VFW_E_INVALIDSUBTYPE)
			A(VFW_E_NEED_OWNER)
			A(VFW_E_ENUM_OUT_OF_SYNC)
			A(VFW_E_ALREADY_CONNECTED)
			A(VFW_E_FILTER_ACTIVE)
			A(VFW_E_NO_TYPES)
			A(VFW_E_NO_ACCEPTABLE_TYPES)
			A(VFW_E_INVALID_DIRECTION)
			A(VFW_E_NOT_CONNECTED)
			A(VFW_E_NO_ALLOCATOR)
			A(VFW_E_RUNTIME_ERROR)
			A(VFW_E_BUFFER_NOTSET)
			A(VFW_E_BUFFER_OVERFLOW)
			A(VFW_E_BADALIGN)
			A(VFW_E_ALREADY_COMMITTED)
			A(VFW_E_BUFFERS_OUTSTANDING)
			A(VFW_E_NOT_COMMITTED)
			A(VFW_E_SIZENOTSET)
			A(VFW_E_NO_CLOCK)
			A(VFW_E_NO_SINK)
			A(VFW_E_NO_INTERFACE)
			A(VFW_E_NOT_FOUND)
			A(VFW_E_CANNOT_CONNECT)
			A(VFW_E_CANNOT_RENDER)
			A(VFW_E_CHANGING_FORMAT)
			A(VFW_E_NO_COLOR_KEY_SET)
			A(VFW_E_NOT_OVERLAY_CONNECTION)
			A(VFW_E_NOT_SAMPLE_CONNECTION)
			A(VFW_E_PALETTE_SET)
			A(VFW_E_COLOR_KEY_SET)
			A(VFW_E_NO_COLOR_KEY_FOUND)
			A(VFW_E_NO_PALETTE_AVAILABLE)
			A(VFW_E_NO_DISPLAY_PALETTE)
			A(VFW_E_TOO_MANY_COLORS)
			A(VFW_E_STATE_CHANGED)
			A(VFW_E_NOT_STOPPED)
			A(VFW_E_NOT_PAUSED)
			A(VFW_E_NOT_RUNNING)
			A(VFW_E_WRONG_STATE)
			A(VFW_E_START_TIME_AFTER_END)
			A(VFW_E_INVALID_RECT)
			A(VFW_E_TYPE_NOT_ACCEPTED)
			A(VFW_E_SAMPLE_REJECTED)
			A(VFW_E_SAMPLE_REJECTED_EOS)
			A(VFW_E_DUPLICATE_NAME)
			A(VFW_S_DUPLICATE_NAME)
			A(VFW_E_TIMEOUT)
			A(VFW_E_INVALID_FILE_FORMAT)
			A(VFW_E_ENUM_OUT_OF_RANGE)
			A(VFW_E_CIRCULAR_GRAPH)
			A(VFW_E_NOT_ALLOWED_TO_SAVE)
			A(VFW_E_TIME_ALREADY_PASSED)
			A(VFW_E_ALREADY_CANCELLED)
			A(VFW_E_CORRUPT_GRAPH_FILE)
			A(VFW_E_ADVISE_ALREADY_SET)
			A(VFW_S_STATE_INTERMEDIATE)
			A(VFW_E_NO_MODEX_AVAILABLE)
			A(VFW_E_NO_ADVISE_SET)
			A(VFW_E_NO_FULLSCREEN)
			A(VFW_E_IN_FULLSCREEN_MODE)
			A(VFW_E_UNKNOWN_FILE_TYPE)
			A(VFW_E_CANNOT_LOAD_SOURCE_FILTER)
			A(VFW_S_PARTIAL_RENDER)
			A(VFW_E_FILE_TOO_SHORT)
			A(VFW_E_INVALID_FILE_VERSION)
			A(VFW_S_SOME_DATA_IGNORED)
			A(VFW_S_CONNECTIONS_DEFERRED)
			A(VFW_E_INVALID_CLSID)
			A(VFW_E_INVALID_MEDIA_TYPE)
			A(VFW_E_BAD_KEY)
			A(VFW_S_NO_MORE_ITEMS)
			A(VFW_E_SAMPLE_TIME_NOT_SET)
			A(VFW_S_RESOURCE_NOT_NEEDED)
			A(VFW_E_MEDIA_TIME_NOT_SET)
			A(VFW_E_NO_TIME_FORMAT_SET)
			A(VFW_E_MONO_AUDIO_HW)
			A(VFW_S_MEDIA_TYPE_IGNORED)
			A(VFW_E_NO_DECOMPRESSOR)
			A(VFW_E_NO_AUDIO_HARDWARE)
			A(VFW_S_VIDEO_NOT_RENDERED)
			A(VFW_S_AUDIO_NOT_RENDERED)
			A(VFW_E_RPZA)
			A(VFW_S_RPZA)
			A(VFW_E_PROCESSOR_NOT_SUITABLE)
			A(VFW_E_UNSUPPORTED_AUDIO)
			A(VFW_E_UNSUPPORTED_VIDEO)
			A(VFW_E_MPEG_NOT_CONSTRAINED)
			A(VFW_E_NOT_IN_GRAPH)
			A(VFW_S_ESTIMATED)
			A(VFW_E_NO_TIME_FORMAT)
			A(VFW_E_READ_ONLY)
			A(VFW_S_RESERVED)
			A(VFW_E_BUFFER_UNDERFLOW)
			A(VFW_E_UNSUPPORTED_STREAM)
			A(VFW_E_NO_TRANSPORT)
			A(VFW_S_STREAM_OFF)
			A(VFW_S_CANT_CUE)
			A(VFW_E_BAD_VIDEOCD)
			A(VFW_S_NO_STOP_TIME)
			A(VFW_E_OUT_OF_VIDEO_MEMORY)
			A(VFW_E_VP_NEGOTIATION_FAILED)
			A(VFW_E_DDRAW_CAPS_NOT_SUITABLE)
			A(VFW_E_NO_VP_HARDWARE)
			A(VFW_E_NO_CAPTURE_HARDWARE)
			A(VFW_E_DVD_OPERATION_INHIBITED)
			A(VFW_E_DVD_INVALIDDOMAIN)
			A(VFW_E_DVD_NO_BUTTON)
			A(VFW_E_DVD_GRAPHNOTREADY)
			A(VFW_E_DVD_RENDERFAIL)
			A(VFW_E_DVD_DECNOTENOUGH)
			A(VFW_E_DDRAW_VERSION_NOT_SUITABLE)
			A(VFW_E_COPYPROT_FAILED)
			A(VFW_S_NOPREVIEWPIN)
			A(VFW_E_TIME_EXPIRED)
			A(VFW_S_DVD_NON_ONE_SEQUENTIAL)
			A(VFW_E_DVD_WRONG_SPEED)
			A(VFW_E_DVD_MENU_DOES_NOT_EXIST)
			A(VFW_E_DVD_CMD_CANCELLED)
			A(VFW_E_DVD_STATE_WRONG_VERSION)
			A(VFW_E_DVD_STATE_CORRUPT)
			A(VFW_E_DVD_STATE_WRONG_DISC)
			A(VFW_E_DVD_INCOMPATIBLE_REGION)
			A(VFW_E_DVD_NO_ATTRIBUTES)
			A(VFW_E_DVD_NO_GOUP_PGC)
			A(VFW_E_DVD_LOW_PARENTAL_LEVEL)
			A(VFW_E_DVD_NOT_IN_KARAOKE_MODE)
			A(VFW_S_DVD_CHANNEL_CONTENTS_NOT_AVAILABLE)
			A(VFW_S_DVD_NOT_ACCURATE)
			A(VFW_E_FRAME_STEP_UNSUPPORTED)
			A(VFW_E_DVD_STREAM_DISABLED)
			A(VFW_E_DVD_TITLE_UNKNOWN)
			A(VFW_E_DVD_INVALID_DISC)
			A(VFW_E_DVD_NO_RESUME_INFORMATION)
			A(VFW_E_PIN_ALREADY_BLOCKED_ON_THIS_THREAD)
			A(VFW_E_PIN_ALREADY_BLOCKED)
			A(VFW_E_CERTIFICATION_FAILURE)
			A(VFW_E_VMR_NOT_IN_MIXER_MODE)
			A(VFW_E_VMR_NO_AP_SUPPLIED)
			A(VFW_E_VMR_NO_DEINTERLACE_HW)
			A(VFW_E_VMR_NO_PROCAMP_HW)
			A(VFW_E_DVD_VMR9_INCOMPATIBLEDEC)
			A(VFW_E_VMR_NO_AP_SUPPLIED)
			A(VFW_E_VMR_NO_DEINTERLACE_HW)
			A(VFW_E_VMR_NO_PROCAMP_HW)
			A(VFW_E_DVD_VMR9_INCOMPATIBLEDEC)
			A(VFW_E_NO_COPP_HW)
			A(VFW_E_DVD_NONBLOCKING)
			A(VFW_E_DVD_TOO_MANY_RENDERERS_IN_FILTER_GRAPH)
			A(VFW_E_DVD_NON_EVR_RENDERER_IN_FILTER_GRAPH)
			A(VFW_E_DVD_RESOLUTION_ERROR)
			A(VFW_E_CODECAPI_LINEAR_RANGE)
			A(VFW_E_CODECAPI_ENUMERATED)
			A(VFW_E_CODECAPI_NO_DEFAULT)
			A(VFW_E_CODECAPI_NO_CURRENT_VALUE)
			A(VFW_E_DVD_CHAPTER_DOES_NOT_EXIST)
			#undef A
		};
		for(SIZE_T nIndex = 0; nIndex < DIM(g_pMap); nIndex++)
			if(g_pMap[nIndex].nValue == nValue)
				return CString(g_pMap[nIndex].pszName);
		return _T("");
	}
	static BOOL IsWmResult(HRESULT nResult, CString* psMessage = NULL)
	{
		if(HRESULT_FACILITY(nResult) != FACILITY_NS)
			return FALSE;
		// SUGG: Cache loaded libraries
		HMODULE hModule = NULL;
		const SCODE nCode = HRESULT_CODE(nResult);
		CString sMessage;
		if(nCode >= 2000 && nCode < 3000) // 2000 -  2999 = ASF (defined in ASFERR.MC)
			sMessage = AtlFormatSystemMessage(CDataLibraryMap::LoadLibrary(_T("asferror.dll")), nResult);
		else
			sMessage = AtlFormatSystemMessage(CDataLibraryMap::LoadLibrary(_T("wmerror.dll")), nResult);
		if(sMessage.IsEmpty())
			return FALSE;
		if(psMessage)
			*psMessage = sMessage;
		return TRUE;
	}
	static BOOL IsMfResult(HRESULT nResult, CString* psMessage = NULL)
	{
		if(HRESULT_FACILITY(nResult) != FACILITY_MF)
			return FALSE;
		const CString sMessage = AtlFormatSystemMessage(CDataLibraryMap::LoadLibrary(_T("mferror.dll")), nResult);
		if(sMessage.IsEmpty())
			return FALSE;
		if(psMessage)
			*psMessage = sMessage;
		return TRUE;
	}
	static CString LookupMfIdentifier(HRESULT nValue)
	{
		static const struct { HRESULT nValue; LPCSTR pszName; } g_pMap[] = 
		{
			#define A(x) { x, #x },
			A(MF_E_PLATFORM_NOT_INITIALIZED)
			A(MF_E_BUFFERTOOSMALL)
			A(MF_E_INVALIDREQUEST)
			A(MF_E_INVALIDSTREAMNUMBER)
			A(MF_E_INVALIDMEDIATYPE)
			A(MF_E_NOTACCEPTING)
			A(MF_E_NOT_INITIALIZED)
			A(MF_E_UNSUPPORTED_REPRESENTATION)
			A(MF_E_NO_MORE_TYPES)
			A(MF_E_UNSUPPORTED_SERVICE)
			A(MF_E_UNEXPECTED)
			A(MF_E_INVALIDNAME)
			A(MF_E_INVALIDTYPE)
			A(MF_E_INVALID_FILE_FORMAT)
			A(MF_E_INVALIDINDEX)
			A(MF_E_INVALID_TIMESTAMP)
			A(MF_E_UNSUPPORTED_SCHEME)
			A(MF_E_UNSUPPORTED_BYTESTREAM_TYPE)
			A(MF_E_UNSUPPORTED_TIME_FORMAT)
			A(MF_E_NO_SAMPLE_TIMESTAMP)
			A(MF_E_NO_SAMPLE_DURATION)
			A(MF_E_INVALID_STREAM_DATA)
			A(MF_E_RT_UNAVAILABLE)
			A(MF_E_UNSUPPORTED_RATE)
			A(MF_E_THINNING_UNSUPPORTED)
			A(MF_E_REVERSE_UNSUPPORTED)
			A(MF_E_UNSUPPORTED_RATE_TRANSITION)
			A(MF_E_RATE_CHANGE_PREEMPTED)
			A(MF_E_NOT_FOUND)
			A(MF_E_NOT_AVAILABLE)
			A(MF_E_NO_CLOCK)
			A(MF_S_MULTIPLE_BEGIN)
			A(MF_E_MULTIPLE_BEGIN)
			A(MF_E_MULTIPLE_SUBSCRIBERS)
			A(MF_E_TIMER_ORPHANED)
			A(MF_E_STATE_TRANSITION_PENDING)
			A(MF_E_UNSUPPORTED_STATE_TRANSITION)
			A(MF_E_UNRECOVERABLE_ERROR_OCCURRED)
			A(MF_E_SAMPLE_HAS_TOO_MANY_BUFFERS)
			A(MF_E_SAMPLE_NOT_WRITABLE)
			A(MF_E_INVALID_KEY)
			A(MF_E_BAD_STARTUP_VERSION)
			A(MF_E_UNSUPPORTED_CAPTION)
			A(MF_E_INVALID_POSITION)
			A(MF_E_ATTRIBUTENOTFOUND)
			A(MF_E_PROPERTY_TYPE_NOT_ALLOWED)
			A(MF_E_PROPERTY_TYPE_NOT_SUPPORTED)
			A(MF_E_PROPERTY_EMPTY)
			A(MF_E_PROPERTY_NOT_EMPTY)
			A(MF_E_PROPERTY_VECTOR_NOT_ALLOWED)
			A(MF_E_PROPERTY_VECTOR_REQUIRED)
			A(MF_E_OPERATION_CANCELLED)
			A(MF_E_BYTESTREAM_NOT_SEEKABLE)
			A(MF_E_DISABLED_IN_SAFEMODE)
			A(MF_E_CANNOT_PARSE_BYTESTREAM)
			A(MF_E_SOURCERESOLVER_MUTUALLY_EXCLUSIVE_FLAGS)
			A(MF_E_MEDIAPROC_WRONGSTATE)
			A(MF_E_RT_THROUGHPUT_NOT_AVAILABLE)
			A(MF_E_RT_TOO_MANY_CLASSES)
			A(MF_E_RT_WOULDBLOCK)
			A(MF_E_NO_BITPUMP)
			A(MF_E_RT_OUTOFMEMORY)
			A(MF_E_RT_WORKQUEUE_CLASS_NOT_SPECIFIED)
			A(MF_E_INSUFFICIENT_BUFFER)
			A(MF_E_CANNOT_CREATE_SINK)
			A(MF_E_BYTESTREAM_UNKNOWN_LENGTH)
			A(MF_E_SESSION_PAUSEWHILESTOPPED)
			A(MF_S_ACTIVATE_REPLACED)
			A(MF_E_FORMAT_CHANGE_NOT_SUPPORTED)
			A(MF_E_INVALID_WORKQUEUE)
			A(MF_E_DRM_UNSUPPORTED)
			A(MF_E_UNAUTHORIZED)
			A(MF_E_OUT_OF_RANGE)
			A(MF_E_INVALID_CODEC_MERIT)
			A(MF_E_HW_MFT_FAILED_START_STREAMING)
			A(MF_S_ASF_PARSEINPROGRESS)
			A(MF_E_ASF_PARSINGINCOMPLETE)
			A(MF_E_ASF_MISSINGDATA)
			A(MF_E_ASF_INVALIDDATA)
			A(MF_E_ASF_OPAQUEPACKET)
			A(MF_E_ASF_NOINDEX)
			A(MF_E_ASF_OUTOFRANGE)
			A(MF_E_ASF_INDEXNOTLOADED)
			A(MF_E_ASF_TOO_MANY_PAYLOADS)
			A(MF_E_ASF_UNSUPPORTED_STREAM_TYPE)
			A(MF_E_ASF_DROPPED_PACKET)
			A(MF_E_NO_EVENTS_AVAILABLE)
			A(MF_E_INVALID_STATE_TRANSITION)
			A(MF_E_END_OF_STREAM)
			A(MF_E_SHUTDOWN)
			A(MF_E_MP3_NOTFOUND)
			A(MF_E_MP3_OUTOFDATA)
			A(MF_E_MP3_NOTMP3)
			A(MF_E_MP3_NOTSUPPORTED)
			A(MF_E_NO_DURATION)
			A(MF_E_INVALID_FORMAT)
			A(MF_E_PROPERTY_NOT_FOUND)
			A(MF_E_PROPERTY_READ_ONLY)
			A(MF_E_PROPERTY_NOT_ALLOWED)
			A(MF_E_MEDIA_SOURCE_NOT_STARTED)
			A(MF_E_UNSUPPORTED_FORMAT)
			A(MF_E_MP3_BAD_CRC)
			A(MF_E_NOT_PROTECTED)
			A(MF_E_MEDIA_SOURCE_WRONGSTATE)
			A(MF_E_MEDIA_SOURCE_NO_STREAMS_SELECTED)
			A(MF_E_CANNOT_FIND_KEYFRAME_SAMPLE)
			A(MF_E_NETWORK_RESOURCE_FAILURE)
			A(MF_E_NET_WRITE)
			A(MF_E_NET_READ)
			A(MF_E_NET_REQUIRE_NETWORK)
			A(MF_E_NET_REQUIRE_ASYNC)
			A(MF_E_NET_BWLEVEL_NOT_SUPPORTED)
			A(MF_E_NET_STREAMGROUPS_NOT_SUPPORTED)
			A(MF_E_NET_MANUALSS_NOT_SUPPORTED)
			A(MF_E_NET_INVALID_PRESENTATION_DESCRIPTOR)
			A(MF_E_NET_CACHESTREAM_NOT_FOUND)
			A(MF_I_MANUAL_PROXY)
			A(MF_E_NET_REQUIRE_INPUT)
			A(MF_E_NET_REDIRECT)
			A(MF_E_NET_REDIRECT_TO_PROXY)
			A(MF_E_NET_TOO_MANY_REDIRECTS)
			A(MF_E_NET_TIMEOUT)
			A(MF_E_NET_CLIENT_CLOSE)
			A(MF_E_NET_BAD_CONTROL_DATA)
			A(MF_E_NET_INCOMPATIBLE_SERVER)
			A(MF_E_NET_UNSAFE_URL)
			A(MF_E_NET_CACHE_NO_DATA)
			A(MF_E_NET_EOL)
			A(MF_E_NET_BAD_REQUEST)
			A(MF_E_NET_INTERNAL_SERVER_ERROR)
			A(MF_E_NET_SESSION_NOT_FOUND)
			A(MF_E_NET_NOCONNECTION)
			A(MF_E_NET_CONNECTION_FAILURE)
			A(MF_E_NET_INCOMPATIBLE_PUSHSERVER)
			A(MF_E_NET_SERVER_ACCESSDENIED)
			A(MF_E_NET_PROXY_ACCESSDENIED)
			A(MF_E_NET_CANNOTCONNECT)
			A(MF_E_NET_INVALID_PUSH_TEMPLATE)
			A(MF_E_NET_INVALID_PUSH_PUBLISHING_POINT)
			A(MF_E_NET_BUSY)
			A(MF_E_NET_RESOURCE_GONE)
			A(MF_E_NET_ERROR_FROM_PROXY)
			A(MF_E_NET_PROXY_TIMEOUT)
			A(MF_E_NET_SERVER_UNAVAILABLE)
			A(MF_E_NET_TOO_MUCH_DATA)
			A(MF_E_NET_SESSION_INVALID)
			A(MF_E_OFFLINE_MODE)
			A(MF_E_NET_UDP_BLOCKED)
			A(MF_E_NET_UNSUPPORTED_CONFIGURATION)
			A(MF_E_NET_PROTOCOL_DISABLED)
			A(MF_E_ALREADY_INITIALIZED)
			A(MF_E_BANDWIDTH_OVERRUN)
			A(MF_E_LATE_SAMPLE)
			A(MF_E_FLUSH_NEEDED)
			A(MF_E_INVALID_PROFILE)
			A(MF_E_INDEX_NOT_COMMITTED)
			A(MF_E_NO_INDEX)
			A(MF_E_CANNOT_INDEX_IN_PLACE)
			A(MF_E_MISSING_ASF_LEAKYBUCKET)
			A(MF_E_INVALID_ASF_STREAMID)
			A(MF_E_STREAMSINK_REMOVED)
			A(MF_E_STREAMSINKS_OUT_OF_SYNC)
			A(MF_E_STREAMSINKS_FIXED)
			A(MF_E_STREAMSINK_EXISTS)
			A(MF_E_SAMPLEALLOCATOR_CANCELED)
			A(MF_E_SAMPLEALLOCATOR_EMPTY)
			A(MF_E_SINK_ALREADYSTOPPED)
			A(MF_E_ASF_FILESINK_BITRATE_UNKNOWN)
			A(MF_E_SINK_NO_STREAMS)
			A(MF_S_SINK_NOT_FINALIZED)
			A(MF_E_METADATA_TOO_LONG)
			A(MF_E_SINK_NO_SAMPLES_PROCESSED)
			A(MF_E_VIDEO_REN_NO_PROCAMP_HW)
			A(MF_E_VIDEO_REN_NO_DEINTERLACE_HW)
			A(MF_E_VIDEO_REN_COPYPROT_FAILED)
			A(MF_E_VIDEO_REN_SURFACE_NOT_SHARED)
			A(MF_E_VIDEO_DEVICE_LOCKED)
			A(MF_E_NEW_VIDEO_DEVICE)
			A(MF_E_NO_VIDEO_SAMPLE_AVAILABLE)
			A(MF_E_NO_AUDIO_PLAYBACK_DEVICE)
			A(MF_E_AUDIO_PLAYBACK_DEVICE_IN_USE)
			A(MF_E_AUDIO_PLAYBACK_DEVICE_INVALIDATED)
			A(MF_E_AUDIO_SERVICE_NOT_RUNNING)
			A(MF_E_TOPO_INVALID_OPTIONAL_NODE)
			A(MF_E_TOPO_CANNOT_FIND_DECRYPTOR)
			A(MF_E_TOPO_CODEC_NOT_FOUND)
			A(MF_E_TOPO_CANNOT_CONNECT)
			A(MF_E_TOPO_UNSUPPORTED)
			A(MF_E_TOPO_INVALID_TIME_ATTRIBUTES)
			A(MF_E_TOPO_LOOPS_IN_TOPOLOGY)
			A(MF_E_TOPO_MISSING_PRESENTATION_DESCRIPTOR)
			A(MF_E_TOPO_MISSING_STREAM_DESCRIPTOR)
			A(MF_E_TOPO_STREAM_DESCRIPTOR_NOT_SELECTED)
			A(MF_E_TOPO_MISSING_SOURCE)
			A(MF_E_TOPO_SINK_ACTIVATES_UNSUPPORTED)
			A(MF_E_SEQUENCER_UNKNOWN_SEGMENT_ID)
			A(MF_S_SEQUENCER_CONTEXT_CANCELED)
			A(MF_E_NO_SOURCE_IN_CACHE)
			A(MF_S_SEQUENCER_SEGMENT_AT_END_OF_STREAM)
			A(MF_E_TRANSFORM_TYPE_NOT_SET)
			A(MF_E_TRANSFORM_STREAM_CHANGE)
			A(MF_E_TRANSFORM_INPUT_REMAINING)
			A(MF_E_TRANSFORM_PROFILE_MISSING)
			A(MF_E_TRANSFORM_PROFILE_INVALID_OR_CORRUPT)
			A(MF_E_TRANSFORM_PROFILE_TRUNCATED)
			A(MF_E_TRANSFORM_PROPERTY_PID_NOT_RECOGNIZED)
			A(MF_E_TRANSFORM_PROPERTY_VARIANT_TYPE_WRONG)
			A(MF_E_TRANSFORM_PROPERTY_NOT_WRITEABLE)
			A(MF_E_TRANSFORM_PROPERTY_ARRAY_VALUE_WRONG_NUM_DIM)
			A(MF_E_TRANSFORM_PROPERTY_VALUE_SIZE_WRONG)
			A(MF_E_TRANSFORM_PROPERTY_VALUE_OUT_OF_RANGE)
			A(MF_E_TRANSFORM_PROPERTY_VALUE_INCOMPATIBLE)
			A(MF_E_TRANSFORM_NOT_POSSIBLE_FOR_CURRENT_OUTPUT_MEDIATYPE)
			A(MF_E_TRANSFORM_NOT_POSSIBLE_FOR_CURRENT_INPUT_MEDIATYPE)
			A(MF_E_TRANSFORM_NOT_POSSIBLE_FOR_CURRENT_MEDIATYPE_COMBINATION)
			A(MF_E_TRANSFORM_CONFLICTS_WITH_OTHER_CURRENTLY_ENABLED_FEATURES)
			A(MF_E_TRANSFORM_NEED_MORE_INPUT)
			A(MF_E_TRANSFORM_NOT_POSSIBLE_FOR_CURRENT_SPKR_CONFIG)
			A(MF_E_TRANSFORM_CANNOT_CHANGE_MEDIATYPE_WHILE_PROCESSING)
			A(MF_S_TRANSFORM_DO_NOT_PROPAGATE_EVENT)
			A(MF_E_UNSUPPORTED_D3D_TYPE)
			A(MF_E_TRANSFORM_ASYNC_LOCKED)
			A(MF_E_TRANSFORM_CANNOT_INITIALIZE_ACM_DRIVER)
			A(MF_E_LICENSE_INCORRECT_RIGHTS)
			A(MF_E_LICENSE_OUTOFDATE)
			A(MF_E_LICENSE_REQUIRED)
			A(MF_E_DRM_HARDWARE_INCONSISTENT)
			A(MF_E_NO_CONTENT_PROTECTION_MANAGER)
			A(MF_E_LICENSE_RESTORE_NO_RIGHTS)
			A(MF_E_BACKUP_RESTRICTED_LICENSE)
			A(MF_E_LICENSE_RESTORE_NEEDS_INDIVIDUALIZATION)
			A(MF_S_PROTECTION_NOT_REQUIRED)
			A(MF_E_COMPONENT_REVOKED)
			A(MF_E_TRUST_DISABLED)
			A(MF_E_WMDRMOTA_NO_ACTION)
			A(MF_E_WMDRMOTA_ACTION_ALREADY_SET)
			A(MF_E_WMDRMOTA_DRM_HEADER_NOT_AVAILABLE)
			A(MF_E_WMDRMOTA_DRM_ENCRYPTION_SCHEME_NOT_SUPPORTED)
			A(MF_E_WMDRMOTA_ACTION_MISMATCH)
			A(MF_E_WMDRMOTA_INVALID_POLICY)
			A(MF_E_POLICY_UNSUPPORTED)
			A(MF_E_OPL_NOT_SUPPORTED)
			A(MF_E_TOPOLOGY_VERIFICATION_FAILED)
			A(MF_E_SIGNATURE_VERIFICATION_FAILED)
			A(MF_E_DEBUGGING_NOT_ALLOWED)
			A(MF_E_CODE_EXPIRED)
			A(MF_E_GRL_VERSION_TOO_LOW)
			A(MF_E_GRL_RENEWAL_NOT_FOUND)
			A(MF_E_GRL_EXTENSIBLE_ENTRY_NOT_FOUND)
			A(MF_E_KERNEL_UNTRUSTED)
			A(MF_E_PEAUTH_UNTRUSTED)
			A(MF_E_NON_PE_PROCESS)
			A(MF_E_REBOOT_REQUIRED)
			A(MF_S_WAIT_FOR_POLICY_SET)
			A(MF_S_VIDEO_DISABLED_WITH_UNKNOWN_SOFTWARE_OUTPUT)
			A(MF_E_GRL_INVALID_FORMAT)
			A(MF_E_GRL_UNRECOGNIZED_FORMAT)
			A(MF_E_ALL_PROCESS_RESTART_REQUIRED)
			A(MF_E_PROCESS_RESTART_REQUIRED)
			A(MF_E_USERMODE_UNTRUSTED)
			A(MF_E_PEAUTH_SESSION_NOT_STARTED)
			A(MF_E_PEAUTH_PUBLICKEY_REVOKED)
			A(MF_E_GRL_ABSENT)
			A(MF_S_PE_TRUSTED)
			A(MF_E_PE_UNTRUSTED)
			A(MF_E_PEAUTH_NOT_STARTED)
			A(MF_E_INCOMPATIBLE_SAMPLE_PROTECTION)
			A(MF_E_PE_SESSIONS_MAXED)
			A(MF_E_HIGH_SECURITY_LEVEL_CONTENT_NOT_ALLOWED)
			A(MF_E_TEST_SIGNED_COMPONENTS_NOT_ALLOWED)
			A(MF_E_ITA_UNSUPPORTED_ACTION)
			A(MF_E_ITA_ERROR_PARSING_SAP_PARAMETERS)
			A(MF_E_POLICY_MGR_ACTION_OUTOFBOUNDS)
			A(MF_E_BAD_OPL_STRUCTURE_FORMAT)
			A(MF_E_ITA_UNRECOGNIZED_ANALOG_VIDEO_PROTECTION_GUID)
			A(MF_E_NO_PMP_HOST)
			A(MF_E_ITA_OPL_DATA_NOT_INITIALIZED)
			A(MF_E_ITA_UNRECOGNIZED_ANALOG_VIDEO_OUTPUT)
			A(MF_E_ITA_UNRECOGNIZED_DIGITAL_VIDEO_OUTPUT)
			A(MF_E_CLOCK_INVALID_CONTINUITY_KEY)
			A(MF_E_CLOCK_NO_TIME_SOURCE)
			A(MF_E_CLOCK_STATE_ALREADY_SET)
			A(MF_E_CLOCK_NOT_SIMPLE)
			A(MF_S_CLOCK_STOPPED)
			A(MF_E_NO_MORE_DROP_MODES)
			A(MF_E_NO_MORE_QUALITY_LEVELS)
			A(MF_E_DROPTIME_NOT_SUPPORTED)
			A(MF_E_QUALITYKNOB_WAIT_LONGER)
			A(MF_E_QM_INVALIDSTATE)
			A(MF_E_TRANSCODE_NO_CONTAINERTYPE)
			A(MF_E_TRANSCODE_PROFILE_NO_MATCHING_STREAMS)
			A(MF_E_TRANSCODE_NO_MATCHING_ENCODER)
			A(MF_E_ALLOCATOR_NOT_INITIALIZED)
			A(MF_E_ALLOCATOR_NOT_COMMITED)
			A(MF_E_ALLOCATOR_ALREADY_COMMITED)
			A(MF_E_STREAM_ERROR)
			A(MF_E_INVALID_STREAM_STATE)
			A(MF_E_HW_STREAM_NOT_CONNECTED)
			#undef A
		};
		for(SIZE_T nIndex = 0; nIndex < DIM(g_pMap); nIndex++)
			if(g_pMap[nIndex].nValue == nValue)
				return CString(g_pMap[nIndex].pszName);
		return _T("");
	}
	static BOOL IsWs2Result(HRESULT nResult, CString* psMessage = NULL)
	{
		if(HRESULT_FACILITY(nResult) != FACILITY_WIN32)
			return FALSE;
		const SCODE nCode = HRESULT_CODE(nResult);
		if(nCode < 10000 || nCode >= 11150) // WSABASEERR, ...
			return FALSE;
		const CString sMessage = AtlFormatSystemMessage(CDataLibraryMap::LoadLibrary(_T("ws2_32.dll")), nResult);
		if(sMessage.IsEmpty())
			return FALSE;
		if(psMessage)
			*psMessage = sMessage;
		return TRUE;
	}
	static BOOL IsWinHttpResult(HRESULT nResult, CString* psMessage = NULL)
	{
		if(HRESULT_FACILITY(nResult) != FACILITY_WIN32)
			return FALSE;
		const LONG nCode = HRESULT_CODE(nResult);
		if(nCode < 12000 || nCode >= 12200) // WINHTTP_ERROR_BASE, WINHTTP_ERROR_LAST
			return FALSE;
		const CString sMessage = AtlFormatSystemMessage(CDataLibraryMap::LoadLibrary(_T("winhttp.dll")), nResult);
		if(sMessage.IsEmpty())
			return FALSE;
		if(psMessage)
			*psMessage = sMessage;
		return TRUE;
	}
	static BOOL IsWinInetResult(HRESULT nResult, CString* psMessage = NULL)
	{
		if(HRESULT_FACILITY(nResult) != FACILITY_WIN32)
			return FALSE;
		const LONG nCode = HRESULT_CODE(nResult);
		if(nCode < 12000 || nCode >= 12200) // INTERNET_ERROR_BASE, INTERNET_ERROR_LAST
			return FALSE;
		const CString sMessage = AtlFormatSystemMessage(CDataLibraryMap::LoadLibrary(_T("wininet.dll")), nResult);
		if(sMessage.IsEmpty())
			return FALSE;
		if(psMessage)
			*psMessage = sMessage;
		return TRUE;
	}
	static CString LookupSystemIdentifier(HRESULT nValue)
	{
		if(HRESULT_SEVERITY(nValue) == SEVERITY_ERROR && HRESULT_FACILITY(nValue) == FACILITY_WIN32)
			nValue = HRESULT_CODE(nValue);
		if(nValue < 0 || nValue >= 16384)
			return _T("");
		static const struct { HRESULT nValue; LPCSTR pszName; } g_pMap[] = 
		{
			#define A(x) { x, #x },
			A(ERROR_SUCCESS)
			A(ERROR_INVALID_FUNCTION)
			A(ERROR_FILE_NOT_FOUND)
			A(ERROR_PATH_NOT_FOUND)
			A(ERROR_TOO_MANY_OPEN_FILES)
			A(ERROR_ACCESS_DENIED)
			A(ERROR_INVALID_HANDLE)
			A(ERROR_ARENA_TRASHED)
			A(ERROR_NOT_ENOUGH_MEMORY)
			A(ERROR_INVALID_BLOCK)
			A(ERROR_BAD_ENVIRONMENT)
			A(ERROR_BAD_FORMAT)
			A(ERROR_INVALID_ACCESS)
			A(ERROR_INVALID_DATA)
			A(ERROR_OUTOFMEMORY)
			A(ERROR_INVALID_DRIVE)
			A(ERROR_CURRENT_DIRECTORY)
			A(ERROR_NOT_SAME_DEVICE)
			A(ERROR_NO_MORE_FILES)
			A(ERROR_WRITE_PROTECT)
			A(ERROR_BAD_UNIT)
			A(ERROR_NOT_READY)
			A(ERROR_BAD_COMMAND)
			A(ERROR_CRC)
			A(ERROR_BAD_LENGTH)
			A(ERROR_SEEK)
			A(ERROR_NOT_DOS_DISK)
			A(ERROR_SECTOR_NOT_FOUND)
			A(ERROR_OUT_OF_PAPER)
			A(ERROR_WRITE_FAULT)
			A(ERROR_READ_FAULT)
			A(ERROR_GEN_FAILURE)
			A(ERROR_SHARING_VIOLATION)
			A(ERROR_LOCK_VIOLATION)
			A(ERROR_WRONG_DISK)
			A(ERROR_SHARING_BUFFER_EXCEEDED)
			A(ERROR_HANDLE_EOF)
			A(ERROR_HANDLE_DISK_FULL)
			A(ERROR_NOT_SUPPORTED)
			A(ERROR_REM_NOT_LIST)
			A(ERROR_DUP_NAME)
			A(ERROR_BAD_NETPATH)
			A(ERROR_NETWORK_BUSY)
			A(ERROR_DEV_NOT_EXIST)
			A(ERROR_TOO_MANY_CMDS)
			A(ERROR_ADAP_HDW_ERR)
			A(ERROR_BAD_NET_RESP)
			A(ERROR_UNEXP_NET_ERR)
			A(ERROR_BAD_REM_ADAP)
			A(ERROR_PRINTQ_FULL)
			A(ERROR_NO_SPOOL_SPACE)
			A(ERROR_PRINT_CANCELLED)
			A(ERROR_NETNAME_DELETED)
			A(ERROR_NETWORK_ACCESS_DENIED)
			A(ERROR_BAD_DEV_TYPE)
			A(ERROR_BAD_NET_NAME)
			A(ERROR_TOO_MANY_NAMES)
			A(ERROR_TOO_MANY_SESS)
			A(ERROR_SHARING_PAUSED)
			A(ERROR_REQ_NOT_ACCEP)
			A(ERROR_REDIR_PAUSED)
			A(ERROR_FILE_EXISTS)
			A(ERROR_CANNOT_MAKE)
			A(ERROR_FAIL_I24)
			A(ERROR_OUT_OF_STRUCTURES)
			A(ERROR_ALREADY_ASSIGNED)
			A(ERROR_INVALID_PASSWORD)
			A(ERROR_INVALID_PARAMETER)
			A(ERROR_NET_WRITE_FAULT)
			A(ERROR_NO_PROC_SLOTS)
			A(ERROR_TOO_MANY_SEMAPHORES)
			A(ERROR_EXCL_SEM_ALREADY_OWNED)
			A(ERROR_SEM_IS_SET)
			A(ERROR_TOO_MANY_SEM_REQUESTS)
			A(ERROR_INVALID_AT_INTERRUPT_TIME)
			A(ERROR_SEM_OWNER_DIED)
			A(ERROR_SEM_USER_LIMIT)
			A(ERROR_DISK_CHANGE)
			A(ERROR_DRIVE_LOCKED)
			A(ERROR_BROKEN_PIPE)
			A(ERROR_OPEN_FAILED)
			A(ERROR_BUFFER_OVERFLOW)
			A(ERROR_DISK_FULL)
			A(ERROR_NO_MORE_SEARCH_HANDLES)
			A(ERROR_INVALID_TARGET_HANDLE)
			A(ERROR_INVALID_CATEGORY)
			A(ERROR_INVALID_VERIFY_SWITCH)
			A(ERROR_BAD_DRIVER_LEVEL)
			A(ERROR_CALL_NOT_IMPLEMENTED)
			A(ERROR_SEM_TIMEOUT)
			A(ERROR_INSUFFICIENT_BUFFER)
			A(ERROR_INVALID_NAME)
			A(ERROR_INVALID_LEVEL)
			A(ERROR_NO_VOLUME_LABEL)
			A(ERROR_MOD_NOT_FOUND)
			A(ERROR_PROC_NOT_FOUND)
			A(ERROR_WAIT_NO_CHILDREN)
			A(ERROR_CHILD_NOT_COMPLETE)
			A(ERROR_DIRECT_ACCESS_HANDLE)
			A(ERROR_NEGATIVE_SEEK)
			A(ERROR_SEEK_ON_DEVICE)
			A(ERROR_IS_JOIN_TARGET)
			A(ERROR_IS_JOINED)
			A(ERROR_IS_SUBSTED)
			A(ERROR_NOT_JOINED)
			A(ERROR_NOT_SUBSTED)
			A(ERROR_JOIN_TO_JOIN)
			A(ERROR_SUBST_TO_SUBST)
			A(ERROR_JOIN_TO_SUBST)
			A(ERROR_SUBST_TO_JOIN)
			A(ERROR_BUSY_DRIVE)
			A(ERROR_SAME_DRIVE)
			A(ERROR_DIR_NOT_ROOT)
			A(ERROR_DIR_NOT_EMPTY)
			A(ERROR_IS_SUBST_PATH)
			A(ERROR_IS_JOIN_PATH)
			A(ERROR_PATH_BUSY)
			A(ERROR_IS_SUBST_TARGET)
			A(ERROR_SYSTEM_TRACE)
			A(ERROR_INVALID_EVENT_COUNT)
			A(ERROR_TOO_MANY_MUXWAITERS)
			A(ERROR_INVALID_LIST_FORMAT)
			A(ERROR_LABEL_TOO_LONG)
			A(ERROR_TOO_MANY_TCBS)
			A(ERROR_SIGNAL_REFUSED)
			A(ERROR_DISCARDED)
			A(ERROR_NOT_LOCKED)
			A(ERROR_BAD_THREADID_ADDR)
			A(ERROR_BAD_ARGUMENTS)
			A(ERROR_BAD_PATHNAME)
			A(ERROR_SIGNAL_PENDING)
			A(ERROR_MAX_THRDS_REACHED)
			A(ERROR_LOCK_FAILED)
			A(ERROR_BUSY)
			A(ERROR_CANCEL_VIOLATION)
			A(ERROR_ATOMIC_LOCKS_NOT_SUPPORTED)
			A(ERROR_INVALID_SEGMENT_NUMBER)
			A(ERROR_INVALID_ORDINAL)
			A(ERROR_ALREADY_EXISTS)
			A(ERROR_INVALID_FLAG_NUMBER)
			A(ERROR_SEM_NOT_FOUND)
			A(ERROR_INVALID_STARTING_CODESEG)
			A(ERROR_INVALID_STACKSEG)
			A(ERROR_INVALID_MODULETYPE)
			A(ERROR_INVALID_EXE_SIGNATURE)
			A(ERROR_EXE_MARKED_INVALID)
			A(ERROR_BAD_EXE_FORMAT)
			A(ERROR_ITERATED_DATA_EXCEEDS_64k)
			A(ERROR_INVALID_MINALLOCSIZE)
			A(ERROR_DYNLINK_FROM_INVALID_RING)
			A(ERROR_IOPL_NOT_ENABLED)
			A(ERROR_INVALID_SEGDPL)
			A(ERROR_AUTODATASEG_EXCEEDS_64k)
			A(ERROR_RING2SEG_MUST_BE_MOVABLE)
			A(ERROR_RELOC_CHAIN_XEEDS_SEGLIM)
			A(ERROR_INFLOOP_IN_RELOC_CHAIN)
			A(ERROR_ENVVAR_NOT_FOUND)
			A(ERROR_NO_SIGNAL_SENT)
			A(ERROR_FILENAME_EXCED_RANGE)
			A(ERROR_RING2_STACK_IN_USE)
			A(ERROR_META_EXPANSION_TOO_LONG)
			A(ERROR_INVALID_SIGNAL_NUMBER)
			A(ERROR_THREAD_1_INACTIVE)
			A(ERROR_LOCKED)
			A(ERROR_TOO_MANY_MODULES)
			A(ERROR_NESTING_NOT_ALLOWED)
			A(ERROR_EXE_MACHINE_TYPE_MISMATCH)
			A(ERROR_EXE_CANNOT_MODIFY_SIGNED_BINARY)
			A(ERROR_EXE_CANNOT_MODIFY_STRONG_SIGNED_BINARY)
			A(ERROR_FILE_CHECKED_OUT)
			A(ERROR_CHECKOUT_REQUIRED)
			A(ERROR_BAD_FILE_TYPE)
			A(ERROR_FILE_TOO_LARGE)
			A(ERROR_FORMS_AUTH_REQUIRED)
			A(ERROR_VIRUS_INFECTED)
			A(ERROR_VIRUS_DELETED)
			A(ERROR_PIPE_LOCAL)
			A(ERROR_BAD_PIPE)
			A(ERROR_PIPE_BUSY)
			A(ERROR_NO_DATA)
			A(ERROR_PIPE_NOT_CONNECTED)
			A(ERROR_MORE_DATA)
			A(ERROR_VC_DISCONNECTED)
			A(ERROR_INVALID_EA_NAME)
			A(ERROR_EA_LIST_INCONSISTENT)
			A(ERROR_NO_MORE_ITEMS)
			A(ERROR_CANNOT_COPY)
			A(ERROR_DIRECTORY)
			A(ERROR_EAS_DIDNT_FIT)
			A(ERROR_EA_FILE_CORRUPT)
			A(ERROR_EA_TABLE_FULL)
			A(ERROR_INVALID_EA_HANDLE)
			A(ERROR_EAS_NOT_SUPPORTED)
			A(ERROR_NOT_OWNER)
			A(ERROR_TOO_MANY_POSTS)
			A(ERROR_PARTIAL_COPY)
			A(ERROR_OPLOCK_NOT_GRANTED)
			A(ERROR_INVALID_OPLOCK_PROTOCOL)
			A(ERROR_DISK_TOO_FRAGMENTED)
			A(ERROR_DELETE_PENDING)
			A(ERROR_INCOMPATIBLE_WITH_GLOBAL_SHORT_NAME_REGISTRY_SETTING)
			A(ERROR_SHORT_NAMES_NOT_ENABLED_ON_VOLUME)
			A(ERROR_SECURITY_STREAM_IS_INCONSISTENT)
			A(ERROR_INVALID_LOCK_RANGE)
			A(ERROR_IMAGE_SUBSYSTEM_NOT_PRESENT)
			A(ERROR_NOTIFICATION_GUID_ALREADY_DEFINED)
			A(ERROR_MR_MID_NOT_FOUND)
			A(ERROR_SCOPE_NOT_FOUND)
			A(ERROR_FAIL_NOACTION_REBOOT)
			A(ERROR_FAIL_SHUTDOWN)
			A(ERROR_FAIL_RESTART)
			A(ERROR_MAX_SESSIONS_REACHED)
			A(ERROR_THREAD_MODE_ALREADY_BACKGROUND)
			A(ERROR_THREAD_MODE_NOT_BACKGROUND)
			A(ERROR_PROCESS_MODE_ALREADY_BACKGROUND)
			A(ERROR_PROCESS_MODE_NOT_BACKGROUND)
			A(ERROR_INVALID_ADDRESS)
			A(ERROR_USER_PROFILE_LOAD)
			A(ERROR_ARITHMETIC_OVERFLOW)
			A(ERROR_PIPE_CONNECTED)
			A(ERROR_PIPE_LISTENING)
			A(ERROR_VERIFIER_STOP)
			A(ERROR_ABIOS_ERROR)
			A(ERROR_WX86_WARNING)
			A(ERROR_WX86_ERROR)
			A(ERROR_TIMER_NOT_CANCELED)
			A(ERROR_UNWIND)
			A(ERROR_BAD_STACK)
			A(ERROR_INVALID_UNWIND_TARGET)
			A(ERROR_INVALID_PORT_ATTRIBUTES)
			A(ERROR_PORT_MESSAGE_TOO_LONG)
			A(ERROR_INVALID_QUOTA_LOWER)
			A(ERROR_DEVICE_ALREADY_ATTACHED)
			A(ERROR_INSTRUCTION_MISALIGNMENT)
			A(ERROR_PROFILING_NOT_STARTED)
			A(ERROR_PROFILING_NOT_STOPPED)
			A(ERROR_COULD_NOT_INTERPRET)
			A(ERROR_PROFILING_AT_LIMIT)
			A(ERROR_CANT_WAIT)
			A(ERROR_CANT_TERMINATE_SELF)
			A(ERROR_UNEXPECTED_MM_CREATE_ERR)
			A(ERROR_UNEXPECTED_MM_MAP_ERROR)
			A(ERROR_UNEXPECTED_MM_EXTEND_ERR)
			A(ERROR_BAD_FUNCTION_TABLE)
			A(ERROR_NO_GUID_TRANSLATION)
			A(ERROR_INVALID_LDT_SIZE)
			A(ERROR_INVALID_LDT_OFFSET)
			A(ERROR_INVALID_LDT_DESCRIPTOR)
			A(ERROR_TOO_MANY_THREADS)
			A(ERROR_THREAD_NOT_IN_PROCESS)
			A(ERROR_PAGEFILE_QUOTA_EXCEEDED)
			A(ERROR_LOGON_SERVER_CONFLICT)
			A(ERROR_SYNCHRONIZATION_REQUIRED)
			A(ERROR_NET_OPEN_FAILED)
			A(ERROR_IO_PRIVILEGE_FAILED)
			A(ERROR_CONTROL_C_EXIT)
			A(ERROR_MISSING_SYSTEMFILE)
			A(ERROR_UNHANDLED_EXCEPTION)
			A(ERROR_APP_INIT_FAILURE)
			A(ERROR_PAGEFILE_CREATE_FAILED)
			A(ERROR_INVALID_IMAGE_HASH)
			A(ERROR_NO_PAGEFILE)
			A(ERROR_ILLEGAL_FLOAT_CONTEXT)
			A(ERROR_NO_EVENT_PAIR)
			A(ERROR_DOMAIN_CTRLR_CONFIG_ERROR)
			A(ERROR_ILLEGAL_CHARACTER)
			A(ERROR_UNDEFINED_CHARACTER)
			A(ERROR_FLOPPY_VOLUME)
			A(ERROR_BIOS_FAILED_TO_CONNECT_INTERRUPT)
			A(ERROR_BACKUP_CONTROLLER)
			A(ERROR_MUTANT_LIMIT_EXCEEDED)
			A(ERROR_FS_DRIVER_REQUIRED)
			A(ERROR_CANNOT_LOAD_REGISTRY_FILE)
			A(ERROR_DEBUG_ATTACH_FAILED)
			A(ERROR_SYSTEM_PROCESS_TERMINATED)
			A(ERROR_DATA_NOT_ACCEPTED)
			A(ERROR_VDM_HARD_ERROR)
			A(ERROR_DRIVER_CANCEL_TIMEOUT)
			A(ERROR_REPLY_MESSAGE_MISMATCH)
			A(ERROR_LOST_WRITEBEHIND_DATA)
			A(ERROR_CLIENT_SERVER_PARAMETERS_INVALID)
			A(ERROR_NOT_TINY_STREAM)
			A(ERROR_STACK_OVERFLOW_READ)
			A(ERROR_CONVERT_TO_LARGE)
			A(ERROR_FOUND_OUT_OF_SCOPE)
			A(ERROR_ALLOCATE_BUCKET)
			A(ERROR_MARSHALL_OVERFLOW)
			A(ERROR_INVALID_VARIANT)
			A(ERROR_BAD_COMPRESSION_BUFFER)
			A(ERROR_AUDIT_FAILED)
			A(ERROR_TIMER_RESOLUTION_NOT_SET)
			A(ERROR_INSUFFICIENT_LOGON_INFO)
			A(ERROR_BAD_DLL_ENTRYPOINT)
			A(ERROR_BAD_SERVICE_ENTRYPOINT)
			A(ERROR_IP_ADDRESS_CONFLICT1)
			A(ERROR_IP_ADDRESS_CONFLICT2)
			A(ERROR_REGISTRY_QUOTA_LIMIT)
			A(ERROR_NO_CALLBACK_ACTIVE)
			A(ERROR_PWD_TOO_SHORT)
			A(ERROR_PWD_TOO_RECENT)
			A(ERROR_PWD_HISTORY_CONFLICT)
			A(ERROR_UNSUPPORTED_COMPRESSION)
			A(ERROR_INVALID_HW_PROFILE)
			A(ERROR_INVALID_PLUGPLAY_DEVICE_PATH)
			A(ERROR_QUOTA_LIST_INCONSISTENT)
			A(ERROR_EVALUATION_EXPIRATION)
			A(ERROR_ILLEGAL_DLL_RELOCATION)
			A(ERROR_DLL_INIT_FAILED_LOGOFF)
			A(ERROR_VALIDATE_CONTINUE)
			A(ERROR_NO_MORE_MATCHES)
			A(ERROR_RANGE_LIST_CONFLICT)
			A(ERROR_SERVER_SID_MISMATCH)
			A(ERROR_CANT_ENABLE_DENY_ONLY)
			A(ERROR_FLOAT_MULTIPLE_FAULTS)
			A(ERROR_FLOAT_MULTIPLE_TRAPS)
			A(ERROR_NOINTERFACE)
			A(ERROR_DRIVER_FAILED_SLEEP)
			A(ERROR_CORRUPT_SYSTEM_FILE)
			A(ERROR_COMMITMENT_MINIMUM)
			A(ERROR_PNP_RESTART_ENUMERATION)
			A(ERROR_SYSTEM_IMAGE_BAD_SIGNATURE)
			A(ERROR_PNP_REBOOT_REQUIRED)
			A(ERROR_INSUFFICIENT_POWER)
			A(ERROR_MULTIPLE_FAULT_VIOLATION)
			A(ERROR_SYSTEM_SHUTDOWN)
			A(ERROR_PORT_NOT_SET)
			A(ERROR_DS_VERSION_CHECK_FAILURE)
			A(ERROR_RANGE_NOT_FOUND)
			A(ERROR_NOT_SAFE_MODE_DRIVER)
			A(ERROR_FAILED_DRIVER_ENTRY)
			A(ERROR_DEVICE_ENUMERATION_ERROR)
			A(ERROR_MOUNT_POINT_NOT_RESOLVED)
			A(ERROR_INVALID_DEVICE_OBJECT_PARAMETER)
			A(ERROR_MCA_OCCURED)
			A(ERROR_DRIVER_DATABASE_ERROR)
			A(ERROR_SYSTEM_HIVE_TOO_LARGE)
			A(ERROR_DRIVER_FAILED_PRIOR_UNLOAD)
			A(ERROR_VOLSNAP_PREPARE_HIBERNATE)
			A(ERROR_HIBERNATION_FAILURE)
			A(ERROR_FILE_SYSTEM_LIMITATION)
			A(ERROR_ASSERTION_FAILURE)
			A(ERROR_ACPI_ERROR)
			A(ERROR_WOW_ASSERTION)
			A(ERROR_PNP_BAD_MPS_TABLE)
			A(ERROR_PNP_TRANSLATION_FAILED)
			A(ERROR_PNP_IRQ_TRANSLATION_FAILED)
			A(ERROR_PNP_INVALID_ID)
			A(ERROR_WAKE_SYSTEM_DEBUGGER)
			A(ERROR_HANDLES_CLOSED)
			A(ERROR_EXTRANEOUS_INFORMATION)
			A(ERROR_RXACT_COMMIT_NECESSARY)
			A(ERROR_MEDIA_CHECK)
			A(ERROR_GUID_SUBSTITUTION_MADE)
			A(ERROR_STOPPED_ON_SYMLINK)
			A(ERROR_LONGJUMP)
			A(ERROR_PLUGPLAY_QUERY_VETOED)
			A(ERROR_UNWIND_CONSOLIDATE)
			A(ERROR_REGISTRY_HIVE_RECOVERED)
			A(ERROR_DLL_MIGHT_BE_INSECURE)
			A(ERROR_DLL_MIGHT_BE_INCOMPATIBLE)
			A(ERROR_DBG_EXCEPTION_NOT_HANDLED)
			A(ERROR_DBG_REPLY_LATER)
			A(ERROR_DBG_UNABLE_TO_PROVIDE_HANDLE)
			A(ERROR_DBG_TERMINATE_THREAD)
			A(ERROR_DBG_TERMINATE_PROCESS)
			A(ERROR_DBG_CONTROL_C)
			A(ERROR_DBG_PRINTEXCEPTION_C)
			A(ERROR_DBG_RIPEXCEPTION)
			A(ERROR_DBG_CONTROL_BREAK)
			A(ERROR_DBG_COMMAND_EXCEPTION)
			A(ERROR_OBJECT_NAME_EXISTS)
			A(ERROR_THREAD_WAS_SUSPENDED)
			A(ERROR_IMAGE_NOT_AT_BASE)
			A(ERROR_RXACT_STATE_CREATED)
			A(ERROR_SEGMENT_NOTIFICATION)
			A(ERROR_BAD_CURRENT_DIRECTORY)
			A(ERROR_FT_READ_RECOVERY_FROM_BACKUP)
			A(ERROR_FT_WRITE_RECOVERY)
			A(ERROR_IMAGE_MACHINE_TYPE_MISMATCH)
			A(ERROR_RECEIVE_PARTIAL)
			A(ERROR_RECEIVE_EXPEDITED)
			A(ERROR_RECEIVE_PARTIAL_EXPEDITED)
			A(ERROR_EVENT_DONE)
			A(ERROR_EVENT_PENDING)
			A(ERROR_CHECKING_FILE_SYSTEM)
			A(ERROR_FATAL_APP_EXIT)
			A(ERROR_PREDEFINED_HANDLE)
			A(ERROR_WAS_UNLOCKED)
			A(ERROR_SERVICE_NOTIFICATION)
			A(ERROR_WAS_LOCKED)
			A(ERROR_LOG_HARD_ERROR)
			A(ERROR_ALREADY_WIN32)
			A(ERROR_IMAGE_MACHINE_TYPE_MISMATCH_EXE)
			A(ERROR_NO_YIELD_PERFORMED)
			A(ERROR_TIMER_RESUME_IGNORED)
			A(ERROR_ARBITRATION_UNHANDLED)
			A(ERROR_CARDBUS_NOT_SUPPORTED)
			A(ERROR_MP_PROCESSOR_MISMATCH)
			A(ERROR_HIBERNATED)
			A(ERROR_RESUME_HIBERNATION)
			A(ERROR_FIRMWARE_UPDATED)
			A(ERROR_DRIVERS_LEAKING_LOCKED_PAGES)
			A(ERROR_WAKE_SYSTEM)
			A(ERROR_WAIT_1)
			A(ERROR_WAIT_2)
			A(ERROR_WAIT_3)
			A(ERROR_WAIT_63)
			A(ERROR_ABANDONED_WAIT_0)
			A(ERROR_ABANDONED_WAIT_63)
			A(ERROR_USER_APC)
			A(ERROR_KERNEL_APC)
			A(ERROR_ALERTED)
			A(ERROR_ELEVATION_REQUIRED)
			A(ERROR_REPARSE)
			A(ERROR_OPLOCK_BREAK_IN_PROGRESS)
			A(ERROR_VOLUME_MOUNTED)
			A(ERROR_RXACT_COMMITTED)
			A(ERROR_NOTIFY_CLEANUP)
			A(ERROR_PRIMARY_TRANSPORT_CONNECT_FAILED)
			A(ERROR_PAGE_FAULT_TRANSITION)
			A(ERROR_PAGE_FAULT_DEMAND_ZERO)
			A(ERROR_PAGE_FAULT_COPY_ON_WRITE)
			A(ERROR_PAGE_FAULT_GUARD_PAGE)
			A(ERROR_PAGE_FAULT_PAGING_FILE)
			A(ERROR_CACHE_PAGE_LOCKED)
			A(ERROR_CRASH_DUMP)
			A(ERROR_BUFFER_ALL_ZEROS)
			A(ERROR_REPARSE_OBJECT)
			A(ERROR_RESOURCE_REQUIREMENTS_CHANGED)
			A(ERROR_TRANSLATION_COMPLETE)
			A(ERROR_NOTHING_TO_TERMINATE)
			A(ERROR_PROCESS_NOT_IN_JOB)
			A(ERROR_PROCESS_IN_JOB)
			A(ERROR_VOLSNAP_HIBERNATE_READY)
			A(ERROR_FSFILTER_OP_COMPLETED_SUCCESSFULLY)
			A(ERROR_INTERRUPT_VECTOR_ALREADY_CONNECTED)
			A(ERROR_INTERRUPT_STILL_CONNECTED)
			A(ERROR_WAIT_FOR_OPLOCK)
			A(ERROR_DBG_EXCEPTION_HANDLED)
			A(ERROR_DBG_CONTINUE)
			A(ERROR_CALLBACK_POP_STACK)
			A(ERROR_COMPRESSION_DISABLED)
			A(ERROR_CANTFETCHBACKWARDS)
			A(ERROR_CANTSCROLLBACKWARDS)
			A(ERROR_ROWSNOTRELEASED)
			A(ERROR_BAD_ACCESSOR_FLAGS)
			A(ERROR_ERRORS_ENCOUNTERED)
			A(ERROR_NOT_CAPABLE)
			A(ERROR_REQUEST_OUT_OF_SEQUENCE)
			A(ERROR_VERSION_PARSE_ERROR)
			A(ERROR_BADSTARTPOSITION)
			A(ERROR_MEMORY_HARDWARE)
			A(ERROR_DISK_REPAIR_DISABLED)
			A(ERROR_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE)
			A(ERROR_SYSTEM_POWERSTATE_TRANSITION)
			A(ERROR_SYSTEM_POWERSTATE_COMPLEX_TRANSITION)
			A(ERROR_MCA_EXCEPTION)
			A(ERROR_ACCESS_AUDIT_BY_POLICY)
			A(ERROR_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY)
			A(ERROR_ABANDON_HIBERFILE)
			A(ERROR_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED)
			A(ERROR_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR)
			A(ERROR_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR)
			A(ERROR_BAD_MCFG_TABLE)
			A(ERROR_OPLOCK_SWITCHED_TO_NEW_HANDLE)
			A(ERROR_CANNOT_GRANT_REQUESTED_OPLOCK)
			A(ERROR_CANNOT_BREAK_OPLOCK)
			A(ERROR_OPLOCK_HANDLE_CLOSED)
			A(ERROR_NO_ACE_CONDITION)
			A(ERROR_INVALID_ACE_CONDITION)
			A(ERROR_EA_ACCESS_DENIED)
			A(ERROR_OPERATION_ABORTED)
			A(ERROR_IO_INCOMPLETE)
			A(ERROR_IO_PENDING)
			A(ERROR_NOACCESS)
			A(ERROR_SWAPERROR)
			A(ERROR_STACK_OVERFLOW)
			A(ERROR_INVALID_MESSAGE)
			A(ERROR_CAN_NOT_COMPLETE)
			A(ERROR_INVALID_FLAGS)
			A(ERROR_UNRECOGNIZED_VOLUME)
			A(ERROR_FILE_INVALID)
			A(ERROR_FULLSCREEN_MODE)
			A(ERROR_NO_TOKEN)
			A(ERROR_BADDB)
			A(ERROR_BADKEY)
			A(ERROR_CANTOPEN)
			A(ERROR_CANTREAD)
			A(ERROR_CANTWRITE)
			A(ERROR_REGISTRY_RECOVERED)
			A(ERROR_REGISTRY_CORRUPT)
			A(ERROR_REGISTRY_IO_FAILED)
			A(ERROR_NOT_REGISTRY_FILE)
			A(ERROR_KEY_DELETED)
			A(ERROR_NO_LOG_SPACE)
			A(ERROR_KEY_HAS_CHILDREN)
			A(ERROR_CHILD_MUST_BE_VOLATILE)
			A(ERROR_NOTIFY_ENUM_DIR)
			A(ERROR_DEPENDENT_SERVICES_RUNNING)
			A(ERROR_INVALID_SERVICE_CONTROL)
			A(ERROR_SERVICE_REQUEST_TIMEOUT)
			A(ERROR_SERVICE_NO_THREAD)
			A(ERROR_SERVICE_DATABASE_LOCKED)
			A(ERROR_SERVICE_ALREADY_RUNNING)
			A(ERROR_INVALID_SERVICE_ACCOUNT)
			A(ERROR_SERVICE_DISABLED)
			A(ERROR_CIRCULAR_DEPENDENCY)
			A(ERROR_SERVICE_DOES_NOT_EXIST)
			A(ERROR_SERVICE_CANNOT_ACCEPT_CTRL)
			A(ERROR_SERVICE_NOT_ACTIVE)
			A(ERROR_FAILED_SERVICE_CONTROLLER_CONNECT)
			A(ERROR_EXCEPTION_IN_SERVICE)
			A(ERROR_DATABASE_DOES_NOT_EXIST)
			A(ERROR_SERVICE_SPECIFIC_ERROR)
			A(ERROR_PROCESS_ABORTED)
			A(ERROR_SERVICE_DEPENDENCY_FAIL)
			A(ERROR_SERVICE_LOGON_FAILED)
			A(ERROR_SERVICE_START_HANG)
			A(ERROR_INVALID_SERVICE_LOCK)
			A(ERROR_SERVICE_MARKED_FOR_DELETE)
			A(ERROR_SERVICE_EXISTS)
			A(ERROR_ALREADY_RUNNING_LKG)
			A(ERROR_SERVICE_DEPENDENCY_DELETED)
			A(ERROR_BOOT_ALREADY_ACCEPTED)
			A(ERROR_SERVICE_NEVER_STARTED)
			A(ERROR_DUPLICATE_SERVICE_NAME)
			A(ERROR_DIFFERENT_SERVICE_ACCOUNT)
			A(ERROR_CANNOT_DETECT_DRIVER_FAILURE)
			A(ERROR_CANNOT_DETECT_PROCESS_ABORT)
			A(ERROR_NO_RECOVERY_PROGRAM)
			A(ERROR_SERVICE_NOT_IN_EXE)
			A(ERROR_NOT_SAFEBOOT_SERVICE)
			A(ERROR_END_OF_MEDIA)
			A(ERROR_FILEMARK_DETECTED)
			A(ERROR_BEGINNING_OF_MEDIA)
			A(ERROR_SETMARK_DETECTED)
			A(ERROR_NO_DATA_DETECTED)
			A(ERROR_PARTITION_FAILURE)
			A(ERROR_INVALID_BLOCK_LENGTH)
			A(ERROR_DEVICE_NOT_PARTITIONED)
			A(ERROR_UNABLE_TO_LOCK_MEDIA)
			A(ERROR_UNABLE_TO_UNLOAD_MEDIA)
			A(ERROR_MEDIA_CHANGED)
			A(ERROR_BUS_RESET)
			A(ERROR_NO_MEDIA_IN_DRIVE)
			A(ERROR_NO_UNICODE_TRANSLATION)
			A(ERROR_DLL_INIT_FAILED)
			A(ERROR_SHUTDOWN_IN_PROGRESS)
			A(ERROR_NO_SHUTDOWN_IN_PROGRESS)
			A(ERROR_IO_DEVICE)
			A(ERROR_SERIAL_NO_DEVICE)
			A(ERROR_IRQ_BUSY)
			A(ERROR_MORE_WRITES)
			A(ERROR_COUNTER_TIMEOUT)
			A(ERROR_FLOPPY_ID_MARK_NOT_FOUND)
			A(ERROR_FLOPPY_WRONG_CYLINDER)
			A(ERROR_FLOPPY_UNKNOWN_ERROR)
			A(ERROR_FLOPPY_BAD_REGISTERS)
			A(ERROR_DISK_RECALIBRATE_FAILED)
			A(ERROR_DISK_OPERATION_FAILED)
			A(ERROR_DISK_RESET_FAILED)
			A(ERROR_EOM_OVERFLOW)
			A(ERROR_NOT_ENOUGH_SERVER_MEMORY)
			A(ERROR_POSSIBLE_DEADLOCK)
			A(ERROR_MAPPED_ALIGNMENT)
			A(ERROR_SET_POWER_STATE_VETOED)
			A(ERROR_SET_POWER_STATE_FAILED)
			A(ERROR_TOO_MANY_LINKS)
			A(ERROR_OLD_WIN_VERSION)
			A(ERROR_APP_WRONG_OS)
			A(ERROR_SINGLE_INSTANCE_APP)
			A(ERROR_RMODE_APP)
			A(ERROR_INVALID_DLL)
			A(ERROR_NO_ASSOCIATION)
			A(ERROR_DDE_FAIL)
			A(ERROR_DLL_NOT_FOUND)
			A(ERROR_NO_MORE_USER_HANDLES)
			A(ERROR_MESSAGE_SYNC_ONLY)
			A(ERROR_SOURCE_ELEMENT_EMPTY)
			A(ERROR_DESTINATION_ELEMENT_FULL)
			A(ERROR_ILLEGAL_ELEMENT_ADDRESS)
			A(ERROR_MAGAZINE_NOT_PRESENT)
			A(ERROR_DEVICE_REINITIALIZATION_NEEDED)
			A(ERROR_DEVICE_REQUIRES_CLEANING)
			A(ERROR_DEVICE_DOOR_OPEN)
			A(ERROR_DEVICE_NOT_CONNECTED)
			A(ERROR_NOT_FOUND)
			A(ERROR_NO_MATCH)
			A(ERROR_SET_NOT_FOUND)
			A(ERROR_POINT_NOT_FOUND)
			A(ERROR_NO_TRACKING_SERVICE)
			A(ERROR_NO_VOLUME_ID)
			A(ERROR_UNABLE_TO_REMOVE_REPLACED)
			A(ERROR_UNABLE_TO_MOVE_REPLACEMENT)
			A(ERROR_UNABLE_TO_MOVE_REPLACEMENT_2)
			A(ERROR_JOURNAL_DELETE_IN_PROGRESS)
			A(ERROR_JOURNAL_NOT_ACTIVE)
			A(ERROR_POTENTIAL_FILE_FOUND)
			A(ERROR_JOURNAL_ENTRY_DELETED)
			A(ERROR_SHUTDOWN_IS_SCHEDULED)
			A(ERROR_SHUTDOWN_USERS_LOGGED_ON)
			A(ERROR_BAD_DEVICE)
			A(ERROR_CONNECTION_UNAVAIL)
			A(ERROR_DEVICE_ALREADY_REMEMBERED)
			A(ERROR_NO_NET_OR_BAD_PATH)
			A(ERROR_BAD_PROVIDER)
			A(ERROR_CANNOT_OPEN_PROFILE)
			A(ERROR_BAD_PROFILE)
			A(ERROR_NOT_CONTAINER)
			A(ERROR_EXTENDED_ERROR)
			A(ERROR_INVALID_GROUPNAME)
			A(ERROR_INVALID_COMPUTERNAME)
			A(ERROR_INVALID_EVENTNAME)
			A(ERROR_INVALID_DOMAINNAME)
			A(ERROR_INVALID_SERVICENAME)
			A(ERROR_INVALID_NETNAME)
			A(ERROR_INVALID_SHARENAME)
			A(ERROR_INVALID_PASSWORDNAME)
			A(ERROR_INVALID_MESSAGENAME)
			A(ERROR_INVALID_MESSAGEDEST)
			A(ERROR_SESSION_CREDENTIAL_CONFLICT)
			A(ERROR_REMOTE_SESSION_LIMIT_EXCEEDED)
			A(ERROR_DUP_DOMAINNAME)
			A(ERROR_NO_NETWORK)
			A(ERROR_CANCELLED)
			A(ERROR_USER_MAPPED_FILE)
			A(ERROR_CONNECTION_REFUSED)
			A(ERROR_GRACEFUL_DISCONNECT)
			A(ERROR_ADDRESS_ALREADY_ASSOCIATED)
			A(ERROR_ADDRESS_NOT_ASSOCIATED)
			A(ERROR_CONNECTION_INVALID)
			A(ERROR_CONNECTION_ACTIVE)
			A(ERROR_NETWORK_UNREACHABLE)
			A(ERROR_HOST_UNREACHABLE)
			A(ERROR_PROTOCOL_UNREACHABLE)
			A(ERROR_PORT_UNREACHABLE)
			A(ERROR_REQUEST_ABORTED)
			A(ERROR_CONNECTION_ABORTED)
			A(ERROR_RETRY)
			A(ERROR_CONNECTION_COUNT_LIMIT)
			A(ERROR_LOGIN_TIME_RESTRICTION)
			A(ERROR_LOGIN_WKSTA_RESTRICTION)
			A(ERROR_INCORRECT_ADDRESS)
			A(ERROR_ALREADY_REGISTERED)
			A(ERROR_SERVICE_NOT_FOUND)
			A(ERROR_NOT_AUTHENTICATED)
			A(ERROR_NOT_LOGGED_ON)
			A(ERROR_CONTINUE)
			A(ERROR_ALREADY_INITIALIZED)
			A(ERROR_NO_MORE_DEVICES)
			A(ERROR_NO_SUCH_SITE)
			A(ERROR_DOMAIN_CONTROLLER_EXISTS)
			A(ERROR_ONLY_IF_CONNECTED)
			A(ERROR_OVERRIDE_NOCHANGES)
			A(ERROR_BAD_USER_PROFILE)
			A(ERROR_NOT_SUPPORTED_ON_SBS)
			A(ERROR_SERVER_SHUTDOWN_IN_PROGRESS)
			A(ERROR_HOST_DOWN)
			A(ERROR_NON_ACCOUNT_SID)
			A(ERROR_NON_DOMAIN_SID)
			A(ERROR_APPHELP_BLOCK)
			A(ERROR_ACCESS_DISABLED_BY_POLICY)
			A(ERROR_REG_NAT_CONSUMPTION)
			A(ERROR_CSCSHARE_OFFLINE)
			A(ERROR_PKINIT_FAILURE)
			A(ERROR_SMARTCARD_SUBSYSTEM_FAILURE)
			A(ERROR_DOWNGRADE_DETECTED)
			A(ERROR_MACHINE_LOCKED)
			A(ERROR_CALLBACK_SUPPLIED_INVALID_DATA)
			A(ERROR_SYNC_FOREGROUND_REFRESH_REQUIRED)
			A(ERROR_DRIVER_BLOCKED)
			A(ERROR_INVALID_IMPORT_OF_NON_DLL)
			A(ERROR_ACCESS_DISABLED_WEBBLADE)
			A(ERROR_ACCESS_DISABLED_WEBBLADE_TAMPER)
			A(ERROR_RECOVERY_FAILURE)
			A(ERROR_ALREADY_FIBER)
			A(ERROR_ALREADY_THREAD)
			A(ERROR_STACK_BUFFER_OVERRUN)
			A(ERROR_PARAMETER_QUOTA_EXCEEDED)
			A(ERROR_DEBUGGER_INACTIVE)
			A(ERROR_DELAY_LOAD_FAILED)
			A(ERROR_VDM_DISALLOWED)
			A(ERROR_UNIDENTIFIED_ERROR)
			A(ERROR_INVALID_CRUNTIME_PARAMETER)
			A(ERROR_BEYOND_VDL)
			A(ERROR_INCOMPATIBLE_SERVICE_SID_TYPE)
			A(ERROR_DRIVER_PROCESS_TERMINATED)
			A(ERROR_IMPLEMENTATION_LIMIT)
			A(ERROR_PROCESS_IS_PROTECTED)
			A(ERROR_SERVICE_NOTIFY_CLIENT_LAGGING)
			A(ERROR_DISK_QUOTA_EXCEEDED)
			A(ERROR_CONTENT_BLOCKED)
			A(ERROR_INCOMPATIBLE_SERVICE_PRIVILEGE)
			A(ERROR_APP_HANG)
			A(ERROR_INVALID_LABEL)
			A(ERROR_NOT_ALL_ASSIGNED)
			A(ERROR_SOME_NOT_MAPPED)
			A(ERROR_NO_QUOTAS_FOR_ACCOUNT)
			A(ERROR_LOCAL_USER_SESSION_KEY)
			A(ERROR_NULL_LM_PASSWORD)
			A(ERROR_UNKNOWN_REVISION)
			A(ERROR_REVISION_MISMATCH)
			A(ERROR_INVALID_OWNER)
			A(ERROR_INVALID_PRIMARY_GROUP)
			A(ERROR_NO_IMPERSONATION_TOKEN)
			A(ERROR_CANT_DISABLE_MANDATORY)
			A(ERROR_NO_LOGON_SERVERS)
			A(ERROR_NO_SUCH_LOGON_SESSION)
			A(ERROR_NO_SUCH_PRIVILEGE)
			A(ERROR_PRIVILEGE_NOT_HELD)
			A(ERROR_INVALID_ACCOUNT_NAME)
			A(ERROR_USER_EXISTS)
			A(ERROR_NO_SUCH_USER)
			A(ERROR_GROUP_EXISTS)
			A(ERROR_NO_SUCH_GROUP)
			A(ERROR_MEMBER_IN_GROUP)
			A(ERROR_MEMBER_NOT_IN_GROUP)
			A(ERROR_LAST_ADMIN)
			A(ERROR_WRONG_PASSWORD)
			A(ERROR_ILL_FORMED_PASSWORD)
			A(ERROR_PASSWORD_RESTRICTION)
			A(ERROR_LOGON_FAILURE)
			A(ERROR_ACCOUNT_RESTRICTION)
			A(ERROR_INVALID_LOGON_HOURS)
			A(ERROR_INVALID_WORKSTATION)
			A(ERROR_PASSWORD_EXPIRED)
			A(ERROR_ACCOUNT_DISABLED)
			A(ERROR_NONE_MAPPED)
			A(ERROR_TOO_MANY_LUIDS_REQUESTED)
			A(ERROR_LUIDS_EXHAUSTED)
			A(ERROR_INVALID_SUB_AUTHORITY)
			A(ERROR_INVALID_ACL)
			A(ERROR_INVALID_SID)
			A(ERROR_INVALID_SECURITY_DESCR)
			A(ERROR_BAD_INHERITANCE_ACL)
			A(ERROR_SERVER_DISABLED)
			A(ERROR_SERVER_NOT_DISABLED)
			A(ERROR_INVALID_ID_AUTHORITY)
			A(ERROR_ALLOTTED_SPACE_EXCEEDED)
			A(ERROR_INVALID_GROUP_ATTRIBUTES)
			A(ERROR_BAD_IMPERSONATION_LEVEL)
			A(ERROR_CANT_OPEN_ANONYMOUS)
			A(ERROR_BAD_VALIDATION_CLASS)
			A(ERROR_BAD_TOKEN_TYPE)
			A(ERROR_NO_SECURITY_ON_OBJECT)
			A(ERROR_CANT_ACCESS_DOMAIN_INFO)
			A(ERROR_INVALID_SERVER_STATE)
			A(ERROR_INVALID_DOMAIN_STATE)
			A(ERROR_INVALID_DOMAIN_ROLE)
			A(ERROR_NO_SUCH_DOMAIN)
			A(ERROR_DOMAIN_EXISTS)
			A(ERROR_DOMAIN_LIMIT_EXCEEDED)
			A(ERROR_INTERNAL_DB_CORRUPTION)
			A(ERROR_INTERNAL_ERROR)
			A(ERROR_GENERIC_NOT_MAPPED)
			A(ERROR_BAD_DESCRIPTOR_FORMAT)
			A(ERROR_NOT_LOGON_PROCESS)
			A(ERROR_LOGON_SESSION_EXISTS)
			A(ERROR_NO_SUCH_PACKAGE)
			A(ERROR_BAD_LOGON_SESSION_STATE)
			A(ERROR_LOGON_SESSION_COLLISION)
			A(ERROR_INVALID_LOGON_TYPE)
			A(ERROR_CANNOT_IMPERSONATE)
			A(ERROR_RXACT_INVALID_STATE)
			A(ERROR_RXACT_COMMIT_FAILURE)
			A(ERROR_SPECIAL_ACCOUNT)
			A(ERROR_SPECIAL_GROUP)
			A(ERROR_SPECIAL_USER)
			A(ERROR_MEMBERS_PRIMARY_GROUP)
			A(ERROR_TOKEN_ALREADY_IN_USE)
			A(ERROR_NO_SUCH_ALIAS)
			A(ERROR_MEMBER_NOT_IN_ALIAS)
			A(ERROR_MEMBER_IN_ALIAS)
			A(ERROR_ALIAS_EXISTS)
			A(ERROR_LOGON_NOT_GRANTED)
			A(ERROR_TOO_MANY_SECRETS)
			A(ERROR_SECRET_TOO_LONG)
			A(ERROR_INTERNAL_DB_ERROR)
			A(ERROR_TOO_MANY_CONTEXT_IDS)
			A(ERROR_LOGON_TYPE_NOT_GRANTED)
			A(ERROR_NT_CROSS_ENCRYPTION_REQUIRED)
			A(ERROR_NO_SUCH_MEMBER)
			A(ERROR_INVALID_MEMBER)
			A(ERROR_TOO_MANY_SIDS)
			A(ERROR_LM_CROSS_ENCRYPTION_REQUIRED)
			A(ERROR_NO_INHERITANCE)
			A(ERROR_FILE_CORRUPT)
			A(ERROR_DISK_CORRUPT)
			A(ERROR_NO_USER_SESSION_KEY)
			A(ERROR_LICENSE_QUOTA_EXCEEDED)
			A(ERROR_WRONG_TARGET_NAME)
			A(ERROR_MUTUAL_AUTH_FAILED)
			A(ERROR_TIME_SKEW)
			A(ERROR_CURRENT_DOMAIN_NOT_ALLOWED)
			A(ERROR_INVALID_WINDOW_HANDLE)
			A(ERROR_INVALID_MENU_HANDLE)
			A(ERROR_INVALID_CURSOR_HANDLE)
			A(ERROR_INVALID_ACCEL_HANDLE)
			A(ERROR_INVALID_HOOK_HANDLE)
			A(ERROR_INVALID_DWP_HANDLE)
			A(ERROR_TLW_WITH_WSCHILD)
			A(ERROR_CANNOT_FIND_WND_CLASS)
			A(ERROR_WINDOW_OF_OTHER_THREAD)
			A(ERROR_HOTKEY_ALREADY_REGISTERED)
			A(ERROR_CLASS_ALREADY_EXISTS)
			A(ERROR_CLASS_DOES_NOT_EXIST)
			A(ERROR_CLASS_HAS_WINDOWS)
			A(ERROR_INVALID_INDEX)
			A(ERROR_INVALID_ICON_HANDLE)
			A(ERROR_PRIVATE_DIALOG_INDEX)
			A(ERROR_LISTBOX_ID_NOT_FOUND)
			A(ERROR_NO_WILDCARD_CHARACTERS)
			A(ERROR_CLIPBOARD_NOT_OPEN)
			A(ERROR_HOTKEY_NOT_REGISTERED)
			A(ERROR_WINDOW_NOT_DIALOG)
			A(ERROR_CONTROL_ID_NOT_FOUND)
			A(ERROR_INVALID_COMBOBOX_MESSAGE)
			A(ERROR_WINDOW_NOT_COMBOBOX)
			A(ERROR_INVALID_EDIT_HEIGHT)
			A(ERROR_DC_NOT_FOUND)
			A(ERROR_INVALID_HOOK_FILTER)
			A(ERROR_INVALID_FILTER_PROC)
			A(ERROR_HOOK_NEEDS_HMOD)
			A(ERROR_GLOBAL_ONLY_HOOK)
			A(ERROR_JOURNAL_HOOK_SET)
			A(ERROR_HOOK_NOT_INSTALLED)
			A(ERROR_INVALID_LB_MESSAGE)
			A(ERROR_SETCOUNT_ON_BAD_LB)
			A(ERROR_LB_WITHOUT_TABSTOPS)
			A(ERROR_DESTROY_OBJECT_OF_OTHER_THREAD)
			A(ERROR_CHILD_WINDOW_MENU)
			A(ERROR_NO_SYSTEM_MENU)
			A(ERROR_INVALID_MSGBOX_STYLE)
			A(ERROR_INVALID_SPI_VALUE)
			A(ERROR_SCREEN_ALREADY_LOCKED)
			A(ERROR_HWNDS_HAVE_DIFF_PARENT)
			A(ERROR_NOT_CHILD_WINDOW)
			A(ERROR_INVALID_GW_COMMAND)
			A(ERROR_INVALID_THREAD_ID)
			A(ERROR_NON_MDICHILD_WINDOW)
			A(ERROR_POPUP_ALREADY_ACTIVE)
			A(ERROR_NO_SCROLLBARS)
			A(ERROR_INVALID_SCROLLBAR_RANGE)
			A(ERROR_INVALID_SHOWWIN_COMMAND)
			A(ERROR_NO_SYSTEM_RESOURCES)
			A(ERROR_NONPAGED_SYSTEM_RESOURCES)
			A(ERROR_PAGED_SYSTEM_RESOURCES)
			A(ERROR_WORKING_SET_QUOTA)
			A(ERROR_PAGEFILE_QUOTA)
			A(ERROR_COMMITMENT_LIMIT)
			A(ERROR_MENU_ITEM_NOT_FOUND)
			A(ERROR_INVALID_KEYBOARD_HANDLE)
			A(ERROR_HOOK_TYPE_NOT_ALLOWED)
			A(ERROR_REQUIRES_INTERACTIVE_WINDOWSTATION)
			A(ERROR_TIMEOUT)
			A(ERROR_INVALID_MONITOR_HANDLE)
			A(ERROR_INCORRECT_SIZE)
			A(ERROR_SYMLINK_CLASS_DISABLED)
			A(ERROR_SYMLINK_NOT_SUPPORTED)
			A(ERROR_XML_PARSE_ERROR)
			A(ERROR_XMLDSIG_ERROR)
			A(ERROR_RESTART_APPLICATION)
			A(ERROR_WRONG_COMPARTMENT)
			A(ERROR_AUTHIP_FAILURE)
			A(ERROR_NO_NVRAM_RESOURCES)
			A(ERROR_EVENTLOG_FILE_CORRUPT)
			A(ERROR_EVENTLOG_CANT_START)
			A(ERROR_LOG_FILE_FULL)
			A(ERROR_EVENTLOG_FILE_CHANGED)
			A(ERROR_INVALID_TASK_NAME)
			A(ERROR_INVALID_TASK_INDEX)
			A(ERROR_THREAD_ALREADY_IN_TASK)
			A(ERROR_INSTALL_SERVICE_FAILURE)
			A(ERROR_INSTALL_USEREXIT)
			A(ERROR_INSTALL_FAILURE)
			A(ERROR_INSTALL_SUSPEND)
			A(ERROR_UNKNOWN_PRODUCT)
			A(ERROR_UNKNOWN_FEATURE)
			A(ERROR_UNKNOWN_COMPONENT)
			A(ERROR_UNKNOWN_PROPERTY)
			A(ERROR_INVALID_HANDLE_STATE)
			A(ERROR_BAD_CONFIGURATION)
			A(ERROR_INDEX_ABSENT)
			A(ERROR_INSTALL_SOURCE_ABSENT)
			A(ERROR_INSTALL_PACKAGE_VERSION)
			A(ERROR_PRODUCT_UNINSTALLED)
			A(ERROR_BAD_QUERY_SYNTAX)
			A(ERROR_INVALID_FIELD)
			A(ERROR_DEVICE_REMOVED)
			A(ERROR_INSTALL_ALREADY_RUNNING)
			A(ERROR_INSTALL_PACKAGE_OPEN_FAILED)
			A(ERROR_INSTALL_PACKAGE_INVALID)
			A(ERROR_INSTALL_UI_FAILURE)
			A(ERROR_INSTALL_LOG_FAILURE)
			A(ERROR_INSTALL_LANGUAGE_UNSUPPORTED)
			A(ERROR_INSTALL_TRANSFORM_FAILURE)
			A(ERROR_INSTALL_PACKAGE_REJECTED)
			A(ERROR_FUNCTION_NOT_CALLED)
			A(ERROR_FUNCTION_FAILED)
			A(ERROR_INVALID_TABLE)
			A(ERROR_DATATYPE_MISMATCH)
			A(ERROR_UNSUPPORTED_TYPE)
			A(ERROR_CREATE_FAILED)
			A(ERROR_INSTALL_TEMP_UNWRITABLE)
			A(ERROR_INSTALL_PLATFORM_UNSUPPORTED)
			A(ERROR_INSTALL_NOTUSED)
			A(ERROR_PATCH_PACKAGE_OPEN_FAILED)
			A(ERROR_PATCH_PACKAGE_INVALID)
			A(ERROR_PATCH_PACKAGE_UNSUPPORTED)
			A(ERROR_PRODUCT_VERSION)
			A(ERROR_INVALID_COMMAND_LINE)
			A(ERROR_INSTALL_REMOTE_DISALLOWED)
			A(ERROR_SUCCESS_REBOOT_INITIATED)
			A(ERROR_PATCH_TARGET_NOT_FOUND)
			A(ERROR_PATCH_PACKAGE_REJECTED)
			A(ERROR_INSTALL_TRANSFORM_REJECTED)
			A(ERROR_INSTALL_REMOTE_PROHIBITED)
			A(ERROR_PATCH_REMOVAL_UNSUPPORTED)
			A(ERROR_UNKNOWN_PATCH)
			A(ERROR_PATCH_NO_SEQUENCE)
			A(ERROR_PATCH_REMOVAL_DISALLOWED)
			A(ERROR_INVALID_PATCH_XML)
			A(ERROR_PATCH_MANAGED_ADVERTISED_PRODUCT)
			A(ERROR_INSTALL_SERVICE_SAFEBOOT)
			A(ERROR_FAIL_FAST_EXCEPTION)
			A(ERROR_INVALID_USER_BUFFER)
			A(ERROR_UNRECOGNIZED_MEDIA)
			A(ERROR_NO_TRUST_LSA_SECRET)
			A(ERROR_NO_TRUST_SAM_ACCOUNT)
			A(ERROR_TRUSTED_DOMAIN_FAILURE)
			A(ERROR_TRUSTED_RELATIONSHIP_FAILURE)
			A(ERROR_TRUST_FAILURE)
			A(ERROR_NETLOGON_NOT_STARTED)
			A(ERROR_ACCOUNT_EXPIRED)
			A(ERROR_REDIRECTOR_HAS_OPEN_HANDLES)
			A(ERROR_PRINTER_DRIVER_ALREADY_INSTALLED)
			A(ERROR_UNKNOWN_PORT)
			A(ERROR_UNKNOWN_PRINTER_DRIVER)
			A(ERROR_UNKNOWN_PRINTPROCESSOR)
			A(ERROR_INVALID_SEPARATOR_FILE)
			A(ERROR_INVALID_PRIORITY)
			A(ERROR_INVALID_PRINTER_NAME)
			A(ERROR_PRINTER_ALREADY_EXISTS)
			A(ERROR_INVALID_PRINTER_COMMAND)
			A(ERROR_INVALID_DATATYPE)
			A(ERROR_INVALID_ENVIRONMENT)
			A(ERROR_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT)
			A(ERROR_NOLOGON_WORKSTATION_TRUST_ACCOUNT)
			A(ERROR_NOLOGON_SERVER_TRUST_ACCOUNT)
			A(ERROR_DOMAIN_TRUST_INCONSISTENT)
			A(ERROR_SERVER_HAS_OPEN_HANDLES)
			A(ERROR_RESOURCE_DATA_NOT_FOUND)
			A(ERROR_RESOURCE_TYPE_NOT_FOUND)
			A(ERROR_RESOURCE_NAME_NOT_FOUND)
			A(ERROR_RESOURCE_LANG_NOT_FOUND)
			A(ERROR_NOT_ENOUGH_QUOTA)
			A(ERROR_INVALID_TIME)
			A(ERROR_INVALID_FORM_NAME)
			A(ERROR_INVALID_FORM_SIZE)
			A(ERROR_ALREADY_WAITING)
			A(ERROR_PRINTER_DELETED)
			A(ERROR_INVALID_PRINTER_STATE)
			A(ERROR_PASSWORD_MUST_CHANGE)
			A(ERROR_DOMAIN_CONTROLLER_NOT_FOUND)
			A(ERROR_ACCOUNT_LOCKED_OUT)
			A(ERROR_NO_SITENAME)
			A(ERROR_CANT_ACCESS_FILE)
			A(ERROR_CANT_RESOLVE_FILENAME)
			A(ERROR_KM_DRIVER_BLOCKED)
			A(ERROR_CONTEXT_EXPIRED)
			A(ERROR_PER_USER_TRUST_QUOTA_EXCEEDED)
			A(ERROR_ALL_USER_TRUST_QUOTA_EXCEEDED)
			A(ERROR_USER_DELETE_TRUST_QUOTA_EXCEEDED)
			A(ERROR_AUTHENTICATION_FIREWALL_FAILED)
			A(ERROR_REMOTE_PRINT_CONNECTIONS_BLOCKED)
			A(ERROR_NTLM_BLOCKED)
			A(ERROR_INVALID_PIXEL_FORMAT)
			A(ERROR_BAD_DRIVER)
			A(ERROR_INVALID_WINDOW_STYLE)
			A(ERROR_METAFILE_NOT_SUPPORTED)
			A(ERROR_TRANSFORM_NOT_SUPPORTED)
			A(ERROR_CLIPPING_NOT_SUPPORTED)
			A(ERROR_INVALID_CMM)
			A(ERROR_INVALID_PROFILE)
			A(ERROR_TAG_NOT_FOUND)
			A(ERROR_TAG_NOT_PRESENT)
			A(ERROR_DUPLICATE_TAG)
			A(ERROR_PROFILE_NOT_ASSOCIATED_WITH_DEVICE)
			A(ERROR_PROFILE_NOT_FOUND)
			A(ERROR_INVALID_COLORSPACE)
			A(ERROR_ICM_NOT_ENABLED)
			A(ERROR_DELETING_ICM_XFORM)
			A(ERROR_INVALID_TRANSFORM)
			A(ERROR_COLORSPACE_MISMATCH)
			A(ERROR_INVALID_COLORINDEX)
			A(ERROR_PROFILE_DOES_NOT_MATCH_DEVICE)
			A(ERROR_CONNECTED_OTHER_PASSWORD)
			A(ERROR_CONNECTED_OTHER_PASSWORD_DEFAULT)
			A(ERROR_BAD_USERNAME)
			A(ERROR_NOT_CONNECTED)
			A(ERROR_OPEN_FILES)
			A(ERROR_ACTIVE_CONNECTIONS)
			A(ERROR_DEVICE_IN_USE)
			A(ERROR_UNKNOWN_PRINT_MONITOR)
			A(ERROR_PRINTER_DRIVER_IN_USE)
			A(ERROR_SPOOL_FILE_NOT_FOUND)
			A(ERROR_SPL_NO_STARTDOC)
			A(ERROR_SPL_NO_ADDJOB)
			A(ERROR_PRINT_PROCESSOR_ALREADY_INSTALLED)
			A(ERROR_PRINT_MONITOR_ALREADY_INSTALLED)
			A(ERROR_INVALID_PRINT_MONITOR)
			A(ERROR_PRINT_MONITOR_IN_USE)
			A(ERROR_PRINTER_HAS_JOBS_QUEUED)
			A(ERROR_SUCCESS_REBOOT_REQUIRED)
			A(ERROR_SUCCESS_RESTART_REQUIRED)
			A(ERROR_PRINTER_NOT_FOUND)
			A(ERROR_PRINTER_DRIVER_WARNED)
			A(ERROR_PRINTER_DRIVER_BLOCKED)
			A(ERROR_PRINTER_DRIVER_PACKAGE_IN_USE)
			A(ERROR_CORE_DRIVER_PACKAGE_NOT_FOUND)
			A(ERROR_FAIL_REBOOT_REQUIRED)
			A(ERROR_FAIL_REBOOT_INITIATED)
			A(ERROR_PRINTER_DRIVER_DOWNLOAD_NEEDED)
			A(ERROR_PRINT_JOB_RESTART_REQUIRED)
			A(ERROR_IO_REISSUE_AS_CACHED)
			A(ERROR_WINS_INTERNAL)
			A(ERROR_CAN_NOT_DEL_LOCAL_WINS)
			A(ERROR_STATIC_INIT)
			A(ERROR_INC_BACKUP)
			A(ERROR_FULL_BACKUP)
			A(ERROR_REC_NON_EXISTENT)
			A(ERROR_RPL_NOT_ALLOWED)
			A(ERROR_DHCP_ADDRESS_CONFLICT)
			A(ERROR_WMI_GUID_NOT_FOUND)
			A(ERROR_WMI_INSTANCE_NOT_FOUND)
			A(ERROR_WMI_ITEMID_NOT_FOUND)
			A(ERROR_WMI_TRY_AGAIN)
			A(ERROR_WMI_DP_NOT_FOUND)
			A(ERROR_WMI_UNRESOLVED_INSTANCE_REF)
			A(ERROR_WMI_ALREADY_ENABLED)
			A(ERROR_WMI_GUID_DISCONNECTED)
			A(ERROR_WMI_SERVER_UNAVAILABLE)
			A(ERROR_WMI_DP_FAILED)
			A(ERROR_WMI_INVALID_MOF)
			A(ERROR_WMI_INVALID_REGINFO)
			A(ERROR_WMI_ALREADY_DISABLED)
			A(ERROR_WMI_READ_ONLY)
			A(ERROR_WMI_SET_FAILURE)
			A(ERROR_INVALID_MEDIA)
			A(ERROR_INVALID_LIBRARY)
			A(ERROR_INVALID_MEDIA_POOL)
			A(ERROR_DRIVE_MEDIA_MISMATCH)
			A(ERROR_MEDIA_OFFLINE)
			A(ERROR_LIBRARY_OFFLINE)
			A(ERROR_EMPTY)
			A(ERROR_NOT_EMPTY)
			A(ERROR_MEDIA_UNAVAILABLE)
			A(ERROR_RESOURCE_DISABLED)
			A(ERROR_INVALID_CLEANER)
			A(ERROR_UNABLE_TO_CLEAN)
			A(ERROR_OBJECT_NOT_FOUND)
			A(ERROR_DATABASE_FAILURE)
			A(ERROR_DATABASE_FULL)
			A(ERROR_MEDIA_INCOMPATIBLE)
			A(ERROR_RESOURCE_NOT_PRESENT)
			A(ERROR_INVALID_OPERATION)
			A(ERROR_MEDIA_NOT_AVAILABLE)
			A(ERROR_DEVICE_NOT_AVAILABLE)
			A(ERROR_REQUEST_REFUSED)
			A(ERROR_INVALID_DRIVE_OBJECT)
			A(ERROR_LIBRARY_FULL)
			A(ERROR_MEDIUM_NOT_ACCESSIBLE)
			A(ERROR_UNABLE_TO_LOAD_MEDIUM)
			A(ERROR_UNABLE_TO_INVENTORY_DRIVE)
			A(ERROR_UNABLE_TO_INVENTORY_SLOT)
			A(ERROR_UNABLE_TO_INVENTORY_TRANSPORT)
			A(ERROR_TRANSPORT_FULL)
			A(ERROR_CONTROLLING_IEPORT)
			A(ERROR_UNABLE_TO_EJECT_MOUNTED_MEDIA)
			A(ERROR_CLEANER_SLOT_SET)
			A(ERROR_CLEANER_SLOT_NOT_SET)
			A(ERROR_CLEANER_CARTRIDGE_SPENT)
			A(ERROR_UNEXPECTED_OMID)
			A(ERROR_CANT_DELETE_LAST_ITEM)
			A(ERROR_MESSAGE_EXCEEDS_MAX_SIZE)
			A(ERROR_VOLUME_CONTAINS_SYS_FILES)
			A(ERROR_INDIGENOUS_TYPE)
			A(ERROR_NO_SUPPORTING_DRIVES)
			A(ERROR_CLEANER_CARTRIDGE_INSTALLED)
			A(ERROR_IEPORT_FULL)
			A(ERROR_FILE_OFFLINE)
			A(ERROR_REMOTE_STORAGE_NOT_ACTIVE)
			A(ERROR_REMOTE_STORAGE_MEDIA_ERROR)
			A(ERROR_NOT_A_REPARSE_POINT)
			A(ERROR_REPARSE_ATTRIBUTE_CONFLICT)
			A(ERROR_INVALID_REPARSE_DATA)
			A(ERROR_REPARSE_TAG_INVALID)
			A(ERROR_REPARSE_TAG_MISMATCH)
			A(ERROR_VOLUME_NOT_SIS_ENABLED)
			A(ERROR_DEPENDENT_RESOURCE_EXISTS)
			A(ERROR_DEPENDENCY_NOT_FOUND)
			A(ERROR_DEPENDENCY_ALREADY_EXISTS)
			A(ERROR_RESOURCE_NOT_ONLINE)
			A(ERROR_HOST_NODE_NOT_AVAILABLE)
			A(ERROR_RESOURCE_NOT_AVAILABLE)
			A(ERROR_RESOURCE_NOT_FOUND)
			A(ERROR_SHUTDOWN_CLUSTER)
			A(ERROR_CANT_EVICT_ACTIVE_NODE)
			A(ERROR_OBJECT_ALREADY_EXISTS)
			A(ERROR_OBJECT_IN_LIST)
			A(ERROR_GROUP_NOT_AVAILABLE)
			A(ERROR_GROUP_NOT_FOUND)
			A(ERROR_GROUP_NOT_ONLINE)
			A(ERROR_HOST_NODE_NOT_RESOURCE_OWNER)
			A(ERROR_HOST_NODE_NOT_GROUP_OWNER)
			A(ERROR_RESMON_CREATE_FAILED)
			A(ERROR_RESMON_ONLINE_FAILED)
			A(ERROR_RESOURCE_ONLINE)
			A(ERROR_QUORUM_RESOURCE)
			A(ERROR_NOT_QUORUM_CAPABLE)
			A(ERROR_CLUSTER_SHUTTING_DOWN)
			A(ERROR_INVALID_STATE)
			A(ERROR_RESOURCE_PROPERTIES_STORED)
			A(ERROR_NOT_QUORUM_CLASS)
			A(ERROR_CORE_RESOURCE)
			A(ERROR_QUORUM_RESOURCE_ONLINE_FAILED)
			A(ERROR_QUORUMLOG_OPEN_FAILED)
			A(ERROR_CLUSTERLOG_CORRUPT)
			A(ERROR_CLUSTERLOG_RECORD_EXCEEDS_MAXSIZE)
			A(ERROR_CLUSTERLOG_EXCEEDS_MAXSIZE)
			A(ERROR_CLUSTERLOG_CHKPOINT_NOT_FOUND)
			A(ERROR_CLUSTERLOG_NOT_ENOUGH_SPACE)
			A(ERROR_QUORUM_OWNER_ALIVE)
			A(ERROR_NETWORK_NOT_AVAILABLE)
			A(ERROR_NODE_NOT_AVAILABLE)
			A(ERROR_ALL_NODES_NOT_AVAILABLE)
			A(ERROR_RESOURCE_FAILED)
			A(ERROR_CLUSTER_INVALID_NODE)
			A(ERROR_CLUSTER_NODE_EXISTS)
			A(ERROR_CLUSTER_JOIN_IN_PROGRESS)
			A(ERROR_CLUSTER_NODE_NOT_FOUND)
			A(ERROR_CLUSTER_LOCAL_NODE_NOT_FOUND)
			A(ERROR_CLUSTER_NETWORK_EXISTS)
			A(ERROR_CLUSTER_NETWORK_NOT_FOUND)
			A(ERROR_CLUSTER_NETINTERFACE_EXISTS)
			A(ERROR_CLUSTER_NETINTERFACE_NOT_FOUND)
			A(ERROR_CLUSTER_INVALID_REQUEST)
			A(ERROR_CLUSTER_INVALID_NETWORK_PROVIDER)
			A(ERROR_CLUSTER_NODE_DOWN)
			A(ERROR_CLUSTER_NODE_UNREACHABLE)
			A(ERROR_CLUSTER_NODE_NOT_MEMBER)
			A(ERROR_CLUSTER_JOIN_NOT_IN_PROGRESS)
			A(ERROR_CLUSTER_INVALID_NETWORK)
			A(ERROR_CLUSTER_NODE_UP)
			A(ERROR_CLUSTER_IPADDR_IN_USE)
			A(ERROR_CLUSTER_NODE_NOT_PAUSED)
			A(ERROR_CLUSTER_NO_SECURITY_CONTEXT)
			A(ERROR_CLUSTER_NETWORK_NOT_INTERNAL)
			A(ERROR_CLUSTER_NODE_ALREADY_UP)
			A(ERROR_CLUSTER_NODE_ALREADY_DOWN)
			A(ERROR_CLUSTER_NETWORK_ALREADY_ONLINE)
			A(ERROR_CLUSTER_NETWORK_ALREADY_OFFLINE)
			A(ERROR_CLUSTER_NODE_ALREADY_MEMBER)
			A(ERROR_CLUSTER_LAST_INTERNAL_NETWORK)
			A(ERROR_CLUSTER_NETWORK_HAS_DEPENDENTS)
			A(ERROR_INVALID_OPERATION_ON_QUORUM)
			A(ERROR_DEPENDENCY_NOT_ALLOWED)
			A(ERROR_CLUSTER_NODE_PAUSED)
			A(ERROR_NODE_CANT_HOST_RESOURCE)
			A(ERROR_CLUSTER_NODE_NOT_READY)
			A(ERROR_CLUSTER_NODE_SHUTTING_DOWN)
			A(ERROR_CLUSTER_JOIN_ABORTED)
			A(ERROR_CLUSTER_INCOMPATIBLE_VERSIONS)
			A(ERROR_CLUSTER_MAXNUM_OF_RESOURCES_EXCEEDED)
			A(ERROR_CLUSTER_SYSTEM_CONFIG_CHANGED)
			A(ERROR_CLUSTER_RESOURCE_TYPE_NOT_FOUND)
			A(ERROR_CLUSTER_RESTYPE_NOT_SUPPORTED)
			A(ERROR_CLUSTER_RESNAME_NOT_FOUND)
			A(ERROR_CLUSTER_NO_RPC_PACKAGES_REGISTERED)
			A(ERROR_CLUSTER_OWNER_NOT_IN_PREFLIST)
			A(ERROR_CLUSTER_DATABASE_SEQMISMATCH)
			A(ERROR_RESMON_INVALID_STATE)
			A(ERROR_CLUSTER_GUM_NOT_LOCKER)
			A(ERROR_QUORUM_DISK_NOT_FOUND)
			A(ERROR_DATABASE_BACKUP_CORRUPT)
			A(ERROR_CLUSTER_NODE_ALREADY_HAS_DFS_ROOT)
			A(ERROR_RESOURCE_PROPERTY_UNCHANGEABLE)
			A(ERROR_CLUSTER_MEMBERSHIP_INVALID_STATE)
			A(ERROR_CLUSTER_QUORUMLOG_NOT_FOUND)
			A(ERROR_CLUSTER_MEMBERSHIP_HALT)
			A(ERROR_CLUSTER_INSTANCE_ID_MISMATCH)
			A(ERROR_CLUSTER_NETWORK_NOT_FOUND_FOR_IP)
			A(ERROR_CLUSTER_PROPERTY_DATA_TYPE_MISMATCH)
			A(ERROR_CLUSTER_EVICT_WITHOUT_CLEANUP)
			A(ERROR_CLUSTER_PARAMETER_MISMATCH)
			A(ERROR_NODE_CANNOT_BE_CLUSTERED)
			A(ERROR_CLUSTER_WRONG_OS_VERSION)
			A(ERROR_CLUSTER_CANT_CREATE_DUP_CLUSTER_NAME)
			A(ERROR_CLUSCFG_ALREADY_COMMITTED)
			A(ERROR_CLUSCFG_ROLLBACK_FAILED)
			A(ERROR_CLUSCFG_SYSTEM_DISK_DRIVE_LETTER_CONFLICT)
			A(ERROR_CLUSTER_OLD_VERSION)
			A(ERROR_CLUSTER_MISMATCHED_COMPUTER_ACCT_NAME)
			A(ERROR_CLUSTER_NO_NET_ADAPTERS)
			A(ERROR_CLUSTER_POISONED)
			A(ERROR_CLUSTER_GROUP_MOVING)
			A(ERROR_CLUSTER_RESOURCE_TYPE_BUSY)
			A(ERROR_RESOURCE_CALL_TIMED_OUT)
			A(ERROR_INVALID_CLUSTER_IPV6_ADDRESS)
			A(ERROR_CLUSTER_INTERNAL_INVALID_FUNCTION)
			A(ERROR_CLUSTER_PARAMETER_OUT_OF_BOUNDS)
			A(ERROR_CLUSTER_PARTIAL_SEND)
			A(ERROR_CLUSTER_REGISTRY_INVALID_FUNCTION)
			A(ERROR_CLUSTER_INVALID_STRING_TERMINATION)
			A(ERROR_CLUSTER_INVALID_STRING_FORMAT)
			A(ERROR_CLUSTER_DATABASE_TRANSACTION_IN_PROGRESS)
			A(ERROR_CLUSTER_DATABASE_TRANSACTION_NOT_IN_PROGRESS)
			A(ERROR_CLUSTER_NULL_DATA)
			A(ERROR_CLUSTER_PARTIAL_READ)
			A(ERROR_CLUSTER_PARTIAL_WRITE)
			A(ERROR_CLUSTER_CANT_DESERIALIZE_DATA)
			A(ERROR_DEPENDENT_RESOURCE_PROPERTY_CONFLICT)
			A(ERROR_CLUSTER_NO_QUORUM)
			A(ERROR_CLUSTER_INVALID_IPV6_NETWORK)
			A(ERROR_CLUSTER_INVALID_IPV6_TUNNEL_NETWORK)
			A(ERROR_QUORUM_NOT_ALLOWED_IN_THIS_GROUP)
			A(ERROR_DEPENDENCY_TREE_TOO_COMPLEX)
			A(ERROR_EXCEPTION_IN_RESOURCE_CALL)
			A(ERROR_CLUSTER_RHS_FAILED_INITIALIZATION)
			A(ERROR_CLUSTER_NOT_INSTALLED)
			A(ERROR_CLUSTER_RESOURCES_MUST_BE_ONLINE_ON_THE_SAME_NODE)
			A(ERROR_CLUSTER_MAX_NODES_IN_CLUSTER)
			A(ERROR_CLUSTER_TOO_MANY_NODES)
			A(ERROR_CLUSTER_OBJECT_ALREADY_USED)
			A(ERROR_NONCORE_GROUPS_FOUND)
			A(ERROR_FILE_SHARE_RESOURCE_CONFLICT)
			A(ERROR_CLUSTER_EVICT_INVALID_REQUEST)
			A(ERROR_CLUSTER_SINGLETON_RESOURCE)
			A(ERROR_CLUSTER_GROUP_SINGLETON_RESOURCE)
			A(ERROR_CLUSTER_RESOURCE_PROVIDER_FAILED)
			A(ERROR_CLUSTER_RESOURCE_CONFIGURATION_ERROR)
			A(ERROR_CLUSTER_GROUP_BUSY)
			A(ERROR_CLUSTER_NOT_SHARED_VOLUME)
			A(ERROR_CLUSTER_INVALID_SECURITY_DESCRIPTOR)
			A(ERROR_CLUSTER_SHARED_VOLUMES_IN_USE)
			A(ERROR_CLUSTER_USE_SHARED_VOLUMES_API)
			A(ERROR_CLUSTER_BACKUP_IN_PROGRESS)
			A(ERROR_NON_CSV_PATH)
			A(ERROR_CSV_VOLUME_NOT_LOCAL)
			A(ERROR_CLUSTER_WATCHDOG_TERMINATING)
			A(ERROR_ENCRYPTION_FAILED)
			A(ERROR_DECRYPTION_FAILED)
			A(ERROR_FILE_ENCRYPTED)
			A(ERROR_NO_RECOVERY_POLICY)
			A(ERROR_NO_EFS)
			A(ERROR_WRONG_EFS)
			A(ERROR_NO_USER_KEYS)
			A(ERROR_FILE_NOT_ENCRYPTED)
			A(ERROR_NOT_EXPORT_FORMAT)
			A(ERROR_FILE_READ_ONLY)
			A(ERROR_DIR_EFS_DISALLOWED)
			A(ERROR_EFS_SERVER_NOT_TRUSTED)
			A(ERROR_BAD_RECOVERY_POLICY)
			A(ERROR_EFS_ALG_BLOB_TOO_BIG)
			A(ERROR_VOLUME_NOT_SUPPORT_EFS)
			A(ERROR_EFS_DISABLED)
			A(ERROR_EFS_VERSION_NOT_SUPPORT)
			A(ERROR_CS_ENCRYPTION_INVALID_SERVER_RESPONSE)
			A(ERROR_CS_ENCRYPTION_UNSUPPORTED_SERVER)
			A(ERROR_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE)
			A(ERROR_CS_ENCRYPTION_NEW_ENCRYPTED_FILE)
			A(ERROR_CS_ENCRYPTION_FILE_NOT_CSE)
			A(ERROR_ENCRYPTION_POLICY_DENIES_OPERATION)
			A(ERROR_NO_BROWSER_SERVERS_FOUND)
			A(ERROR_LOG_SECTOR_INVALID)
			A(ERROR_LOG_SECTOR_PARITY_INVALID)
			A(ERROR_LOG_SECTOR_REMAPPED)
			A(ERROR_LOG_BLOCK_INCOMPLETE)
			A(ERROR_LOG_INVALID_RANGE)
			A(ERROR_LOG_BLOCKS_EXHAUSTED)
			A(ERROR_LOG_READ_CONTEXT_INVALID)
			A(ERROR_LOG_RESTART_INVALID)
			A(ERROR_LOG_BLOCK_VERSION)
			A(ERROR_LOG_BLOCK_INVALID)
			A(ERROR_LOG_READ_MODE_INVALID)
			A(ERROR_LOG_NO_RESTART)
			A(ERROR_LOG_METADATA_CORRUPT)
			A(ERROR_LOG_METADATA_INVALID)
			A(ERROR_LOG_METADATA_INCONSISTENT)
			A(ERROR_LOG_RESERVATION_INVALID)
			A(ERROR_LOG_CANT_DELETE)
			A(ERROR_LOG_CONTAINER_LIMIT_EXCEEDED)
			A(ERROR_LOG_START_OF_LOG)
			A(ERROR_LOG_POLICY_ALREADY_INSTALLED)
			A(ERROR_LOG_POLICY_NOT_INSTALLED)
			A(ERROR_LOG_POLICY_INVALID)
			A(ERROR_LOG_POLICY_CONFLICT)
			A(ERROR_LOG_PINNED_ARCHIVE_TAIL)
			A(ERROR_LOG_RECORD_NONEXISTENT)
			A(ERROR_LOG_RECORDS_RESERVED_INVALID)
			A(ERROR_LOG_SPACE_RESERVED_INVALID)
			A(ERROR_LOG_TAIL_INVALID)
			A(ERROR_LOG_FULL)
			A(ERROR_COULD_NOT_RESIZE_LOG)
			A(ERROR_LOG_MULTIPLEXED)
			A(ERROR_LOG_DEDICATED)
			A(ERROR_LOG_ARCHIVE_NOT_IN_PROGRESS)
			A(ERROR_LOG_ARCHIVE_IN_PROGRESS)
			A(ERROR_LOG_EPHEMERAL)
			A(ERROR_LOG_NOT_ENOUGH_CONTAINERS)
			A(ERROR_LOG_CLIENT_ALREADY_REGISTERED)
			A(ERROR_LOG_CLIENT_NOT_REGISTERED)
			A(ERROR_LOG_FULL_HANDLER_IN_PROGRESS)
			A(ERROR_LOG_CONTAINER_READ_FAILED)
			A(ERROR_LOG_CONTAINER_WRITE_FAILED)
			A(ERROR_LOG_CONTAINER_OPEN_FAILED)
			A(ERROR_LOG_CONTAINER_STATE_INVALID)
			A(ERROR_LOG_STATE_INVALID)
			A(ERROR_LOG_PINNED)
			A(ERROR_LOG_METADATA_FLUSH_FAILED)
			A(ERROR_LOG_INCONSISTENT_SECURITY)
			A(ERROR_LOG_APPENDED_FLUSH_FAILED)
			A(ERROR_LOG_PINNED_RESERVATION)
			A(ERROR_INVALID_TRANSACTION)
			A(ERROR_TRANSACTION_NOT_ACTIVE)
			A(ERROR_TRANSACTION_REQUEST_NOT_VALID)
			A(ERROR_TRANSACTION_NOT_REQUESTED)
			A(ERROR_TRANSACTION_ALREADY_ABORTED)
			A(ERROR_TRANSACTION_ALREADY_COMMITTED)
			A(ERROR_TM_INITIALIZATION_FAILED)
			A(ERROR_RESOURCEMANAGER_READ_ONLY)
			A(ERROR_TRANSACTION_NOT_JOINED)
			A(ERROR_TRANSACTION_SUPERIOR_EXISTS)
			A(ERROR_CRM_PROTOCOL_ALREADY_EXISTS)
			A(ERROR_TRANSACTION_PROPAGATION_FAILED)
			A(ERROR_CRM_PROTOCOL_NOT_FOUND)
			A(ERROR_TRANSACTION_INVALID_MARSHALL_BUFFER)
			A(ERROR_CURRENT_TRANSACTION_NOT_VALID)
			A(ERROR_TRANSACTION_NOT_FOUND)
			A(ERROR_RESOURCEMANAGER_NOT_FOUND)
			A(ERROR_ENLISTMENT_NOT_FOUND)
			A(ERROR_TRANSACTIONMANAGER_NOT_FOUND)
			A(ERROR_TRANSACTIONMANAGER_NOT_ONLINE)
			A(ERROR_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION)
			A(ERROR_TRANSACTION_NOT_ROOT)
			A(ERROR_TRANSACTION_OBJECT_EXPIRED)
			A(ERROR_TRANSACTION_RESPONSE_NOT_ENLISTED)
			A(ERROR_TRANSACTION_RECORD_TOO_LONG)
			A(ERROR_IMPLICIT_TRANSACTION_NOT_SUPPORTED)
			A(ERROR_TRANSACTION_INTEGRITY_VIOLATED)
			A(ERROR_TRANSACTIONMANAGER_IDENTITY_MISMATCH)
			A(ERROR_RM_CANNOT_BE_FROZEN_FOR_SNAPSHOT)
			A(ERROR_TRANSACTION_MUST_WRITETHROUGH)
			A(ERROR_TRANSACTION_NO_SUPERIOR)
			A(ERROR_HEURISTIC_DAMAGE_POSSIBLE)
			A(ERROR_TRANSACTIONAL_CONFLICT)
			A(ERROR_RM_NOT_ACTIVE)
			A(ERROR_RM_METADATA_CORRUPT)
			A(ERROR_DIRECTORY_NOT_RM)
			A(ERROR_TRANSACTIONS_UNSUPPORTED_REMOTE)
			A(ERROR_LOG_RESIZE_INVALID_SIZE)
			A(ERROR_OBJECT_NO_LONGER_EXISTS)
			A(ERROR_STREAM_MINIVERSION_NOT_FOUND)
			A(ERROR_STREAM_MINIVERSION_NOT_VALID)
			A(ERROR_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION)
			A(ERROR_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT)
			A(ERROR_CANT_CREATE_MORE_STREAM_MINIVERSIONS)
			A(ERROR_REMOTE_FILE_VERSION_MISMATCH)
			A(ERROR_HANDLE_NO_LONGER_VALID)
			A(ERROR_NO_TXF_METADATA)
			A(ERROR_LOG_CORRUPTION_DETECTED)
			A(ERROR_CANT_RECOVER_WITH_HANDLE_OPEN)
			A(ERROR_RM_DISCONNECTED)
			A(ERROR_ENLISTMENT_NOT_SUPERIOR)
			A(ERROR_RECOVERY_NOT_NEEDED)
			A(ERROR_RM_ALREADY_STARTED)
			A(ERROR_FILE_IDENTITY_NOT_PERSISTENT)
			A(ERROR_CANT_BREAK_TRANSACTIONAL_DEPENDENCY)
			A(ERROR_CANT_CROSS_RM_BOUNDARY)
			A(ERROR_TXF_DIR_NOT_EMPTY)
			A(ERROR_INDOUBT_TRANSACTIONS_EXIST)
			A(ERROR_TM_VOLATILE)
			A(ERROR_ROLLBACK_TIMER_EXPIRED)
			A(ERROR_TXF_ATTRIBUTE_CORRUPT)
			A(ERROR_EFS_NOT_ALLOWED_IN_TRANSACTION)
			A(ERROR_TRANSACTIONAL_OPEN_NOT_ALLOWED)
			A(ERROR_LOG_GROWTH_FAILED)
			A(ERROR_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE)
			A(ERROR_TXF_METADATA_ALREADY_PRESENT)
			A(ERROR_TRANSACTION_SCOPE_CALLBACKS_NOT_SET)
			A(ERROR_TRANSACTION_REQUIRED_PROMOTION)
			A(ERROR_CANNOT_EXECUTE_FILE_IN_TRANSACTION)
			A(ERROR_TRANSACTIONS_NOT_FROZEN)
			A(ERROR_TRANSACTION_FREEZE_IN_PROGRESS)
			A(ERROR_NOT_SNAPSHOT_VOLUME)
			A(ERROR_NO_SAVEPOINT_WITH_OPEN_FILES)
			A(ERROR_DATA_LOST_REPAIR)
			A(ERROR_SPARSE_NOT_ALLOWED_IN_TRANSACTION)
			A(ERROR_TM_IDENTITY_MISMATCH)
			A(ERROR_FLOATED_SECTION)
			A(ERROR_CANNOT_ACCEPT_TRANSACTED_WORK)
			A(ERROR_CANNOT_ABORT_TRANSACTIONS)
			A(ERROR_BAD_CLUSTERS)
			A(ERROR_COMPRESSION_NOT_ALLOWED_IN_TRANSACTION)
			A(ERROR_VOLUME_DIRTY)
			A(ERROR_NO_LINK_TRACKING_IN_TRANSACTION)
			A(ERROR_OPERATION_NOT_SUPPORTED_IN_TRANSACTION)
			A(ERROR_EXPIRED_HANDLE)
			A(ERROR_TRANSACTION_NOT_ENLISTED)
			A(ERROR_CTX_WINSTATION_NAME_INVALID)
			A(ERROR_CTX_INVALID_PD)
			A(ERROR_CTX_PD_NOT_FOUND)
			A(ERROR_CTX_WD_NOT_FOUND)
			A(ERROR_CTX_CANNOT_MAKE_EVENTLOG_ENTRY)
			A(ERROR_CTX_SERVICE_NAME_COLLISION)
			A(ERROR_CTX_CLOSE_PENDING)
			A(ERROR_CTX_NO_OUTBUF)
			A(ERROR_CTX_MODEM_INF_NOT_FOUND)
			A(ERROR_CTX_INVALID_MODEMNAME)
			A(ERROR_CTX_MODEM_RESPONSE_ERROR)
			A(ERROR_CTX_MODEM_RESPONSE_TIMEOUT)
			A(ERROR_CTX_MODEM_RESPONSE_NO_CARRIER)
			A(ERROR_CTX_MODEM_RESPONSE_NO_DIALTONE)
			A(ERROR_CTX_MODEM_RESPONSE_BUSY)
			A(ERROR_CTX_MODEM_RESPONSE_VOICE)
			A(ERROR_CTX_TD_ERROR)
			A(ERROR_CTX_WINSTATION_NOT_FOUND)
			A(ERROR_CTX_WINSTATION_ALREADY_EXISTS)
			A(ERROR_CTX_WINSTATION_BUSY)
			A(ERROR_CTX_BAD_VIDEO_MODE)
			A(ERROR_CTX_GRAPHICS_INVALID)
			A(ERROR_CTX_LOGON_DISABLED)
			A(ERROR_CTX_NOT_CONSOLE)
			A(ERROR_CTX_CLIENT_QUERY_TIMEOUT)
			A(ERROR_CTX_CONSOLE_DISCONNECT)
			A(ERROR_CTX_CONSOLE_CONNECT)
			A(ERROR_CTX_SHADOW_DENIED)
			A(ERROR_CTX_WINSTATION_ACCESS_DENIED)
			A(ERROR_CTX_INVALID_WD)
			A(ERROR_CTX_SHADOW_INVALID)
			A(ERROR_CTX_SHADOW_DISABLED)
			A(ERROR_CTX_CLIENT_LICENSE_IN_USE)
			A(ERROR_CTX_CLIENT_LICENSE_NOT_SET)
			A(ERROR_CTX_LICENSE_NOT_AVAILABLE)
			A(ERROR_CTX_LICENSE_CLIENT_INVALID)
			A(ERROR_CTX_LICENSE_EXPIRED)
			A(ERROR_CTX_SHADOW_NOT_RUNNING)
			A(ERROR_CTX_SHADOW_ENDED_BY_MODE_CHANGE)
			A(ERROR_ACTIVATION_COUNT_EXCEEDED)
			A(ERROR_CTX_WINSTATIONS_DISABLED)
			A(ERROR_CTX_ENCRYPTION_LEVEL_REQUIRED)
			A(ERROR_CTX_SESSION_IN_USE)
			A(ERROR_CTX_NO_FORCE_LOGOFF)
			A(ERROR_CTX_ACCOUNT_RESTRICTION)
			A(ERROR_RDP_PROTOCOL_ERROR)
			A(ERROR_CTX_CDM_CONNECT)
			A(ERROR_CTX_CDM_DISCONNECT)
			A(ERROR_CTX_SECURITY_LAYER_ERROR)
			A(ERROR_TS_INCOMPATIBLE_SESSIONS)
			A(ERROR_TS_VIDEO_SUBSYSTEM_ERROR)
			A(ERROR_DS_NOT_INSTALLED)
			A(ERROR_DS_MEMBERSHIP_EVALUATED_LOCALLY)
			A(ERROR_DS_NO_ATTRIBUTE_OR_VALUE)
			A(ERROR_DS_INVALID_ATTRIBUTE_SYNTAX)
			A(ERROR_DS_ATTRIBUTE_TYPE_UNDEFINED)
			A(ERROR_DS_ATTRIBUTE_OR_VALUE_EXISTS)
			A(ERROR_DS_BUSY)
			A(ERROR_DS_UNAVAILABLE)
			A(ERROR_DS_NO_RIDS_ALLOCATED)
			A(ERROR_DS_NO_MORE_RIDS)
			A(ERROR_DS_INCORRECT_ROLE_OWNER)
			A(ERROR_DS_RIDMGR_INIT_ERROR)
			A(ERROR_DS_OBJ_CLASS_VIOLATION)
			A(ERROR_DS_CANT_ON_NON_LEAF)
			A(ERROR_DS_CANT_ON_RDN)
			A(ERROR_DS_CANT_MOD_OBJ_CLASS)
			A(ERROR_DS_CROSS_DOM_MOVE_ERROR)
			A(ERROR_DS_GC_NOT_AVAILABLE)
			A(ERROR_SHARED_POLICY)
			A(ERROR_POLICY_OBJECT_NOT_FOUND)
			A(ERROR_POLICY_ONLY_IN_DS)
			A(ERROR_PROMOTION_ACTIVE)
			A(ERROR_NO_PROMOTION_ACTIVE)
			A(ERROR_DS_OPERATIONS_ERROR)
			A(ERROR_DS_PROTOCOL_ERROR)
			A(ERROR_DS_TIMELIMIT_EXCEEDED)
			A(ERROR_DS_SIZELIMIT_EXCEEDED)
			A(ERROR_DS_ADMIN_LIMIT_EXCEEDED)
			A(ERROR_DS_COMPARE_FALSE)
			A(ERROR_DS_COMPARE_TRUE)
			A(ERROR_DS_AUTH_METHOD_NOT_SUPPORTED)
			A(ERROR_DS_STRONG_AUTH_REQUIRED)
			A(ERROR_DS_INAPPROPRIATE_AUTH)
			A(ERROR_DS_AUTH_UNKNOWN)
			A(ERROR_DS_REFERRAL)
			A(ERROR_DS_UNAVAILABLE_CRIT_EXTENSION)
			A(ERROR_DS_CONFIDENTIALITY_REQUIRED)
			A(ERROR_DS_INAPPROPRIATE_MATCHING)
			A(ERROR_DS_CONSTRAINT_VIOLATION)
			A(ERROR_DS_NO_SUCH_OBJECT)
			A(ERROR_DS_ALIAS_PROBLEM)
			A(ERROR_DS_INVALID_DN_SYNTAX)
			A(ERROR_DS_IS_LEAF)
			A(ERROR_DS_ALIAS_DEREF_PROBLEM)
			A(ERROR_DS_UNWILLING_TO_PERFORM)
			A(ERROR_DS_LOOP_DETECT)
			A(ERROR_DS_NAMING_VIOLATION)
			A(ERROR_DS_OBJECT_RESULTS_TOO_LARGE)
			A(ERROR_DS_AFFECTS_MULTIPLE_DSAS)
			A(ERROR_DS_SERVER_DOWN)
			A(ERROR_DS_LOCAL_ERROR)
			A(ERROR_DS_ENCODING_ERROR)
			A(ERROR_DS_DECODING_ERROR)
			A(ERROR_DS_FILTER_UNKNOWN)
			A(ERROR_DS_PARAM_ERROR)
			A(ERROR_DS_NOT_SUPPORTED)
			A(ERROR_DS_NO_RESULTS_RETURNED)
			A(ERROR_DS_CONTROL_NOT_FOUND)
			A(ERROR_DS_CLIENT_LOOP)
			A(ERROR_DS_REFERRAL_LIMIT_EXCEEDED)
			A(ERROR_DS_SORT_CONTROL_MISSING)
			A(ERROR_DS_OFFSET_RANGE_ERROR)
			A(ERROR_DS_ROOT_MUST_BE_NC)
			A(ERROR_DS_ADD_REPLICA_INHIBITED)
			A(ERROR_DS_ATT_NOT_DEF_IN_SCHEMA)
			A(ERROR_DS_MAX_OBJ_SIZE_EXCEEDED)
			A(ERROR_DS_OBJ_STRING_NAME_EXISTS)
			A(ERROR_DS_NO_RDN_DEFINED_IN_SCHEMA)
			A(ERROR_DS_RDN_DOESNT_MATCH_SCHEMA)
			A(ERROR_DS_NO_REQUESTED_ATTS_FOUND)
			A(ERROR_DS_USER_BUFFER_TO_SMALL)
			A(ERROR_DS_ATT_IS_NOT_ON_OBJ)
			A(ERROR_DS_ILLEGAL_MOD_OPERATION)
			A(ERROR_DS_OBJ_TOO_LARGE)
			A(ERROR_DS_BAD_INSTANCE_TYPE)
			A(ERROR_DS_MASTERDSA_REQUIRED)
			A(ERROR_DS_OBJECT_CLASS_REQUIRED)
			A(ERROR_DS_MISSING_REQUIRED_ATT)
			A(ERROR_DS_ATT_NOT_DEF_FOR_CLASS)
			A(ERROR_DS_ATT_ALREADY_EXISTS)
			A(ERROR_DS_CANT_ADD_ATT_VALUES)
			A(ERROR_DS_SINGLE_VALUE_CONSTRAINT)
			A(ERROR_DS_RANGE_CONSTRAINT)
			A(ERROR_DS_ATT_VAL_ALREADY_EXISTS)
			A(ERROR_DS_CANT_REM_MISSING_ATT)
			A(ERROR_DS_CANT_REM_MISSING_ATT_VAL)
			A(ERROR_DS_ROOT_CANT_BE_SUBREF)
			A(ERROR_DS_NO_CHAINING)
			A(ERROR_DS_NO_CHAINED_EVAL)
			A(ERROR_DS_NO_PARENT_OBJECT)
			A(ERROR_DS_PARENT_IS_AN_ALIAS)
			A(ERROR_DS_CANT_MIX_MASTER_AND_REPS)
			A(ERROR_DS_CHILDREN_EXIST)
			A(ERROR_DS_OBJ_NOT_FOUND)
			A(ERROR_DS_ALIASED_OBJ_MISSING)
			A(ERROR_DS_BAD_NAME_SYNTAX)
			A(ERROR_DS_ALIAS_POINTS_TO_ALIAS)
			A(ERROR_DS_CANT_DEREF_ALIAS)
			A(ERROR_DS_OUT_OF_SCOPE)
			A(ERROR_DS_OBJECT_BEING_REMOVED)
			A(ERROR_DS_CANT_DELETE_DSA_OBJ)
			A(ERROR_DS_GENERIC_ERROR)
			A(ERROR_DS_DSA_MUST_BE_INT_MASTER)
			A(ERROR_DS_CLASS_NOT_DSA)
			A(ERROR_DS_INSUFF_ACCESS_RIGHTS)
			A(ERROR_DS_ILLEGAL_SUPERIOR)
			A(ERROR_DS_ATTRIBUTE_OWNED_BY_SAM)
			A(ERROR_DS_NAME_TOO_MANY_PARTS)
			A(ERROR_DS_NAME_TOO_LONG)
			A(ERROR_DS_NAME_VALUE_TOO_LONG)
			A(ERROR_DS_NAME_UNPARSEABLE)
			A(ERROR_DS_NAME_TYPE_UNKNOWN)
			A(ERROR_DS_NOT_AN_OBJECT)
			A(ERROR_DS_SEC_DESC_TOO_SHORT)
			A(ERROR_DS_SEC_DESC_INVALID)
			A(ERROR_DS_NO_DELETED_NAME)
			A(ERROR_DS_SUBREF_MUST_HAVE_PARENT)
			A(ERROR_DS_NCNAME_MUST_BE_NC)
			A(ERROR_DS_CANT_ADD_SYSTEM_ONLY)
			A(ERROR_DS_CLASS_MUST_BE_CONCRETE)
			A(ERROR_DS_INVALID_DMD)
			A(ERROR_DS_OBJ_GUID_EXISTS)
			A(ERROR_DS_NOT_ON_BACKLINK)
			A(ERROR_DS_NO_CROSSREF_FOR_NC)
			A(ERROR_DS_SHUTTING_DOWN)
			A(ERROR_DS_UNKNOWN_OPERATION)
			A(ERROR_DS_INVALID_ROLE_OWNER)
			A(ERROR_DS_COULDNT_CONTACT_FSMO)
			A(ERROR_DS_CROSS_NC_DN_RENAME)
			A(ERROR_DS_CANT_MOD_SYSTEM_ONLY)
			A(ERROR_DS_REPLICATOR_ONLY)
			A(ERROR_DS_OBJ_CLASS_NOT_DEFINED)
			A(ERROR_DS_OBJ_CLASS_NOT_SUBCLASS)
			A(ERROR_DS_NAME_REFERENCE_INVALID)
			A(ERROR_DS_CROSS_REF_EXISTS)
			A(ERROR_DS_CANT_DEL_MASTER_CROSSREF)
			A(ERROR_DS_SUBTREE_NOTIFY_NOT_NC_HEAD)
			A(ERROR_DS_NOTIFY_FILTER_TOO_COMPLEX)
			A(ERROR_DS_DUP_RDN)
			A(ERROR_DS_DUP_OID)
			A(ERROR_DS_DUP_MAPI_ID)
			A(ERROR_DS_DUP_SCHEMA_ID_GUID)
			A(ERROR_DS_DUP_LDAP_DISPLAY_NAME)
			A(ERROR_DS_SEMANTIC_ATT_TEST)
			A(ERROR_DS_SYNTAX_MISMATCH)
			A(ERROR_DS_EXISTS_IN_MUST_HAVE)
			A(ERROR_DS_EXISTS_IN_MAY_HAVE)
			A(ERROR_DS_NONEXISTENT_MAY_HAVE)
			A(ERROR_DS_NONEXISTENT_MUST_HAVE)
			A(ERROR_DS_AUX_CLS_TEST_FAIL)
			A(ERROR_DS_NONEXISTENT_POSS_SUP)
			A(ERROR_DS_SUB_CLS_TEST_FAIL)
			A(ERROR_DS_BAD_RDN_ATT_ID_SYNTAX)
			A(ERROR_DS_EXISTS_IN_AUX_CLS)
			A(ERROR_DS_EXISTS_IN_SUB_CLS)
			A(ERROR_DS_EXISTS_IN_POSS_SUP)
			A(ERROR_DS_RECALCSCHEMA_FAILED)
			A(ERROR_DS_TREE_DELETE_NOT_FINISHED)
			A(ERROR_DS_CANT_DELETE)
			A(ERROR_DS_ATT_SCHEMA_REQ_ID)
			A(ERROR_DS_BAD_ATT_SCHEMA_SYNTAX)
			A(ERROR_DS_CANT_CACHE_ATT)
			A(ERROR_DS_CANT_CACHE_CLASS)
			A(ERROR_DS_CANT_REMOVE_ATT_CACHE)
			A(ERROR_DS_CANT_REMOVE_CLASS_CACHE)
			A(ERROR_DS_CANT_RETRIEVE_DN)
			A(ERROR_DS_MISSING_SUPREF)
			A(ERROR_DS_CANT_RETRIEVE_INSTANCE)
			A(ERROR_DS_CODE_INCONSISTENCY)
			A(ERROR_DS_DATABASE_ERROR)
			A(ERROR_DS_GOVERNSID_MISSING)
			A(ERROR_DS_MISSING_EXPECTED_ATT)
			A(ERROR_DS_NCNAME_MISSING_CR_REF)
			A(ERROR_DS_SECURITY_CHECKING_ERROR)
			A(ERROR_DS_SCHEMA_NOT_LOADED)
			A(ERROR_DS_SCHEMA_ALLOC_FAILED)
			A(ERROR_DS_ATT_SCHEMA_REQ_SYNTAX)
			A(ERROR_DS_GCVERIFY_ERROR)
			A(ERROR_DS_DRA_SCHEMA_MISMATCH)
			A(ERROR_DS_CANT_FIND_DSA_OBJ)
			A(ERROR_DS_CANT_FIND_EXPECTED_NC)
			A(ERROR_DS_CANT_FIND_NC_IN_CACHE)
			A(ERROR_DS_CANT_RETRIEVE_CHILD)
			A(ERROR_DS_SECURITY_ILLEGAL_MODIFY)
			A(ERROR_DS_CANT_REPLACE_HIDDEN_REC)
			A(ERROR_DS_BAD_HIERARCHY_FILE)
			A(ERROR_DS_BUILD_HIERARCHY_TABLE_FAILED)
			A(ERROR_DS_CONFIG_PARAM_MISSING)
			A(ERROR_DS_COUNTING_AB_INDICES_FAILED)
			A(ERROR_DS_HIERARCHY_TABLE_MALLOC_FAILED)
			A(ERROR_DS_INTERNAL_FAILURE)
			A(ERROR_DS_UNKNOWN_ERROR)
			A(ERROR_DS_ROOT_REQUIRES_CLASS_TOP)
			A(ERROR_DS_REFUSING_FSMO_ROLES)
			A(ERROR_DS_MISSING_FSMO_SETTINGS)
			A(ERROR_DS_UNABLE_TO_SURRENDER_ROLES)
			A(ERROR_DS_DRA_GENERIC)
			A(ERROR_DS_DRA_INVALID_PARAMETER)
			A(ERROR_DS_DRA_BUSY)
			A(ERROR_DS_DRA_BAD_DN)
			A(ERROR_DS_DRA_BAD_NC)
			A(ERROR_DS_DRA_DN_EXISTS)
			A(ERROR_DS_DRA_INTERNAL_ERROR)
			A(ERROR_DS_DRA_INCONSISTENT_DIT)
			A(ERROR_DS_DRA_CONNECTION_FAILED)
			A(ERROR_DS_DRA_BAD_INSTANCE_TYPE)
			A(ERROR_DS_DRA_OUT_OF_MEM)
			A(ERROR_DS_DRA_MAIL_PROBLEM)
			A(ERROR_DS_DRA_REF_ALREADY_EXISTS)
			A(ERROR_DS_DRA_REF_NOT_FOUND)
			A(ERROR_DS_DRA_OBJ_IS_REP_SOURCE)
			A(ERROR_DS_DRA_DB_ERROR)
			A(ERROR_DS_DRA_NO_REPLICA)
			A(ERROR_DS_DRA_ACCESS_DENIED)
			A(ERROR_DS_DRA_NOT_SUPPORTED)
			A(ERROR_DS_DRA_RPC_CANCELLED)
			A(ERROR_DS_DRA_SOURCE_DISABLED)
			A(ERROR_DS_DRA_SINK_DISABLED)
			A(ERROR_DS_DRA_NAME_COLLISION)
			A(ERROR_DS_DRA_SOURCE_REINSTALLED)
			A(ERROR_DS_DRA_MISSING_PARENT)
			A(ERROR_DS_DRA_PREEMPTED)
			A(ERROR_DS_DRA_ABANDON_SYNC)
			A(ERROR_DS_DRA_SHUTDOWN)
			A(ERROR_DS_DRA_INCOMPATIBLE_PARTIAL_SET)
			A(ERROR_DS_DRA_SOURCE_IS_PARTIAL_REPLICA)
			A(ERROR_DS_DRA_EXTN_CONNECTION_FAILED)
			A(ERROR_DS_INSTALL_SCHEMA_MISMATCH)
			A(ERROR_DS_DUP_LINK_ID)
			A(ERROR_DS_NAME_ERROR_RESOLVING)
			A(ERROR_DS_NAME_ERROR_NOT_FOUND)
			A(ERROR_DS_NAME_ERROR_NOT_UNIQUE)
			A(ERROR_DS_NAME_ERROR_NO_MAPPING)
			A(ERROR_DS_NAME_ERROR_DOMAIN_ONLY)
			A(ERROR_DS_NAME_ERROR_NO_SYNTACTICAL_MAPPING)
			A(ERROR_DS_CONSTRUCTED_ATT_MOD)
			A(ERROR_DS_WRONG_OM_OBJ_CLASS)
			A(ERROR_DS_DRA_REPL_PENDING)
			A(ERROR_DS_DS_REQUIRED)
			A(ERROR_DS_INVALID_LDAP_DISPLAY_NAME)
			A(ERROR_DS_NON_BASE_SEARCH)
			A(ERROR_DS_CANT_RETRIEVE_ATTS)
			A(ERROR_DS_BACKLINK_WITHOUT_LINK)
			A(ERROR_DS_EPOCH_MISMATCH)
			A(ERROR_DS_SRC_NAME_MISMATCH)
			A(ERROR_DS_SRC_AND_DST_NC_IDENTICAL)
			A(ERROR_DS_DST_NC_MISMATCH)
			A(ERROR_DS_NOT_AUTHORITIVE_FOR_DST_NC)
			A(ERROR_DS_SRC_GUID_MISMATCH)
			A(ERROR_DS_CANT_MOVE_DELETED_OBJECT)
			A(ERROR_DS_PDC_OPERATION_IN_PROGRESS)
			A(ERROR_DS_CROSS_DOMAIN_CLEANUP_REQD)
			A(ERROR_DS_ILLEGAL_XDOM_MOVE_OPERATION)
			A(ERROR_DS_CANT_WITH_ACCT_GROUP_MEMBERSHPS)
			A(ERROR_DS_NC_MUST_HAVE_NC_PARENT)
			A(ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE)
			A(ERROR_DS_DST_DOMAIN_NOT_NATIVE)
			A(ERROR_DS_MISSING_INFRASTRUCTURE_CONTAINER)
			A(ERROR_DS_CANT_MOVE_ACCOUNT_GROUP)
			A(ERROR_DS_CANT_MOVE_RESOURCE_GROUP)
			A(ERROR_DS_INVALID_SEARCH_FLAG)
			A(ERROR_DS_NO_TREE_DELETE_ABOVE_NC)
			A(ERROR_DS_COULDNT_LOCK_TREE_FOR_DELETE)
			A(ERROR_DS_COULDNT_IDENTIFY_OBJECTS_FOR_TREE_DELETE)
			A(ERROR_DS_SAM_INIT_FAILURE)
			A(ERROR_DS_SENSITIVE_GROUP_VIOLATION)
			A(ERROR_DS_CANT_MOD_PRIMARYGROUPID)
			A(ERROR_DS_ILLEGAL_BASE_SCHEMA_MOD)
			A(ERROR_DS_NONSAFE_SCHEMA_CHANGE)
			A(ERROR_DS_SCHEMA_UPDATE_DISALLOWED)
			A(ERROR_DS_CANT_CREATE_UNDER_SCHEMA)
			A(ERROR_DS_INSTALL_NO_SRC_SCH_VERSION)
			A(ERROR_DS_INSTALL_NO_SCH_VERSION_IN_INIFILE)
			A(ERROR_DS_INVALID_GROUP_TYPE)
			A(ERROR_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN)
			A(ERROR_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN)
			A(ERROR_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER)
			A(ERROR_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER)
			A(ERROR_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER)
			A(ERROR_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER)
			A(ERROR_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER)
			A(ERROR_DS_HAVE_PRIMARY_MEMBERS)
			A(ERROR_DS_STRING_SD_CONVERSION_FAILED)
			A(ERROR_DS_NAMING_MASTER_GC)
			A(ERROR_DS_DNS_LOOKUP_FAILURE)
			A(ERROR_DS_COULDNT_UPDATE_SPNS)
			A(ERROR_DS_CANT_RETRIEVE_SD)
			A(ERROR_DS_KEY_NOT_UNIQUE)
			A(ERROR_DS_WRONG_LINKED_ATT_SYNTAX)
			A(ERROR_DS_SAM_NEED_BOOTKEY_PASSWORD)
			A(ERROR_DS_SAM_NEED_BOOTKEY_FLOPPY)
			A(ERROR_DS_CANT_START)
			A(ERROR_DS_INIT_FAILURE)
			A(ERROR_DS_NO_PKT_PRIVACY_ON_CONNECTION)
			A(ERROR_DS_SOURCE_DOMAIN_IN_FOREST)
			A(ERROR_DS_DESTINATION_DOMAIN_NOT_IN_FOREST)
			A(ERROR_DS_DESTINATION_AUDITING_NOT_ENABLED)
			A(ERROR_DS_CANT_FIND_DC_FOR_SRC_DOMAIN)
			A(ERROR_DS_SRC_OBJ_NOT_GROUP_OR_USER)
			A(ERROR_DS_SRC_SID_EXISTS_IN_FOREST)
			A(ERROR_DS_SRC_AND_DST_OBJECT_CLASS_MISMATCH)
			A(ERROR_SAM_INIT_FAILURE)
			A(ERROR_DS_DRA_SCHEMA_INFO_SHIP)
			A(ERROR_DS_DRA_SCHEMA_CONFLICT)
			A(ERROR_DS_DRA_EARLIER_SCHEMA_CONFLICT)
			A(ERROR_DS_DRA_OBJ_NC_MISMATCH)
			A(ERROR_DS_NC_STILL_HAS_DSAS)
			A(ERROR_DS_GC_REQUIRED)
			A(ERROR_DS_LOCAL_MEMBER_OF_LOCAL_ONLY)
			A(ERROR_DS_NO_FPO_IN_UNIVERSAL_GROUPS)
			A(ERROR_DS_CANT_ADD_TO_GC)
			A(ERROR_DS_NO_CHECKPOINT_WITH_PDC)
			A(ERROR_DS_SOURCE_AUDITING_NOT_ENABLED)
			A(ERROR_DS_CANT_CREATE_IN_NONDOMAIN_NC)
			A(ERROR_DS_INVALID_NAME_FOR_SPN)
			A(ERROR_DS_FILTER_USES_CONTRUCTED_ATTRS)
			A(ERROR_DS_UNICODEPWD_NOT_IN_QUOTES)
			A(ERROR_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED)
			A(ERROR_DS_MUST_BE_RUN_ON_DST_DC)
			A(ERROR_DS_SRC_DC_MUST_BE_SP4_OR_GREATER)
			A(ERROR_DS_CANT_TREE_DELETE_CRITICAL_OBJ)
			A(ERROR_DS_INIT_FAILURE_CONSOLE)
			A(ERROR_DS_SAM_INIT_FAILURE_CONSOLE)
			A(ERROR_DS_FOREST_VERSION_TOO_HIGH)
			A(ERROR_DS_DOMAIN_VERSION_TOO_HIGH)
			A(ERROR_DS_FOREST_VERSION_TOO_LOW)
			A(ERROR_DS_DOMAIN_VERSION_TOO_LOW)
			A(ERROR_DS_INCOMPATIBLE_VERSION)
			A(ERROR_DS_LOW_DSA_VERSION)
			A(ERROR_DS_NO_BEHAVIOR_VERSION_IN_MIXEDDOMAIN)
			A(ERROR_DS_NOT_SUPPORTED_SORT_ORDER)
			A(ERROR_DS_NAME_NOT_UNIQUE)
			A(ERROR_DS_MACHINE_ACCOUNT_CREATED_PRENT4)
			A(ERROR_DS_OUT_OF_VERSION_STORE)
			A(ERROR_DS_INCOMPATIBLE_CONTROLS_USED)
			A(ERROR_DS_NO_REF_DOMAIN)
			A(ERROR_DS_RESERVED_LINK_ID)
			A(ERROR_DS_LINK_ID_NOT_AVAILABLE)
			A(ERROR_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER)
			A(ERROR_DS_MODIFYDN_DISALLOWED_BY_INSTANCE_TYPE)
			A(ERROR_DS_NO_OBJECT_MOVE_IN_SCHEMA_NC)
			A(ERROR_DS_MODIFYDN_DISALLOWED_BY_FLAG)
			A(ERROR_DS_MODIFYDN_WRONG_GRANDPARENT)
			A(ERROR_DS_NAME_ERROR_TRUST_REFERRAL)
			A(ERROR_NOT_SUPPORTED_ON_STANDARD_SERVER)
			A(ERROR_DS_CANT_ACCESS_REMOTE_PART_OF_AD)
			A(ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE_V2)
			A(ERROR_DS_THREAD_LIMIT_EXCEEDED)
			A(ERROR_DS_NOT_CLOSEST)
			A(ERROR_DS_CANT_DERIVE_SPN_WITHOUT_SERVER_REF)
			A(ERROR_DS_SINGLE_USER_MODE_FAILED)
			A(ERROR_DS_NTDSCRIPT_SYNTAX_ERROR)
			A(ERROR_DS_NTDSCRIPT_PROCESS_ERROR)
			A(ERROR_DS_DIFFERENT_REPL_EPOCHS)
			A(ERROR_DS_DRS_EXTENSIONS_CHANGED)
			A(ERROR_DS_REPLICA_SET_CHANGE_NOT_ALLOWED_ON_DISABLED_CR)
			A(ERROR_DS_NO_MSDS_INTID)
			A(ERROR_DS_DUP_MSDS_INTID)
			A(ERROR_DS_EXISTS_IN_RDNATTID)
			A(ERROR_DS_AUTHORIZATION_FAILED)
			A(ERROR_DS_INVALID_SCRIPT)
			A(ERROR_DS_REMOTE_CROSSREF_OP_FAILED)
			A(ERROR_DS_CROSS_REF_BUSY)
			A(ERROR_DS_CANT_DERIVE_SPN_FOR_DELETED_DOMAIN)
			A(ERROR_DS_CANT_DEMOTE_WITH_WRITEABLE_NC)
			A(ERROR_DS_DUPLICATE_ID_FOUND)
			A(ERROR_DS_INSUFFICIENT_ATTR_TO_CREATE_OBJECT)
			A(ERROR_DS_GROUP_CONVERSION_ERROR)
			A(ERROR_DS_CANT_MOVE_APP_BASIC_GROUP)
			A(ERROR_DS_CANT_MOVE_APP_QUERY_GROUP)
			A(ERROR_DS_ROLE_NOT_VERIFIED)
			A(ERROR_DS_WKO_CONTAINER_CANNOT_BE_SPECIAL)
			A(ERROR_DS_DOMAIN_RENAME_IN_PROGRESS)
			A(ERROR_DS_EXISTING_AD_CHILD_NC)
			A(ERROR_DS_REPL_LIFETIME_EXCEEDED)
			A(ERROR_DS_DISALLOWED_IN_SYSTEM_CONTAINER)
			A(ERROR_DS_LDAP_SEND_QUEUE_FULL)
			A(ERROR_DS_DRA_OUT_SCHEDULE_WINDOW)
			A(ERROR_DS_POLICY_NOT_KNOWN)
			A(ERROR_NO_SITE_SETTINGS_OBJECT)
			A(ERROR_NO_SECRETS)
			A(ERROR_NO_WRITABLE_DC_FOUND)
			A(ERROR_DS_NO_SERVER_OBJECT)
			A(ERROR_DS_NO_NTDSA_OBJECT)
			A(ERROR_DS_NON_ASQ_SEARCH)
			A(ERROR_DS_AUDIT_FAILURE)
			A(ERROR_DS_INVALID_SEARCH_FLAG_SUBTREE)
			A(ERROR_DS_INVALID_SEARCH_FLAG_TUPLE)
			A(ERROR_DS_HIERARCHY_TABLE_TOO_DEEP)
			A(ERROR_DS_DRA_CORRUPT_UTD_VECTOR)
			A(ERROR_DS_DRA_SECRETS_DENIED)
			A(ERROR_DS_RESERVED_MAPI_ID)
			A(ERROR_DS_MAPI_ID_NOT_AVAILABLE)
			A(ERROR_DS_DRA_MISSING_KRBTGT_SECRET)
			A(ERROR_DS_DOMAIN_NAME_EXISTS_IN_FOREST)
			A(ERROR_DS_FLAT_NAME_EXISTS_IN_FOREST)
			A(ERROR_INVALID_USER_PRINCIPAL_NAME)
			A(ERROR_DS_OID_MAPPED_GROUP_CANT_HAVE_MEMBERS)
			A(ERROR_DS_OID_NOT_FOUND)
			A(ERROR_DS_DRA_RECYCLED_TARGET)
			A(ERROR_IPSEC_QM_POLICY_EXISTS)
			A(ERROR_IPSEC_QM_POLICY_NOT_FOUND)
			A(ERROR_IPSEC_QM_POLICY_IN_USE)
			A(ERROR_IPSEC_MM_POLICY_EXISTS)
			A(ERROR_IPSEC_MM_POLICY_NOT_FOUND)
			A(ERROR_IPSEC_MM_POLICY_IN_USE)
			A(ERROR_IPSEC_MM_FILTER_EXISTS)
			A(ERROR_IPSEC_MM_FILTER_NOT_FOUND)
			A(ERROR_IPSEC_TRANSPORT_FILTER_EXISTS)
			A(ERROR_IPSEC_TRANSPORT_FILTER_NOT_FOUND)
			A(ERROR_IPSEC_MM_AUTH_EXISTS)
			A(ERROR_IPSEC_MM_AUTH_NOT_FOUND)
			A(ERROR_IPSEC_MM_AUTH_IN_USE)
			A(ERROR_IPSEC_DEFAULT_MM_POLICY_NOT_FOUND)
			A(ERROR_IPSEC_DEFAULT_MM_AUTH_NOT_FOUND)
			A(ERROR_IPSEC_DEFAULT_QM_POLICY_NOT_FOUND)
			A(ERROR_IPSEC_TUNNEL_FILTER_EXISTS)
			A(ERROR_IPSEC_TUNNEL_FILTER_NOT_FOUND)
			A(ERROR_IPSEC_MM_FILTER_PENDING_DELETION)
			A(ERROR_IPSEC_TRANSPORT_FILTER_PENDING_DELETION)
			A(ERROR_IPSEC_TUNNEL_FILTER_PENDING_DELETION)
			A(ERROR_IPSEC_MM_POLICY_PENDING_DELETION)
			A(ERROR_IPSEC_MM_AUTH_PENDING_DELETION)
			A(ERROR_IPSEC_QM_POLICY_PENDING_DELETION)
			A(ERROR_IPSEC_IKE_NEG_STATUS_BEGIN)
			A(ERROR_IPSEC_IKE_AUTH_FAIL)
			A(ERROR_IPSEC_IKE_ATTRIB_FAIL)
			A(ERROR_IPSEC_IKE_NEGOTIATION_PENDING)
			A(ERROR_IPSEC_IKE_GENERAL_PROCESSING_ERROR)
			A(ERROR_IPSEC_IKE_TIMED_OUT)
			A(ERROR_IPSEC_IKE_NO_CERT)
			A(ERROR_IPSEC_IKE_SA_DELETED)
			A(ERROR_IPSEC_IKE_SA_REAPED)
			A(ERROR_IPSEC_IKE_MM_ACQUIRE_DROP)
			A(ERROR_IPSEC_IKE_QM_ACQUIRE_DROP)
			A(ERROR_IPSEC_IKE_QUEUE_DROP_MM)
			A(ERROR_IPSEC_IKE_QUEUE_DROP_NO_MM)
			A(ERROR_IPSEC_IKE_DROP_NO_RESPONSE)
			A(ERROR_IPSEC_IKE_MM_DELAY_DROP)
			A(ERROR_IPSEC_IKE_QM_DELAY_DROP)
			A(ERROR_IPSEC_IKE_ERROR)
			A(ERROR_IPSEC_IKE_CRL_FAILED)
			A(ERROR_IPSEC_IKE_INVALID_KEY_USAGE)
			A(ERROR_IPSEC_IKE_INVALID_CERT_TYPE)
			A(ERROR_IPSEC_IKE_NO_PRIVATE_KEY)
			A(ERROR_IPSEC_IKE_SIMULTANEOUS_REKEY)
			A(ERROR_IPSEC_IKE_DH_FAIL)
			A(ERROR_IPSEC_IKE_CRITICAL_PAYLOAD_NOT_RECOGNIZED)
			A(ERROR_IPSEC_IKE_INVALID_HEADER)
			A(ERROR_IPSEC_IKE_NO_POLICY)
			A(ERROR_IPSEC_IKE_INVALID_SIGNATURE)
			A(ERROR_IPSEC_IKE_KERBEROS_ERROR)
			A(ERROR_IPSEC_IKE_NO_PUBLIC_KEY)
			A(ERROR_IPSEC_IKE_PROCESS_ERR)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_SA)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_PROP)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_TRANS)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_KE)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_ID)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_CERT)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_CERT_REQ)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_HASH)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_SIG)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_NONCE)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_NOTIFY)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_DELETE)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_VENDOR)
			A(ERROR_IPSEC_IKE_INVALID_PAYLOAD)
			A(ERROR_IPSEC_IKE_LOAD_SOFT_SA)
			A(ERROR_IPSEC_IKE_SOFT_SA_TORN_DOWN)
			A(ERROR_IPSEC_IKE_INVALID_COOKIE)
			A(ERROR_IPSEC_IKE_NO_PEER_CERT)
			A(ERROR_IPSEC_IKE_PEER_CRL_FAILED)
			A(ERROR_IPSEC_IKE_POLICY_CHANGE)
			A(ERROR_IPSEC_IKE_NO_MM_POLICY)
			A(ERROR_IPSEC_IKE_NOTCBPRIV)
			A(ERROR_IPSEC_IKE_SECLOADFAIL)
			A(ERROR_IPSEC_IKE_FAILSSPINIT)
			A(ERROR_IPSEC_IKE_FAILQUERYSSP)
			A(ERROR_IPSEC_IKE_SRVACQFAIL)
			A(ERROR_IPSEC_IKE_SRVQUERYCRED)
			A(ERROR_IPSEC_IKE_GETSPIFAIL)
			A(ERROR_IPSEC_IKE_INVALID_FILTER)
			A(ERROR_IPSEC_IKE_OUT_OF_MEMORY)
			A(ERROR_IPSEC_IKE_ADD_UPDATE_KEY_FAILED)
			A(ERROR_IPSEC_IKE_INVALID_POLICY)
			A(ERROR_IPSEC_IKE_UNKNOWN_DOI)
			A(ERROR_IPSEC_IKE_INVALID_SITUATION)
			A(ERROR_IPSEC_IKE_DH_FAILURE)
			A(ERROR_IPSEC_IKE_INVALID_GROUP)
			A(ERROR_IPSEC_IKE_ENCRYPT)
			A(ERROR_IPSEC_IKE_DECRYPT)
			A(ERROR_IPSEC_IKE_POLICY_MATCH)
			A(ERROR_IPSEC_IKE_UNSUPPORTED_ID)
			A(ERROR_IPSEC_IKE_INVALID_HASH)
			A(ERROR_IPSEC_IKE_INVALID_HASH_ALG)
			A(ERROR_IPSEC_IKE_INVALID_HASH_SIZE)
			A(ERROR_IPSEC_IKE_INVALID_ENCRYPT_ALG)
			A(ERROR_IPSEC_IKE_INVALID_AUTH_ALG)
			A(ERROR_IPSEC_IKE_INVALID_SIG)
			A(ERROR_IPSEC_IKE_LOAD_FAILED)
			A(ERROR_IPSEC_IKE_RPC_DELETE)
			A(ERROR_IPSEC_IKE_BENIGN_REINIT)
			A(ERROR_IPSEC_IKE_INVALID_RESPONDER_LIFETIME_NOTIFY)
			A(ERROR_IPSEC_IKE_INVALID_MAJOR_VERSION)
			A(ERROR_IPSEC_IKE_INVALID_CERT_KEYLEN)
			A(ERROR_IPSEC_IKE_MM_LIMIT)
			A(ERROR_IPSEC_IKE_NEGOTIATION_DISABLED)
			A(ERROR_IPSEC_IKE_QM_LIMIT)
			A(ERROR_IPSEC_IKE_MM_EXPIRED)
			A(ERROR_IPSEC_IKE_PEER_MM_ASSUMED_INVALID)
			A(ERROR_IPSEC_IKE_CERT_CHAIN_POLICY_MISMATCH)
			A(ERROR_IPSEC_IKE_UNEXPECTED_MESSAGE_ID)
			A(ERROR_IPSEC_IKE_INVALID_AUTH_PAYLOAD)
			A(ERROR_IPSEC_IKE_DOS_COOKIE_SENT)
			A(ERROR_IPSEC_IKE_SHUTTING_DOWN)
			A(ERROR_IPSEC_IKE_CGA_AUTH_FAILED)
			A(ERROR_IPSEC_IKE_PROCESS_ERR_NATOA)
			A(ERROR_IPSEC_IKE_INVALID_MM_FOR_QM)
			A(ERROR_IPSEC_IKE_QM_EXPIRED)
			A(ERROR_IPSEC_IKE_TOO_MANY_FILTERS)
			A(ERROR_IPSEC_IKE_NEG_STATUS_END)
			A(ERROR_IPSEC_IKE_KILL_DUMMY_NAP_TUNNEL)
			A(ERROR_IPSEC_IKE_INNER_IP_ASSIGNMENT_FAILURE)
			A(ERROR_IPSEC_IKE_REQUIRE_CP_PAYLOAD_MISSING)
			A(ERROR_IPSEC_KEY_MODULE_IMPERSONATION_NEGOTIATION_PENDING)
			A(ERROR_IPSEC_IKE_COEXISTENCE_SUPPRESS)
			A(ERROR_IPSEC_IKE_RATELIMIT_DROP)
			A(ERROR_IPSEC_IKE_PEER_DOESNT_SUPPORT_MOBIKE)
			A(ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE)
			A(ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_FAILURE)
			A(ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE_WITH_OPTIONAL_RETRY)
			A(ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_AND_CERTMAP_FAILURE)
			A(ERROR_IPSEC_IKE_NEG_STATUS_EXTENDED_END)
			A(ERROR_IPSEC_BAD_SPI)
			A(ERROR_IPSEC_SA_LIFETIME_EXPIRED)
			A(ERROR_IPSEC_WRONG_SA)
			A(ERROR_IPSEC_REPLAY_CHECK_FAILED)
			A(ERROR_IPSEC_INVALID_PACKET)
			A(ERROR_IPSEC_INTEGRITY_CHECK_FAILED)
			A(ERROR_IPSEC_CLEAR_TEXT_DROP)
			A(ERROR_IPSEC_AUTH_FIREWALL_DROP)
			A(ERROR_IPSEC_THROTTLE_DROP)
			A(ERROR_IPSEC_DOSP_BLOCK)
			A(ERROR_IPSEC_DOSP_RECEIVED_MULTICAST)
			A(ERROR_IPSEC_DOSP_INVALID_PACKET)
			A(ERROR_IPSEC_DOSP_STATE_LOOKUP_FAILED)
			A(ERROR_IPSEC_DOSP_MAX_ENTRIES)
			A(ERROR_IPSEC_DOSP_KEYMOD_NOT_ALLOWED)
			A(ERROR_IPSEC_DOSP_NOT_INSTALLED)
			A(ERROR_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES)
			A(ERROR_SXS_SECTION_NOT_FOUND)
			A(ERROR_SXS_CANT_GEN_ACTCTX)
			A(ERROR_SXS_INVALID_ACTCTXDATA_FORMAT)
			A(ERROR_SXS_ASSEMBLY_NOT_FOUND)
			A(ERROR_SXS_MANIFEST_FORMAT_ERROR)
			A(ERROR_SXS_MANIFEST_PARSE_ERROR)
			A(ERROR_SXS_ACTIVATION_CONTEXT_DISABLED)
			A(ERROR_SXS_KEY_NOT_FOUND)
			A(ERROR_SXS_VERSION_CONFLICT)
			A(ERROR_SXS_WRONG_SECTION_TYPE)
			A(ERROR_SXS_THREAD_QUERIES_DISABLED)
			A(ERROR_SXS_PROCESS_DEFAULT_ALREADY_SET)
			A(ERROR_SXS_UNKNOWN_ENCODING_GROUP)
			A(ERROR_SXS_UNKNOWN_ENCODING)
			A(ERROR_SXS_INVALID_XML_NAMESPACE_URI)
			A(ERROR_SXS_ROOT_MANIFEST_DEPENDENCY_NOT_INSTALLED)
			A(ERROR_SXS_LEAF_MANIFEST_DEPENDENCY_NOT_INSTALLED)
			A(ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE)
			A(ERROR_SXS_MANIFEST_MISSING_REQUIRED_DEFAULT_NAMESPACE)
			A(ERROR_SXS_MANIFEST_INVALID_REQUIRED_DEFAULT_NAMESPACE)
			A(ERROR_SXS_PRIVATE_MANIFEST_CROSS_PATH_WITH_REPARSE_POINT)
			A(ERROR_SXS_DUPLICATE_DLL_NAME)
			A(ERROR_SXS_DUPLICATE_WINDOWCLASS_NAME)
			A(ERROR_SXS_DUPLICATE_CLSID)
			A(ERROR_SXS_DUPLICATE_IID)
			A(ERROR_SXS_DUPLICATE_TLBID)
			A(ERROR_SXS_DUPLICATE_PROGID)
			A(ERROR_SXS_DUPLICATE_ASSEMBLY_NAME)
			A(ERROR_SXS_FILE_HASH_MISMATCH)
			A(ERROR_SXS_POLICY_PARSE_ERROR)
			A(ERROR_SXS_XML_E_MISSINGQUOTE)
			A(ERROR_SXS_XML_E_COMMENTSYNTAX)
			A(ERROR_SXS_XML_E_BADSTARTNAMECHAR)
			A(ERROR_SXS_XML_E_BADNAMECHAR)
			A(ERROR_SXS_XML_E_BADCHARINSTRING)
			A(ERROR_SXS_XML_E_XMLDECLSYNTAX)
			A(ERROR_SXS_XML_E_BADCHARDATA)
			A(ERROR_SXS_XML_E_MISSINGWHITESPACE)
			A(ERROR_SXS_XML_E_EXPECTINGTAGEND)
			A(ERROR_SXS_XML_E_MISSINGSEMICOLON)
			A(ERROR_SXS_XML_E_UNBALANCEDPAREN)
			A(ERROR_SXS_XML_E_INTERNALERROR)
			A(ERROR_SXS_XML_E_UNEXPECTED_WHITESPACE)
			A(ERROR_SXS_XML_E_INCOMPLETE_ENCODING)
			A(ERROR_SXS_XML_E_MISSING_PAREN)
			A(ERROR_SXS_XML_E_EXPECTINGCLOSEQUOTE)
			A(ERROR_SXS_XML_E_MULTIPLE_COLONS)
			A(ERROR_SXS_XML_E_INVALID_DECIMAL)
			A(ERROR_SXS_XML_E_INVALID_HEXIDECIMAL)
			A(ERROR_SXS_XML_E_INVALID_UNICODE)
			A(ERROR_SXS_XML_E_WHITESPACEORQUESTIONMARK)
			A(ERROR_SXS_XML_E_UNEXPECTEDENDTAG)
			A(ERROR_SXS_XML_E_UNCLOSEDTAG)
			A(ERROR_SXS_XML_E_DUPLICATEATTRIBUTE)
			A(ERROR_SXS_XML_E_MULTIPLEROOTS)
			A(ERROR_SXS_XML_E_INVALIDATROOTLEVEL)
			A(ERROR_SXS_XML_E_BADXMLDECL)
			A(ERROR_SXS_XML_E_MISSINGROOT)
			A(ERROR_SXS_XML_E_UNEXPECTEDEOF)
			A(ERROR_SXS_XML_E_BADPEREFINSUBSET)
			A(ERROR_SXS_XML_E_UNCLOSEDSTARTTAG)
			A(ERROR_SXS_XML_E_UNCLOSEDENDTAG)
			A(ERROR_SXS_XML_E_UNCLOSEDSTRING)
			A(ERROR_SXS_XML_E_UNCLOSEDCOMMENT)
			A(ERROR_SXS_XML_E_UNCLOSEDDECL)
			A(ERROR_SXS_XML_E_UNCLOSEDCDATA)
			A(ERROR_SXS_XML_E_RESERVEDNAMESPACE)
			A(ERROR_SXS_XML_E_INVALIDENCODING)
			A(ERROR_SXS_XML_E_INVALIDSWITCH)
			A(ERROR_SXS_XML_E_BADXMLCASE)
			A(ERROR_SXS_XML_E_INVALID_STANDALONE)
			A(ERROR_SXS_XML_E_UNEXPECTED_STANDALONE)
			A(ERROR_SXS_XML_E_INVALID_VERSION)
			A(ERROR_SXS_XML_E_MISSINGEQUALS)
			A(ERROR_SXS_PROTECTION_RECOVERY_FAILED)
			A(ERROR_SXS_PROTECTION_PUBLIC_KEY_TOO_SHORT)
			A(ERROR_SXS_PROTECTION_CATALOG_NOT_VALID)
			A(ERROR_SXS_UNTRANSLATABLE_HRESULT)
			A(ERROR_SXS_PROTECTION_CATALOG_FILE_MISSING)
			A(ERROR_SXS_MISSING_ASSEMBLY_IDENTITY_ATTRIBUTE)
			A(ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE_NAME)
			A(ERROR_SXS_ASSEMBLY_MISSING)
			A(ERROR_SXS_CORRUPT_ACTIVATION_STACK)
			A(ERROR_SXS_CORRUPTION)
			A(ERROR_SXS_EARLY_DEACTIVATION)
			A(ERROR_SXS_INVALID_DEACTIVATION)
			A(ERROR_SXS_MULTIPLE_DEACTIVATION)
			A(ERROR_SXS_PROCESS_TERMINATION_REQUESTED)
			A(ERROR_SXS_RELEASE_ACTIVATION_CONTEXT)
			A(ERROR_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY)
			A(ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE)
			A(ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME)
			A(ERROR_SXS_IDENTITY_DUPLICATE_ATTRIBUTE)
			A(ERROR_SXS_IDENTITY_PARSE_ERROR)
			A(ERROR_MALFORMED_SUBSTITUTION_STRING)
			A(ERROR_SXS_INCORRECT_PUBLIC_KEY_TOKEN)
			A(ERROR_UNMAPPED_SUBSTITUTION_STRING)
			A(ERROR_SXS_ASSEMBLY_NOT_LOCKED)
			A(ERROR_SXS_COMPONENT_STORE_CORRUPT)
			A(ERROR_ADVANCED_INSTALLER_FAILED)
			A(ERROR_XML_ENCODING_MISMATCH)
			A(ERROR_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT)
			A(ERROR_SXS_IDENTITIES_DIFFERENT)
			A(ERROR_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT)
			A(ERROR_SXS_FILE_NOT_PART_OF_ASSEMBLY)
			A(ERROR_SXS_MANIFEST_TOO_BIG)
			A(ERROR_SXS_SETTING_NOT_REGISTERED)
			A(ERROR_SXS_TRANSACTION_CLOSURE_INCOMPLETE)
			A(ERROR_SMI_PRIMITIVE_INSTALLER_FAILED)
			A(ERROR_GENERIC_COMMAND_FAILED)
			A(ERROR_SXS_FILE_HASH_MISSING)
			A(ERROR_EVT_INVALID_CHANNEL_PATH)
			A(ERROR_EVT_INVALID_QUERY)
			A(ERROR_EVT_PUBLISHER_METADATA_NOT_FOUND)
			A(ERROR_EVT_EVENT_TEMPLATE_NOT_FOUND)
			A(ERROR_EVT_INVALID_PUBLISHER_NAME)
			A(ERROR_EVT_INVALID_EVENT_DATA)
			A(ERROR_EVT_CHANNEL_NOT_FOUND)
			A(ERROR_EVT_MALFORMED_XML_TEXT)
			A(ERROR_EVT_SUBSCRIPTION_TO_DIRECT_CHANNEL)
			A(ERROR_EVT_CONFIGURATION_ERROR)
			A(ERROR_EVT_QUERY_RESULT_STALE)
			A(ERROR_EVT_QUERY_RESULT_INVALID_POSITION)
			A(ERROR_EVT_NON_VALIDATING_MSXML)
			A(ERROR_EVT_FILTER_ALREADYSCOPED)
			A(ERROR_EVT_FILTER_NOTELTSET)
			A(ERROR_EVT_FILTER_INVARG)
			A(ERROR_EVT_FILTER_INVTEST)
			A(ERROR_EVT_FILTER_INVTYPE)
			A(ERROR_EVT_FILTER_PARSEERR)
			A(ERROR_EVT_FILTER_UNSUPPORTEDOP)
			A(ERROR_EVT_FILTER_UNEXPECTEDTOKEN)
			A(ERROR_EVT_INVALID_OPERATION_OVER_ENABLED_DIRECT_CHANNEL)
			A(ERROR_EVT_INVALID_CHANNEL_PROPERTY_VALUE)
			A(ERROR_EVT_INVALID_PUBLISHER_PROPERTY_VALUE)
			A(ERROR_EVT_CHANNEL_CANNOT_ACTIVATE)
			A(ERROR_EVT_FILTER_TOO_COMPLEX)
			A(ERROR_EVT_MESSAGE_NOT_FOUND)
			A(ERROR_EVT_MESSAGE_ID_NOT_FOUND)
			A(ERROR_EVT_UNRESOLVED_VALUE_INSERT)
			A(ERROR_EVT_UNRESOLVED_PARAMETER_INSERT)
			A(ERROR_EVT_MAX_INSERTS_REACHED)
			A(ERROR_EVT_EVENT_DEFINITION_NOT_FOUND)
			A(ERROR_EVT_MESSAGE_LOCALE_NOT_FOUND)
			A(ERROR_EVT_VERSION_TOO_OLD)
			A(ERROR_EVT_VERSION_TOO_NEW)
			A(ERROR_EVT_CANNOT_OPEN_CHANNEL_OF_QUERY)
			A(ERROR_EVT_PUBLISHER_DISABLED)
			A(ERROR_EVT_FILTER_OUT_OF_RANGE)
			A(ERROR_EC_SUBSCRIPTION_CANNOT_ACTIVATE)
			A(ERROR_EC_LOG_DISABLED)
			A(ERROR_EC_CIRCULAR_FORWARDING)
			A(ERROR_EC_CREDSTORE_FULL)
			A(ERROR_EC_CRED_NOT_FOUND)
			A(ERROR_EC_NO_ACTIVE_CHANNEL)
			A(ERROR_MUI_FILE_NOT_FOUND)
			A(ERROR_MUI_INVALID_FILE)
			A(ERROR_MUI_INVALID_RC_CONFIG)
			A(ERROR_MUI_INVALID_LOCALE_NAME)
			A(ERROR_MUI_INVALID_ULTIMATEFALLBACK_NAME)
			A(ERROR_MUI_FILE_NOT_LOADED)
			A(ERROR_RESOURCE_ENUM_USER_STOP)
			A(ERROR_MUI_INTLSETTINGS_UILANG_NOT_INSTALLED)
			A(ERROR_MUI_INTLSETTINGS_INVALID_LOCALE_NAME)
			// Start of Monitor Configuration API error codes
			// 15200 to 15249
			A(ERROR_MCA_INVALID_CAPABILITIES_STRING)
			A(ERROR_MCA_INVALID_VCP_VERSION)
			A(ERROR_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION)
			A(ERROR_MCA_MCCS_VERSION_MISMATCH)
			A(ERROR_MCA_UNSUPPORTED_MCCS_VERSION)
			A(ERROR_MCA_INTERNAL_ERROR)
			A(ERROR_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED)
			A(ERROR_MCA_UNSUPPORTED_COLOR_TEMPERATURE)
			// Start of Syspart error codes
			// 15250 - 15299
			A(ERROR_AMBIGUOUS_SYSTEM_DEVICE)
			A(ERROR_SYSTEM_DEVICE_NOT_FOUND)
			// Start of Vortex error codes
			// 15300 - 15320
			A(ERROR_HASH_NOT_SUPPORTED)
			A(ERROR_HASH_NOT_PRESENT)
			#undef A
		};
		for(SIZE_T nIndex = 0; nIndex < DIM(g_pMap); nIndex++)
			if(g_pMap[nIndex].nValue == nValue)
				return CString(g_pMap[nIndex].pszName);
		return _T("");
	}
	static CString LookupHresultSystemIdentifier(HRESULT nValue)
	{
		//if(HRESULT_SEVERITY(nValue) != SEVERITY_ERROR)
		//	return _T("");
		if((UINT) HRESULT_FACILITY(nValue) >= 0x40)
			return _T("");
		static const struct { HRESULT nValue; LPCSTR pszName; } g_pMap[] = 
		{
			#define A(x) { x, #x },
			A(E_UNEXPECTED) // 0x8000FFFFL
			A(E_NOTIMPL) // 0x80004001L
			A(E_OUTOFMEMORY) // 0x8007000EL
			A(E_INVALIDARG) // 0x80070057L
			A(E_NOINTERFACE) // 0x80004002L
			A(E_POINTER) // 0x80004003L
			A(E_HANDLE) // 0x80070006L
			A(E_ABORT) // 0x80004004L
			A(E_FAIL) // 0x80004005L
			A(E_ACCESSDENIED) // 0x80070005L
			A(E_NOTIMPL) // 0x80000001L
			A(E_OUTOFMEMORY) // 0x80000002L
			A(E_INVALIDARG) // 0x80000003L
			A(E_NOINTERFACE) // 0x80000004L
			A(E_POINTER) // 0x80000005L
			A(E_HANDLE) // 0x80000006L
			A(E_ABORT) // 0x80000007L
			A(E_FAIL) // 0x80000008L
			A(E_ACCESSDENIED) // 0x80000009L
			A(E_PENDING) // 0x8000000AL
			A(CO_E_INIT_TLS) // 0x80004006L
			A(CO_E_INIT_SHARED_ALLOCATOR) // 0x80004007L
			A(CO_E_INIT_MEMORY_ALLOCATOR) // 0x80004008L
			A(CO_E_INIT_CLASS_CACHE) // 0x80004009L
			A(CO_E_INIT_RPC_CHANNEL) // 0x8000400AL
			A(CO_E_INIT_TLS_SET_CHANNEL_CONTROL) // 0x8000400BL
			A(CO_E_INIT_TLS_CHANNEL_CONTROL) // 0x8000400CL
			A(CO_E_INIT_UNACCEPTED_USER_ALLOCATOR) // 0x8000400DL
			A(CO_E_INIT_SCM_MUTEX_EXISTS) // 0x8000400EL
			A(CO_E_INIT_SCM_FILE_MAPPING_EXISTS) // 0x8000400FL
			A(CO_E_INIT_SCM_MAP_VIEW_OF_FILE) // 0x80004010L
			A(CO_E_INIT_SCM_EXEC_FAILURE) // 0x80004011L
			A(CO_E_INIT_ONLY_SINGLE_THREADED) // 0x80004012L
			A(CO_E_CANT_REMOTE) // 0x80004013L
			A(CO_E_BAD_SERVER_NAME) // 0x80004014L
			A(CO_E_WRONG_SERVER_IDENTITY) // 0x80004015L
			A(CO_E_OLE1DDE_DISABLED) // 0x80004016L
			A(CO_E_RUNAS_SYNTAX) // 0x80004017L
			A(CO_E_CREATEPROCESS_FAILURE) // 0x80004018L
			A(CO_E_RUNAS_CREATEPROCESS_FAILURE) // 0x80004019L
			A(CO_E_RUNAS_LOGON_FAILURE) // 0x8000401AL
			A(CO_E_LAUNCH_PERMSSION_DENIED) // 0x8000401BL
			A(CO_E_START_SERVICE_FAILURE) // 0x8000401CL
			A(CO_E_REMOTE_COMMUNICATION_FAILURE) // 0x8000401DL
			A(CO_E_SERVER_START_TIMEOUT) // 0x8000401EL
			A(CO_E_CLSREG_INCONSISTENT) // 0x8000401FL
			A(CO_E_IIDREG_INCONSISTENT) // 0x80004020L
			A(CO_E_NOT_SUPPORTED) // 0x80004021L
			A(CO_E_RELOAD_DLL) // 0x80004022L
			A(CO_E_MSI_ERROR) // 0x80004023L
			A(CO_E_ATTEMPT_TO_CREATE_OUTSIDE_CLIENT_CONTEXT) // 0x80004024L
			A(CO_E_SERVER_PAUSED) // 0x80004025L
			A(CO_E_SERVER_NOT_PAUSED) // 0x80004026L
			A(CO_E_CLASS_DISABLED) // 0x80004027L
			A(CO_E_CLRNOTAVAILABLE) // 0x80004028L
			A(CO_E_ASYNC_WORK_REJECTED) // 0x80004029L
			A(CO_E_SERVER_INIT_TIMEOUT) // 0x8000402AL
			A(CO_E_NO_SECCTX_IN_ACTIVATE) // 0x8000402BL
			A(CO_E_TRACKER_CONFIG) // 0x80004030L
			A(CO_E_THREADPOOL_CONFIG) // 0x80004031L
			A(CO_E_SXS_CONFIG) // 0x80004032L
			A(CO_E_MALFORMED_SPN) // 0x80004033L
			A(OLE_E_OLEVERB) // 0x80040000L
			A(OLE_E_ADVF) // 0x80040001L
			A(OLE_E_ENUM_NOMORE) // 0x80040002L
			A(OLE_E_ADVISENOTSUPPORTED) // 0x80040003L
			A(OLE_E_NOCONNECTION) // 0x80040004L
			A(OLE_E_NOTRUNNING) // 0x80040005L
			A(OLE_E_NOCACHE) // 0x80040006L
			A(OLE_E_BLANK) // 0x80040007L
			A(OLE_E_CLASSDIFF) // 0x80040008L
			A(OLE_E_CANT_GETMONIKER) // 0x80040009L
			A(OLE_E_CANT_BINDTOSOURCE) // 0x8004000AL
			A(OLE_E_STATIC) // 0x8004000BL
			A(OLE_E_PROMPTSAVECANCELLED) // 0x8004000CL
			A(OLE_E_INVALIDRECT) // 0x8004000DL
			A(OLE_E_WRONGCOMPOBJ) // 0x8004000EL
			A(OLE_E_INVALIDHWND) // 0x8004000FL
			A(OLE_E_NOT_INPLACEACTIVE) // 0x80040010L
			A(OLE_E_CANTCONVERT) // 0x80040011L
			A(OLE_E_NOSTORAGE) // 0x80040012L
			A(DV_E_FORMATETC) // 0x80040064L
			A(DV_E_DVTARGETDEVICE) // 0x80040065L
			A(DV_E_STGMEDIUM) // 0x80040066L
			A(DV_E_STATDATA) // 0x80040067L
			A(DV_E_LINDEX) // 0x80040068L
			A(DV_E_TYMED) // 0x80040069L
			A(DV_E_CLIPFORMAT) // 0x8004006AL
			A(DV_E_DVASPECT) // 0x8004006BL
			A(DV_E_DVTARGETDEVICE_SIZE) // 0x8004006CL
			A(DV_E_NOIVIEWOBJECT) // 0x8004006DL
			A(DRAGDROP_E_NOTREGISTERED) // 0x80040100L
			A(DRAGDROP_E_ALREADYREGISTERED) // 0x80040101L
			A(DRAGDROP_E_INVALIDHWND) // 0x80040102L
			A(CLASS_E_NOAGGREGATION) // 0x80040110L
			A(CLASS_E_CLASSNOTAVAILABLE) // 0x80040111L
			A(CLASS_E_NOTLICENSED) // 0x80040112L
			A(VIEW_E_DRAW) // 0x80040140L
			A(REGDB_E_READREGDB) // 0x80040150L
			A(REGDB_E_WRITEREGDB) // 0x80040151L
			A(REGDB_E_KEYMISSING) // 0x80040152L
			A(REGDB_E_INVALIDVALUE) // 0x80040153L
			A(REGDB_E_CLASSNOTREG) // 0x80040154L
			A(REGDB_E_IIDNOTREG) // 0x80040155L
			A(REGDB_E_BADTHREADINGMODEL) // 0x80040156L
			A(CAT_E_CATIDNOEXIST) // 0x80040160L
			A(CAT_E_NODESCRIPTION) // 0x80040161L
			A(CS_E_PACKAGE_NOTFOUND) // 0x80040164L
			A(CS_E_NOT_DELETABLE) // 0x80040165L
			A(CS_E_CLASS_NOTFOUND) // 0x80040166L
			A(CS_E_INVALID_VERSION) // 0x80040167L
			A(CS_E_NO_CLASSSTORE) // 0x80040168L
			A(CS_E_OBJECT_NOTFOUND) // 0x80040169L
			A(CS_E_OBJECT_ALREADY_EXISTS) // 0x8004016AL
			A(CS_E_INVALID_PATH) // 0x8004016BL
			A(CS_E_NETWORK_ERROR) // 0x8004016CL
			A(CS_E_ADMIN_LIMIT_EXCEEDED) // 0x8004016DL
			A(CS_E_SCHEMA_MISMATCH) // 0x8004016EL
			A(CS_E_INTERNAL_ERROR) // 0x8004016FL
			A(CACHE_E_NOCACHE_UPDATED) // 0x80040170L
			A(OLEOBJ_E_NOVERBS) // 0x80040180L
			A(OLEOBJ_E_INVALIDVERB) // 0x80040181L
			A(INPLACE_E_NOTUNDOABLE) // 0x800401A0L
			A(INPLACE_E_NOTOOLSPACE) // 0x800401A1L
			A(CONVERT10_E_OLESTREAM_GET) // 0x800401C0L
			A(CONVERT10_E_OLESTREAM_PUT) // 0x800401C1L
			A(CONVERT10_E_OLESTREAM_FMT) // 0x800401C2L
			A(CONVERT10_E_OLESTREAM_BITMAP_TO_DIB) // 0x800401C3L
			A(CONVERT10_E_STG_FMT) // 0x800401C4L
			A(CONVERT10_E_STG_NO_STD_STREAM) // 0x800401C5L
			A(CONVERT10_E_STG_DIB_TO_BITMAP) // 0x800401C6L
			A(CLIPBRD_E_CANT_OPEN) // 0x800401D0L
			A(CLIPBRD_E_CANT_EMPTY) // 0x800401D1L
			A(CLIPBRD_E_CANT_SET) // 0x800401D2L
			A(CLIPBRD_E_BAD_DATA) // 0x800401D3L
			A(CLIPBRD_E_CANT_CLOSE) // 0x800401D4L
			A(MK_E_CONNECTMANUALLY) // 0x800401E0L
			A(MK_E_EXCEEDEDDEADLINE) // 0x800401E1L
			A(MK_E_NEEDGENERIC) // 0x800401E2L
			A(MK_E_UNAVAILABLE) // 0x800401E3L
			A(MK_E_SYNTAX) // 0x800401E4L
			A(MK_E_NOOBJECT) // 0x800401E5L
			A(MK_E_INVALIDEXTENSION) // 0x800401E6L
			A(MK_E_INTERMEDIATEINTERFACENOTSUPPORTED) // 0x800401E7L
			A(MK_E_NOTBINDABLE) // 0x800401E8L
			A(MK_E_NOTBOUND) // 0x800401E9L
			A(MK_E_CANTOPENFILE) // 0x800401EAL
			A(MK_E_MUSTBOTHERUSER) // 0x800401EBL
			A(MK_E_NOINVERSE) // 0x800401ECL
			A(MK_E_NOSTORAGE) // 0x800401EDL
			A(MK_E_NOPREFIX) // 0x800401EEL
			A(MK_E_ENUMERATION_FAILED) // 0x800401EFL
			A(CO_E_NOTINITIALIZED) // 0x800401F0L
			A(CO_E_ALREADYINITIALIZED) // 0x800401F1L
			A(CO_E_CANTDETERMINECLASS) // 0x800401F2L
			A(CO_E_CLASSSTRING) // 0x800401F3L
			A(CO_E_IIDSTRING) // 0x800401F4L
			A(CO_E_APPNOTFOUND) // 0x800401F5L
			A(CO_E_APPSINGLEUSE) // 0x800401F6L
			A(CO_E_ERRORINAPP) // 0x800401F7L
			A(CO_E_DLLNOTFOUND) // 0x800401F8L
			A(CO_E_ERRORINDLL) // 0x800401F9L
			A(CO_E_WRONGOSFORAPP) // 0x800401FAL
			A(CO_E_OBJNOTREG) // 0x800401FBL
			A(CO_E_OBJISREG) // 0x800401FCL
			A(CO_E_OBJNOTCONNECTED) // 0x800401FDL
			A(CO_E_APPDIDNTREG) // 0x800401FEL
			A(CO_E_RELEASED) // 0x800401FFL
			A(EVENT_S_SOME_SUBSCRIBERS_FAILED) // 0x00040200L
			A(EVENT_E_ALL_SUBSCRIBERS_FAILED) // 0x80040201L
			A(EVENT_S_NOSUBSCRIBERS) // 0x00040202L
			A(EVENT_E_QUERYSYNTAX) // 0x80040203L
			A(EVENT_E_QUERYFIELD) // 0x80040204L
			A(EVENT_E_INTERNALEXCEPTION) // 0x80040205L
			A(EVENT_E_INTERNALERROR) // 0x80040206L
			A(EVENT_E_INVALID_PER_USER_SID) // 0x80040207L
			A(EVENT_E_USER_EXCEPTION) // 0x80040208L
			A(EVENT_E_TOO_MANY_METHODS) // 0x80040209L
			A(EVENT_E_MISSING_EVENTCLASS) // 0x8004020AL
			A(EVENT_E_NOT_ALL_REMOVED) // 0x8004020BL
			A(EVENT_E_COMPLUS_NOT_INSTALLED) // 0x8004020CL
			A(EVENT_E_CANT_MODIFY_OR_DELETE_UNCONFIGURED_OBJECT) // 0x8004020DL
			A(EVENT_E_CANT_MODIFY_OR_DELETE_CONFIGURED_OBJECT) // 0x8004020EL
			A(EVENT_E_INVALID_EVENT_CLASS_PARTITION) // 0x8004020FL
			A(EVENT_E_PER_USER_SID_NOT_LOGGED_ON) // 0x80040210L
			A(XACT_E_ALREADYOTHERSINGLEPHASE) // 0x8004D000L
			A(XACT_E_CANTRETAIN) // 0x8004D001L
			A(XACT_E_COMMITFAILED) // 0x8004D002L
			A(XACT_E_COMMITPREVENTED) // 0x8004D003L
			A(XACT_E_HEURISTICABORT) // 0x8004D004L
			A(XACT_E_HEURISTICCOMMIT) // 0x8004D005L
			A(XACT_E_HEURISTICDAMAGE) // 0x8004D006L
			A(XACT_E_HEURISTICDANGER) // 0x8004D007L
			A(XACT_E_ISOLATIONLEVEL) // 0x8004D008L
			A(XACT_E_NOASYNC) // 0x8004D009L
			A(XACT_E_NOENLIST) // 0x8004D00AL
			A(XACT_E_NOISORETAIN) // 0x8004D00BL
			A(XACT_E_NORESOURCE) // 0x8004D00CL
			A(XACT_E_NOTCURRENT) // 0x8004D00DL
			A(XACT_E_NOTRANSACTION) // 0x8004D00EL
			A(XACT_E_NOTSUPPORTED) // 0x8004D00FL
			A(XACT_E_UNKNOWNRMGRID) // 0x8004D010L
			A(XACT_E_WRONGSTATE) // 0x8004D011L
			A(XACT_E_WRONGUOW) // 0x8004D012L
			A(XACT_E_XTIONEXISTS) // 0x8004D013L
			A(XACT_E_NOIMPORTOBJECT) // 0x8004D014L
			A(XACT_E_INVALIDCOOKIE) // 0x8004D015L
			A(XACT_E_INDOUBT) // 0x8004D016L
			A(XACT_E_NOTIMEOUT) // 0x8004D017L
			A(XACT_E_ALREADYINPROGRESS) // 0x8004D018L
			A(XACT_E_ABORTED) // 0x8004D019L
			A(XACT_E_LOGFULL) // 0x8004D01AL
			A(XACT_E_TMNOTAVAILABLE) // 0x8004D01BL
			A(XACT_E_CONNECTION_DOWN) // 0x8004D01CL
			A(XACT_E_CONNECTION_DENIED) // 0x8004D01DL
			A(XACT_E_REENLISTTIMEOUT) // 0x8004D01EL
			A(XACT_E_TIP_CONNECT_FAILED) // 0x8004D01FL
			A(XACT_E_TIP_PROTOCOL_ERROR) // 0x8004D020L
			A(XACT_E_TIP_PULL_FAILED) // 0x8004D021L
			A(XACT_E_DEST_TMNOTAVAILABLE) // 0x8004D022L
			A(XACT_E_TIP_DISABLED) // 0x8004D023L
			A(XACT_E_NETWORK_TX_DISABLED) // 0x8004D024L
			A(XACT_E_PARTNER_NETWORK_TX_DISABLED) // 0x8004D025L
			A(XACT_E_XA_TX_DISABLED) // 0x8004D026L
			A(XACT_E_UNABLE_TO_READ_DTC_CONFIG) // 0x8004D027L
			A(XACT_E_UNABLE_TO_LOAD_DTC_PROXY) // 0x8004D028L
			A(XACT_E_ABORTING) // 0x8004D029L
			A(XACT_E_PUSH_COMM_FAILURE) // 0x8004D02AL
			A(XACT_E_PULL_COMM_FAILURE) // 0x8004D02BL
			A(XACT_E_LU_TX_DISABLED) // 0x8004D02CL
			A(XACT_E_CLERKNOTFOUND) // 0x8004D080L
			A(XACT_E_CLERKEXISTS) // 0x8004D081L
			A(XACT_E_RECOVERYINPROGRESS) // 0x8004D082L
			A(XACT_E_TRANSACTIONCLOSED) // 0x8004D083L
			A(XACT_E_INVALIDLSN) // 0x8004D084L
			A(XACT_E_REPLAYREQUEST) // 0x8004D085L
			A(XACT_S_ASYNC) // 0x0004D000L
			A(XACT_S_DEFECT) // 0x0004D001L
			A(XACT_S_READONLY) // 0x0004D002L
			A(XACT_S_SOMENORETAIN) // 0x0004D003L
			A(XACT_S_OKINFORM) // 0x0004D004L
			A(XACT_S_MADECHANGESCONTENT) // 0x0004D005L
			A(XACT_S_MADECHANGESINFORM) // 0x0004D006L
			A(XACT_S_ALLNORETAIN) // 0x0004D007L
			A(XACT_S_ABORTING) // 0x0004D008L
			A(XACT_S_SINGLEPHASE) // 0x0004D009L
			A(XACT_S_LOCALLY_OK) // 0x0004D00AL
			A(XACT_S_LASTRESOURCEMANAGER) // 0x0004D010L
			A(CONTEXT_E_ABORTED) // 0x8004E002L
			A(CONTEXT_E_ABORTING) // 0x8004E003L
			A(CONTEXT_E_NOCONTEXT) // 0x8004E004L
			A(CONTEXT_E_WOULD_DEADLOCK) // 0x8004E005L
			A(CONTEXT_E_SYNCH_TIMEOUT) // 0x8004E006L
			A(CONTEXT_E_OLDREF) // 0x8004E007L
			A(CONTEXT_E_ROLENOTFOUND) // 0x8004E00CL
			A(CONTEXT_E_TMNOTAVAILABLE) // 0x8004E00FL
			A(CO_E_ACTIVATIONFAILED) // 0x8004E021L
			A(CO_E_ACTIVATIONFAILED_EVENTLOGGED) // 0x8004E022L
			A(CO_E_ACTIVATIONFAILED_CATALOGERROR) // 0x8004E023L
			A(CO_E_ACTIVATIONFAILED_TIMEOUT) // 0x8004E024L
			A(CO_E_INITIALIZATIONFAILED) // 0x8004E025L
			A(CONTEXT_E_NOJIT) // 0x8004E026L
			A(CONTEXT_E_NOTRANSACTION) // 0x8004E027L
			A(CO_E_THREADINGMODEL_CHANGED) // 0x8004E028L
			A(CO_E_NOIISINTRINSICS) // 0x8004E029L
			A(CO_E_NOCOOKIES) // 0x8004E02AL
			A(CO_E_DBERROR) // 0x8004E02BL
			A(CO_E_NOTPOOLED) // 0x8004E02CL
			A(CO_E_NOTCONSTRUCTED) // 0x8004E02DL
			A(CO_E_NOSYNCHRONIZATION) // 0x8004E02EL
			A(CO_E_ISOLEVELMISMATCH) // 0x8004E02FL
			A(CO_E_CALL_OUT_OF_TX_SCOPE_NOT_ALLOWED) // 0x8004E030L
			A(CO_E_EXIT_TRANSACTION_SCOPE_NOT_CALLED) // 0x8004E031L
			A(OLE_S_USEREG) // 0x00040000L
			A(OLE_S_STATIC) // 0x00040001L
			A(OLE_S_MAC_CLIPFORMAT) // 0x00040002L
			A(DRAGDROP_S_DROP) // 0x00040100L
			A(DRAGDROP_S_CANCEL) // 0x00040101L
			A(DRAGDROP_S_USEDEFAULTCURSORS) // 0x00040102L
			A(DATA_S_SAMEFORMATETC) // 0x00040130L
			A(VIEW_S_ALREADY_FROZEN) // 0x00040140L
			A(CACHE_S_FORMATETC_NOTSUPPORTED) // 0x00040170L
			A(CACHE_S_SAMECACHE) // 0x00040171L
			A(CACHE_S_SOMECACHES_NOTUPDATED) // 0x00040172L
			A(OLEOBJ_S_INVALIDVERB) // 0x00040180L
			A(OLEOBJ_S_CANNOT_DOVERB_NOW) // 0x00040181L
			A(OLEOBJ_S_INVALIDHWND) // 0x00040182L
			A(INPLACE_S_TRUNCATED) // 0x000401A0L
			A(CONVERT10_S_NO_PRESENTATION) // 0x000401C0L
			A(MK_S_REDUCED_TO_SELF) // 0x000401E2L
			A(MK_S_ME) // 0x000401E4L
			A(MK_S_HIM) // 0x000401E5L
			A(MK_S_US) // 0x000401E6L
			A(MK_S_MONIKERALREADYREGISTERED) // 0x000401E7L
			A(SCHED_S_TASK_READY) // 0x00041300L
			A(SCHED_S_TASK_RUNNING) // 0x00041301L
			A(SCHED_S_TASK_DISABLED) // 0x00041302L
			A(SCHED_S_TASK_HAS_NOT_RUN) // 0x00041303L
			A(SCHED_S_TASK_NO_MORE_RUNS) // 0x00041304L
			A(SCHED_S_TASK_NOT_SCHEDULED) // 0x00041305L
			A(SCHED_S_TASK_TERMINATED) // 0x00041306L
			A(SCHED_S_TASK_NO_VALID_TRIGGERS) // 0x00041307L
			A(SCHED_S_EVENT_TRIGGER) // 0x00041308L
			A(SCHED_E_TRIGGER_NOT_FOUND) // 0x80041309L
			A(SCHED_E_TASK_NOT_READY) // 0x8004130AL
			A(SCHED_E_TASK_NOT_RUNNING) // 0x8004130BL
			A(SCHED_E_SERVICE_NOT_INSTALLED) // 0x8004130CL
			A(SCHED_E_CANNOT_OPEN_TASK) // 0x8004130DL
			A(SCHED_E_INVALID_TASK) // 0x8004130EL
			A(SCHED_E_ACCOUNT_INFORMATION_NOT_SET) // 0x8004130FL
			A(SCHED_E_ACCOUNT_NAME_NOT_FOUND) // 0x80041310L
			A(SCHED_E_ACCOUNT_DBASE_CORRUPT) // 0x80041311L
			A(SCHED_E_NO_SECURITY_SERVICES) // 0x80041312L
			A(SCHED_E_UNKNOWN_OBJECT_VERSION) // 0x80041313L
			A(SCHED_E_UNSUPPORTED_ACCOUNT_OPTION) // 0x80041314L
			A(SCHED_E_SERVICE_NOT_RUNNING) // 0x80041315L
			A(SCHED_E_UNEXPECTEDNODE) // 0x80041316L
			A(SCHED_E_NAMESPACE) // 0x80041317L
			A(SCHED_E_INVALIDVALUE) // 0x80041318L
			A(SCHED_E_MISSINGNODE) // 0x80041319L
			A(SCHED_E_MALFORMEDXML) // 0x8004131AL
			A(SCHED_S_SOME_TRIGGERS_FAILED) // 0x0004131BL
			A(SCHED_S_BATCH_LOGON_PROBLEM) // 0x0004131CL
			A(SCHED_E_TOO_MANY_NODES) // 0x8004131DL
			A(SCHED_E_PAST_END_BOUNDARY) // 0x8004131EL
			A(SCHED_E_ALREADY_RUNNING) // 0x8004131FL
			A(SCHED_E_USER_NOT_LOGGED_ON) // 0x80041320L
			A(SCHED_E_INVALID_TASK_HASH) // 0x80041321L
			A(SCHED_E_SERVICE_NOT_AVAILABLE) // 0x80041322L
			A(SCHED_E_SERVICE_TOO_BUSY) // 0x80041323L
			A(SCHED_E_TASK_ATTEMPTED) // 0x80041324L
			A(SCHED_S_TASK_QUEUED) // 0x00041325L
			A(SCHED_E_TASK_DISABLED) // 0x80041326L
			A(SCHED_E_TASK_NOT_V1_COMPAT) // 0x80041327L
			A(SCHED_E_START_ON_DEMAND) // 0x80041328L
			A(CO_E_CLASS_CREATE_FAILED) // 0x80080001L
			A(CO_E_SCM_ERROR) // 0x80080002L
			A(CO_E_SCM_RPC_FAILURE) // 0x80080003L
			A(CO_E_BAD_PATH) // 0x80080004L
			A(CO_E_SERVER_EXEC_FAILURE) // 0x80080005L
			A(CO_E_OBJSRV_RPC_FAILURE) // 0x80080006L
			A(MK_E_NO_NORMALIZED) // 0x80080007L
			A(CO_E_SERVER_STOPPING) // 0x80080008L
			A(MEM_E_INVALID_ROOT) // 0x80080009L
			A(MEM_E_INVALID_LINK) // 0x80080010L
			A(MEM_E_INVALID_SIZE) // 0x80080011L
			A(CO_S_NOTALLINTERFACES) // 0x00080012L
			A(CO_S_MACHINENAMENOTFOUND) // 0x00080013L
			A(CO_E_MISSING_DISPLAYNAME) // 0x80080015L
			A(CO_E_RUNAS_VALUE_MUST_BE_AAA) // 0x80080016L
			A(CO_E_ELEVATION_DISABLED) // 0x80080017L
			A(DISP_E_UNKNOWNINTERFACE) // 0x80020001L
			A(DISP_E_MEMBERNOTFOUND) // 0x80020003L
			A(DISP_E_PARAMNOTFOUND) // 0x80020004L
			A(DISP_E_TYPEMISMATCH) // 0x80020005L
			A(DISP_E_UNKNOWNNAME) // 0x80020006L
			A(DISP_E_NONAMEDARGS) // 0x80020007L
			A(DISP_E_BADVARTYPE) // 0x80020008L
			A(DISP_E_EXCEPTION) // 0x80020009L
			A(DISP_E_OVERFLOW) // 0x8002000AL
			A(DISP_E_BADINDEX) // 0x8002000BL
			A(DISP_E_UNKNOWNLCID) // 0x8002000CL
			A(DISP_E_ARRAYISLOCKED) // 0x8002000DL
			A(DISP_E_BADPARAMCOUNT) // 0x8002000EL
			A(DISP_E_PARAMNOTOPTIONAL) // 0x8002000FL
			A(DISP_E_BADCALLEE) // 0x80020010L
			A(DISP_E_NOTACOLLECTION) // 0x80020011L
			A(DISP_E_DIVBYZERO) // 0x80020012L
			A(DISP_E_BUFFERTOOSMALL) // 0x80020013L
			A(TYPE_E_BUFFERTOOSMALL) // 0x80028016L
			A(TYPE_E_FIELDNOTFOUND) // 0x80028017L
			A(TYPE_E_INVDATAREAD) // 0x80028018L
			A(TYPE_E_UNSUPFORMAT) // 0x80028019L
			A(TYPE_E_REGISTRYACCESS) // 0x8002801CL
			A(TYPE_E_LIBNOTREGISTERED) // 0x8002801DL
			A(TYPE_E_UNDEFINEDTYPE) // 0x80028027L
			A(TYPE_E_QUALIFIEDNAMEDISALLOWED) // 0x80028028L
			A(TYPE_E_INVALIDSTATE) // 0x80028029L
			A(TYPE_E_WRONGTYPEKIND) // 0x8002802AL
			A(TYPE_E_ELEMENTNOTFOUND) // 0x8002802BL
			A(TYPE_E_AMBIGUOUSNAME) // 0x8002802CL
			A(TYPE_E_NAMECONFLICT) // 0x8002802DL
			A(TYPE_E_UNKNOWNLCID) // 0x8002802EL
			A(TYPE_E_DLLFUNCTIONNOTFOUND) // 0x8002802FL
			A(TYPE_E_BADMODULEKIND) // 0x800288BDL
			A(TYPE_E_SIZETOOBIG) // 0x800288C5L
			A(TYPE_E_DUPLICATEID) // 0x800288C6L
			A(TYPE_E_INVALIDID) // 0x800288CFL
			A(TYPE_E_TYPEMISMATCH) // 0x80028CA0L
			A(TYPE_E_OUTOFBOUNDS) // 0x80028CA1L
			A(TYPE_E_IOERROR) // 0x80028CA2L
			A(TYPE_E_CANTCREATETMPFILE) // 0x80028CA3L
			A(TYPE_E_CANTLOADLIBRARY) // 0x80029C4AL
			A(TYPE_E_INCONSISTENTPROPFUNCS) // 0x80029C83L
			A(TYPE_E_CIRCULARTYPE) // 0x80029C84L
			A(STG_E_INVALIDFUNCTION) // 0x80030001L
			A(STG_E_FILENOTFOUND) // 0x80030002L
			A(STG_E_PATHNOTFOUND) // 0x80030003L
			A(STG_E_TOOMANYOPENFILES) // 0x80030004L
			A(STG_E_ACCESSDENIED) // 0x80030005L
			A(STG_E_INVALIDHANDLE) // 0x80030006L
			A(STG_E_INSUFFICIENTMEMORY) // 0x80030008L
			A(STG_E_INVALIDPOINTER) // 0x80030009L
			A(STG_E_NOMOREFILES) // 0x80030012L
			A(STG_E_DISKISWRITEPROTECTED) // 0x80030013L
			A(STG_E_SEEKERROR) // 0x80030019L
			A(STG_E_WRITEFAULT) // 0x8003001DL
			A(STG_E_READFAULT) // 0x8003001EL
			A(STG_E_SHAREVIOLATION) // 0x80030020L
			A(STG_E_LOCKVIOLATION) // 0x80030021L
			A(STG_E_FILEALREADYEXISTS) // 0x80030050L
			A(STG_E_INVALIDPARAMETER) // 0x80030057L
			A(STG_E_MEDIUMFULL) // 0x80030070L
			A(STG_E_PROPSETMISMATCHED) // 0x800300F0L
			A(STG_E_ABNORMALAPIEXIT) // 0x800300FAL
			A(STG_E_INVALIDHEADER) // 0x800300FBL
			A(STG_E_INVALIDNAME) // 0x800300FCL
			A(STG_E_UNKNOWN) // 0x800300FDL
			A(STG_E_UNIMPLEMENTEDFUNCTION) // 0x800300FEL
			A(STG_E_INVALIDFLAG) // 0x800300FFL
			A(STG_E_INUSE) // 0x80030100L
			A(STG_E_NOTCURRENT) // 0x80030101L
			A(STG_E_REVERTED) // 0x80030102L
			A(STG_E_CANTSAVE) // 0x80030103L
			A(STG_E_OLDFORMAT) // 0x80030104L
			A(STG_E_OLDDLL) // 0x80030105L
			A(STG_E_SHAREREQUIRED) // 0x80030106L
			A(STG_E_NOTFILEBASEDSTORAGE) // 0x80030107L
			A(STG_E_EXTANTMARSHALLINGS) // 0x80030108L
			A(STG_E_DOCFILECORRUPT) // 0x80030109L
			A(STG_E_BADBASEADDRESS) // 0x80030110L
			A(STG_E_DOCFILETOOLARGE) // 0x80030111L
			A(STG_E_NOTSIMPLEFORMAT) // 0x80030112L
			A(STG_E_INCOMPLETE) // 0x80030201L
			A(STG_E_TERMINATED) // 0x80030202L
			A(STG_S_CONVERTED) // 0x00030200L
			A(STG_S_BLOCK) // 0x00030201L
			A(STG_S_RETRYNOW) // 0x00030202L
			A(STG_S_MONITORING) // 0x00030203L
			A(STG_S_MULTIPLEOPENS) // 0x00030204L
			A(STG_S_CONSOLIDATIONFAILED) // 0x00030205L
			A(STG_S_CANNOTCONSOLIDATE) // 0x00030206L
			A(STG_E_STATUS_COPY_PROTECTION_FAILURE) // 0x80030305L
			A(STG_E_CSS_AUTHENTICATION_FAILURE) // 0x80030306L
			A(STG_E_CSS_KEY_NOT_PRESENT) // 0x80030307L
			A(STG_E_CSS_KEY_NOT_ESTABLISHED) // 0x80030308L
			A(STG_E_CSS_SCRAMBLED_SECTOR) // 0x80030309L
			A(STG_E_CSS_REGION_MISMATCH) // 0x8003030AL
			A(STG_E_RESETS_EXHAUSTED) // 0x8003030BL
			A(RPC_E_CALL_REJECTED) // 0x80010001L
			A(RPC_E_CALL_CANCELED) // 0x80010002L
			A(RPC_E_CANTPOST_INSENDCALL) // 0x80010003L
			A(RPC_E_CANTCALLOUT_INASYNCCALL) // 0x80010004L
			A(RPC_E_CANTCALLOUT_INEXTERNALCALL) // 0x80010005L
			A(RPC_E_CONNECTION_TERMINATED) // 0x80010006L
			A(RPC_E_SERVER_DIED) // 0x80010007L
			A(RPC_E_CLIENT_DIED) // 0x80010008L
			A(RPC_E_INVALID_DATAPACKET) // 0x80010009L
			A(RPC_E_CANTTRANSMIT_CALL) // 0x8001000AL
			A(RPC_E_CLIENT_CANTMARSHAL_DATA) // 0x8001000BL
			A(RPC_E_CLIENT_CANTUNMARSHAL_DATA) // 0x8001000CL
			A(RPC_E_SERVER_CANTMARSHAL_DATA) // 0x8001000DL
			A(RPC_E_SERVER_CANTUNMARSHAL_DATA) // 0x8001000EL
			A(RPC_E_INVALID_DATA) // 0x8001000FL
			A(RPC_E_INVALID_PARAMETER) // 0x80010010L
			A(RPC_E_CANTCALLOUT_AGAIN) // 0x80010011L
			A(RPC_E_SERVER_DIED_DNE) // 0x80010012L
			A(RPC_E_SYS_CALL_FAILED) // 0x80010100L
			A(RPC_E_OUT_OF_RESOURCES) // 0x80010101L
			A(RPC_E_ATTEMPTED_MULTITHREAD) // 0x80010102L
			A(RPC_E_NOT_REGISTERED) // 0x80010103L
			A(RPC_E_FAULT) // 0x80010104L
			A(RPC_E_SERVERFAULT) // 0x80010105L
			A(RPC_E_CHANGED_MODE) // 0x80010106L
			A(RPC_E_INVALIDMETHOD) // 0x80010107L
			A(RPC_E_DISCONNECTED) // 0x80010108L
			A(RPC_E_RETRY) // 0x80010109L
			A(RPC_E_SERVERCALL_RETRYLATER) // 0x8001010AL
			A(RPC_E_SERVERCALL_REJECTED) // 0x8001010BL
			A(RPC_E_INVALID_CALLDATA) // 0x8001010CL
			A(RPC_E_CANTCALLOUT_ININPUTSYNCCALL) // 0x8001010DL
			A(RPC_E_WRONG_THREAD) // 0x8001010EL
			A(RPC_E_THREAD_NOT_INIT) // 0x8001010FL
			A(RPC_E_VERSION_MISMATCH) // 0x80010110L
			A(RPC_E_INVALID_HEADER) // 0x80010111L
			A(RPC_E_INVALID_EXTENSION) // 0x80010112L
			A(RPC_E_INVALID_IPID) // 0x80010113L
			A(RPC_E_INVALID_OBJECT) // 0x80010114L
			A(RPC_S_CALLPENDING) // 0x80010115L
			A(RPC_S_WAITONTIMER) // 0x80010116L
			A(RPC_E_CALL_COMPLETE) // 0x80010117L
			A(RPC_E_UNSECURE_CALL) // 0x80010118L
			A(RPC_E_TOO_LATE) // 0x80010119L
			A(RPC_E_NO_GOOD_SECURITY_PACKAGES) // 0x8001011AL
			A(RPC_E_ACCESS_DENIED) // 0x8001011BL
			A(RPC_E_REMOTE_DISABLED) // 0x8001011CL
			A(RPC_E_INVALID_OBJREF) // 0x8001011DL
			A(RPC_E_NO_CONTEXT) // 0x8001011EL
			A(RPC_E_TIMEOUT) // 0x8001011FL
			A(RPC_E_NO_SYNC) // 0x80010120L
			A(RPC_E_FULLSIC_REQUIRED) // 0x80010121L
			A(RPC_E_INVALID_STD_NAME) // 0x80010122L
			A(CO_E_FAILEDTOIMPERSONATE) // 0x80010123L
			A(CO_E_FAILEDTOGETSECCTX) // 0x80010124L
			A(CO_E_FAILEDTOOPENTHREADTOKEN) // 0x80010125L
			A(CO_E_FAILEDTOGETTOKENINFO) // 0x80010126L
			A(CO_E_TRUSTEEDOESNTMATCHCLIENT) // 0x80010127L
			A(CO_E_FAILEDTOQUERYCLIENTBLANKET) // 0x80010128L
			A(CO_E_FAILEDTOSETDACL) // 0x80010129L
			A(CO_E_ACCESSCHECKFAILED) // 0x8001012AL
			A(CO_E_NETACCESSAPIFAILED) // 0x8001012BL
			A(CO_E_WRONGTRUSTEENAMESYNTAX) // 0x8001012CL
			A(CO_E_INVALIDSID) // 0x8001012DL
			A(CO_E_CONVERSIONFAILED) // 0x8001012EL
			A(CO_E_NOMATCHINGSIDFOUND) // 0x8001012FL
			A(CO_E_LOOKUPACCSIDFAILED) // 0x80010130L
			A(CO_E_NOMATCHINGNAMEFOUND) // 0x80010131L
			A(CO_E_LOOKUPACCNAMEFAILED) // 0x80010132L
			A(CO_E_SETSERLHNDLFAILED) // 0x80010133L
			A(CO_E_FAILEDTOGETWINDIR) // 0x80010134L
			A(CO_E_PATHTOOLONG) // 0x80010135L
			A(CO_E_FAILEDTOGENUUID) // 0x80010136L
			A(CO_E_FAILEDTOCREATEFILE) // 0x80010137L
			A(CO_E_FAILEDTOCLOSEHANDLE) // 0x80010138L
			A(CO_E_EXCEEDSYSACLLIMIT) // 0x80010139L
			A(CO_E_ACESINWRONGORDER) // 0x8001013AL
			A(CO_E_INCOMPATIBLESTREAMVERSION) // 0x8001013BL
			A(CO_E_FAILEDTOOPENPROCESSTOKEN) // 0x8001013CL
			A(CO_E_DECODEFAILED) // 0x8001013DL
			A(CO_E_ACNOTINITIALIZED) // 0x8001013FL
			A(CO_E_CANCEL_DISABLED) // 0x80010140L
			A(RPC_E_UNEXPECTED) // 0x8001FFFFL
			A(ERROR_AUDITING_DISABLED) // 0xC0090001L
			A(ERROR_ALL_SIDS_FILTERED) // 0xC0090002L
			A(ERROR_BIZRULES_NOT_ENABLED) // 0xC0090003L
			A(NTE_BAD_UID) // 0x80090001L
			A(NTE_BAD_HASH) // 0x80090002L
			A(NTE_BAD_KEY) // 0x80090003L
			A(NTE_BAD_LEN) // 0x80090004L
			A(NTE_BAD_DATA) // 0x80090005L
			A(NTE_BAD_SIGNATURE) // 0x80090006L
			A(NTE_BAD_VER) // 0x80090007L
			A(NTE_BAD_ALGID) // 0x80090008L
			A(NTE_BAD_FLAGS) // 0x80090009L
			A(NTE_BAD_TYPE) // 0x8009000AL
			A(NTE_BAD_KEY_STATE) // 0x8009000BL
			A(NTE_BAD_HASH_STATE) // 0x8009000CL
			A(NTE_NO_KEY) // 0x8009000DL
			A(NTE_NO_MEMORY) // 0x8009000EL
			A(NTE_EXISTS) // 0x8009000FL
			A(NTE_PERM) // 0x80090010L
			A(NTE_NOT_FOUND) // 0x80090011L
			A(NTE_DOUBLE_ENCRYPT) // 0x80090012L
			A(NTE_BAD_PROVIDER) // 0x80090013L
			A(NTE_BAD_PROV_TYPE) // 0x80090014L
			A(NTE_BAD_PUBLIC_KEY) // 0x80090015L
			A(NTE_BAD_KEYSET) // 0x80090016L
			A(NTE_PROV_TYPE_NOT_DEF) // 0x80090017L
			A(NTE_PROV_TYPE_ENTRY_BAD) // 0x80090018L
			A(NTE_KEYSET_NOT_DEF) // 0x80090019L
			A(NTE_KEYSET_ENTRY_BAD) // 0x8009001AL
			A(NTE_PROV_TYPE_NO_MATCH) // 0x8009001BL
			A(NTE_SIGNATURE_FILE_BAD) // 0x8009001CL
			A(NTE_PROVIDER_DLL_FAIL) // 0x8009001DL
			A(NTE_PROV_DLL_NOT_FOUND) // 0x8009001EL
			A(NTE_BAD_KEYSET_PARAM) // 0x8009001FL
			A(NTE_FAIL) // 0x80090020L
			A(NTE_SYS_ERR) // 0x80090021L
			A(NTE_SILENT_CONTEXT) // 0x80090022L
			A(NTE_TOKEN_KEYSET_STORAGE_FULL) // 0x80090023L
			A(NTE_TEMPORARY_PROFILE) // 0x80090024L
			A(NTE_FIXEDPARAMETER) // 0x80090025L
			A(NTE_INVALID_HANDLE) // 0x80090026L
			A(NTE_INVALID_PARAMETER) // 0x80090027L
			A(NTE_BUFFER_TOO_SMALL) // 0x80090028L
			A(NTE_NOT_SUPPORTED) // 0x80090029L
			A(NTE_NO_MORE_ITEMS) // 0x8009002AL
			A(NTE_BUFFERS_OVERLAP) // 0x8009002BL
			A(NTE_DECRYPTION_FAILURE) // 0x8009002CL
			A(NTE_INTERNAL_ERROR) // 0x8009002DL
			A(NTE_UI_REQUIRED) // 0x8009002EL
			A(NTE_HMAC_NOT_SUPPORTED) // 0x8009002FL
			A(SEC_E_INSUFFICIENT_MEMORY) // 0x80090300L
			A(SEC_E_INVALID_HANDLE) // 0x80090301L
			A(SEC_E_UNSUPPORTED_FUNCTION) // 0x80090302L
			A(SEC_E_TARGET_UNKNOWN) // 0x80090303L
			A(SEC_E_INTERNAL_ERROR) // 0x80090304L
			A(SEC_E_SECPKG_NOT_FOUND) // 0x80090305L
			A(SEC_E_NOT_OWNER) // 0x80090306L
			A(SEC_E_CANNOT_INSTALL) // 0x80090307L
			A(SEC_E_INVALID_TOKEN) // 0x80090308L
			A(SEC_E_CANNOT_PACK) // 0x80090309L
			A(SEC_E_QOP_NOT_SUPPORTED) // 0x8009030AL
			A(SEC_E_NO_IMPERSONATION) // 0x8009030BL
			A(SEC_E_LOGON_DENIED) // 0x8009030CL
			A(SEC_E_UNKNOWN_CREDENTIALS) // 0x8009030DL
			A(SEC_E_NO_CREDENTIALS) // 0x8009030EL
			A(SEC_E_MESSAGE_ALTERED) // 0x8009030FL
			A(SEC_E_OUT_OF_SEQUENCE) // 0x80090310L
			A(SEC_E_NO_AUTHENTICATING_AUTHORITY) // 0x80090311L
			A(SEC_I_CONTINUE_NEEDED) // 0x00090312L
			A(SEC_I_COMPLETE_NEEDED) // 0x00090313L
			A(SEC_I_COMPLETE_AND_CONTINUE) // 0x00090314L
			A(SEC_I_LOCAL_LOGON) // 0x00090315L
			A(SEC_E_BAD_PKGID) // 0x80090316L
			A(SEC_E_CONTEXT_EXPIRED) // 0x80090317L
			A(SEC_I_CONTEXT_EXPIRED) // 0x00090317L
			A(SEC_E_INCOMPLETE_MESSAGE) // 0x80090318L
			A(SEC_E_INCOMPLETE_CREDENTIALS) // 0x80090320L
			A(SEC_E_BUFFER_TOO_SMALL) // 0x80090321L
			A(SEC_I_INCOMPLETE_CREDENTIALS) // 0x00090320L
			A(SEC_I_RENEGOTIATE) // 0x00090321L
			A(SEC_E_WRONG_PRINCIPAL) // 0x80090322L
			A(SEC_I_NO_LSA_CONTEXT) // 0x00090323L
			A(SEC_E_TIME_SKEW) // 0x80090324L
			A(SEC_E_UNTRUSTED_ROOT) // 0x80090325L
			A(SEC_E_ILLEGAL_MESSAGE) // 0x80090326L
			A(SEC_E_CERT_UNKNOWN) // 0x80090327L
			A(SEC_E_CERT_EXPIRED) // 0x80090328L
			A(SEC_E_ENCRYPT_FAILURE) // 0x80090329L
			A(SEC_E_DECRYPT_FAILURE) // 0x80090330L
			A(SEC_E_ALGORITHM_MISMATCH) // 0x80090331L
			A(SEC_E_SECURITY_QOS_FAILED) // 0x80090332L
			A(SEC_E_UNFINISHED_CONTEXT_DELETED) // 0x80090333L
			A(SEC_E_NO_TGT_REPLY) // 0x80090334L
			A(SEC_E_NO_IP_ADDRESSES) // 0x80090335L
			A(SEC_E_WRONG_CREDENTIAL_HANDLE) // 0x80090336L
			A(SEC_E_CRYPTO_SYSTEM_INVALID) // 0x80090337L
			A(SEC_E_MAX_REFERRALS_EXCEEDED) // 0x80090338L
			A(SEC_E_MUST_BE_KDC) // 0x80090339L
			A(SEC_E_STRONG_CRYPTO_NOT_SUPPORTED) // 0x8009033AL
			A(SEC_E_TOO_MANY_PRINCIPALS) // 0x8009033BL
			A(SEC_E_NO_PA_DATA) // 0x8009033CL
			A(SEC_E_PKINIT_NAME_MISMATCH) // 0x8009033DL
			A(SEC_E_SMARTCARD_LOGON_REQUIRED) // 0x8009033EL
			A(SEC_E_SHUTDOWN_IN_PROGRESS) // 0x8009033FL
			A(SEC_E_KDC_INVALID_REQUEST) // 0x80090340L
			A(SEC_E_KDC_UNABLE_TO_REFER) // 0x80090341L
			A(SEC_E_KDC_UNKNOWN_ETYPE) // 0x80090342L
			A(SEC_E_UNSUPPORTED_PREAUTH) // 0x80090343L
			A(SEC_E_DELEGATION_REQUIRED) // 0x80090345L
			A(SEC_E_BAD_BINDINGS) // 0x80090346L
			A(SEC_E_MULTIPLE_ACCOUNTS) // 0x80090347L
			A(SEC_E_NO_KERB_KEY) // 0x80090348L
			A(SEC_E_CERT_WRONG_USAGE) // 0x80090349L
			A(SEC_E_DOWNGRADE_DETECTED) // 0x80090350L
			A(SEC_E_SMARTCARD_CERT_REVOKED) // 0x80090351L
			A(SEC_E_ISSUING_CA_UNTRUSTED) // 0x80090352L
			A(SEC_E_REVOCATION_OFFLINE_C) // 0x80090353L
			A(SEC_E_PKINIT_CLIENT_FAILURE) // 0x80090354L
			A(SEC_E_SMARTCARD_CERT_EXPIRED) // 0x80090355L
			A(SEC_E_NO_S4U_PROT_SUPPORT) // 0x80090356L
			A(SEC_E_CROSSREALM_DELEGATION_FAILURE) // 0x80090357L
			A(SEC_E_REVOCATION_OFFLINE_KDC) // 0x80090358L
			A(SEC_E_ISSUING_CA_UNTRUSTED_KDC) // 0x80090359L
			A(SEC_E_KDC_CERT_EXPIRED) // 0x8009035AL
			A(SEC_E_KDC_CERT_REVOKED) // 0x8009035BL
			A(SEC_I_SIGNATURE_NEEDED) // 0x0009035CL
			A(SEC_E_INVALID_PARAMETER) // 0x8009035DL
			A(SEC_E_DELEGATION_POLICY) // 0x8009035EL
			A(SEC_E_POLICY_NLTM_ONLY) // 0x8009035FL
			A(SEC_I_NO_RENEGOTIATION) // 0x00090360L
			A(SEC_E_NO_CONTEXT) // 0x80090361L
			A(SEC_E_PKU2U_CERT_FAILURE) // 0x80090362L
			A(SEC_E_MUTUAL_AUTH_FAILED) // 0x80090363L
			A(CRYPT_E_MSG_ERROR) // 0x80091001L
			A(CRYPT_E_UNKNOWN_ALGO) // 0x80091002L
			A(CRYPT_E_OID_FORMAT) // 0x80091003L
			A(CRYPT_E_INVALID_MSG_TYPE) // 0x80091004L
			A(CRYPT_E_UNEXPECTED_ENCODING) // 0x80091005L
			A(CRYPT_E_AUTH_ATTR_MISSING) // 0x80091006L
			A(CRYPT_E_HASH_VALUE) // 0x80091007L
			A(CRYPT_E_INVALID_INDEX) // 0x80091008L
			A(CRYPT_E_ALREADY_DECRYPTED) // 0x80091009L
			A(CRYPT_E_NOT_DECRYPTED) // 0x8009100AL
			A(CRYPT_E_RECIPIENT_NOT_FOUND) // 0x8009100BL
			A(CRYPT_E_CONTROL_TYPE) // 0x8009100CL
			A(CRYPT_E_ISSUER_SERIALNUMBER) // 0x8009100DL
			A(CRYPT_E_SIGNER_NOT_FOUND) // 0x8009100EL
			A(CRYPT_E_ATTRIBUTES_MISSING) // 0x8009100FL
			A(CRYPT_E_STREAM_MSG_NOT_READY) // 0x80091010L
			A(CRYPT_E_STREAM_INSUFFICIENT_DATA) // 0x80091011L
			A(CRYPT_I_NEW_PROTECTION_REQUIRED) // 0x00091012L
			A(CRYPT_E_BAD_LEN) // 0x80092001L
			A(CRYPT_E_BAD_ENCODE) // 0x80092002L
			A(CRYPT_E_FILE_ERROR) // 0x80092003L
			A(CRYPT_E_NOT_FOUND) // 0x80092004L
			A(CRYPT_E_EXISTS) // 0x80092005L
			A(CRYPT_E_NO_PROVIDER) // 0x80092006L
			A(CRYPT_E_SELF_SIGNED) // 0x80092007L
			A(CRYPT_E_DELETED_PREV) // 0x80092008L
			A(CRYPT_E_NO_MATCH) // 0x80092009L
			A(CRYPT_E_UNEXPECTED_MSG_TYPE) // 0x8009200AL
			A(CRYPT_E_NO_KEY_PROPERTY) // 0x8009200BL
			A(CRYPT_E_NO_DECRYPT_CERT) // 0x8009200CL
			A(CRYPT_E_BAD_MSG) // 0x8009200DL
			A(CRYPT_E_NO_SIGNER) // 0x8009200EL
			A(CRYPT_E_PENDING_CLOSE) // 0x8009200FL
			A(CRYPT_E_REVOKED) // 0x80092010L
			A(CRYPT_E_NO_REVOCATION_DLL) // 0x80092011L
			A(CRYPT_E_NO_REVOCATION_CHECK) // 0x80092012L
			A(CRYPT_E_REVOCATION_OFFLINE) // 0x80092013L
			A(CRYPT_E_NOT_IN_REVOCATION_DATABASE) // 0x80092014L
			A(CRYPT_E_INVALID_NUMERIC_STRING) // 0x80092020L
			A(CRYPT_E_INVALID_PRINTABLE_STRING) // 0x80092021L
			A(CRYPT_E_INVALID_IA5_STRING) // 0x80092022L
			A(CRYPT_E_INVALID_X500_STRING) // 0x80092023L
			A(CRYPT_E_NOT_CHAR_STRING) // 0x80092024L
			A(CRYPT_E_FILERESIZED) // 0x80092025L
			A(CRYPT_E_SECURITY_SETTINGS) // 0x80092026L
			A(CRYPT_E_NO_VERIFY_USAGE_DLL) // 0x80092027L
			A(CRYPT_E_NO_VERIFY_USAGE_CHECK) // 0x80092028L
			A(CRYPT_E_VERIFY_USAGE_OFFLINE) // 0x80092029L
			A(CRYPT_E_NOT_IN_CTL) // 0x8009202AL
			A(CRYPT_E_NO_TRUSTED_SIGNER) // 0x8009202BL
			A(CRYPT_E_MISSING_PUBKEY_PARA) // 0x8009202CL
			A(CRYPT_E_OSS_ERROR) // 0x80093000L
			A(OSS_MORE_BUF) // 0x80093001L
			A(OSS_NEGATIVE_UINTEGER) // 0x80093002L
			A(OSS_PDU_RANGE) // 0x80093003L
			A(OSS_MORE_INPUT) // 0x80093004L
			A(OSS_DATA_ERROR) // 0x80093005L
			A(OSS_BAD_ARG) // 0x80093006L
			A(OSS_BAD_VERSION) // 0x80093007L
			A(OSS_OUT_MEMORY) // 0x80093008L
			A(OSS_PDU_MISMATCH) // 0x80093009L
			A(OSS_LIMITED) // 0x8009300AL
			A(OSS_BAD_PTR) // 0x8009300BL
			A(OSS_BAD_TIME) // 0x8009300CL
			A(OSS_INDEFINITE_NOT_SUPPORTED) // 0x8009300DL
			A(OSS_MEM_ERROR) // 0x8009300EL
			A(OSS_BAD_TABLE) // 0x8009300FL
			A(OSS_TOO_LONG) // 0x80093010L
			A(OSS_CONSTRAINT_VIOLATED) // 0x80093011L
			A(OSS_FATAL_ERROR) // 0x80093012L
			A(OSS_ACCESS_SERIALIZATION_ERROR) // 0x80093013L
			A(OSS_NULL_TBL) // 0x80093014L
			A(OSS_NULL_FCN) // 0x80093015L
			A(OSS_BAD_ENCRULES) // 0x80093016L
			A(OSS_UNAVAIL_ENCRULES) // 0x80093017L
			A(OSS_CANT_OPEN_TRACE_WINDOW) // 0x80093018L
			A(OSS_UNIMPLEMENTED) // 0x80093019L
			A(OSS_OID_DLL_NOT_LINKED) // 0x8009301AL
			A(OSS_CANT_OPEN_TRACE_FILE) // 0x8009301BL
			A(OSS_TRACE_FILE_ALREADY_OPEN) // 0x8009301CL
			A(OSS_TABLE_MISMATCH) // 0x8009301DL
			A(OSS_TYPE_NOT_SUPPORTED) // 0x8009301EL
			A(OSS_REAL_DLL_NOT_LINKED) // 0x8009301FL
			A(OSS_REAL_CODE_NOT_LINKED) // 0x80093020L
			A(OSS_OUT_OF_RANGE) // 0x80093021L
			A(OSS_COPIER_DLL_NOT_LINKED) // 0x80093022L
			A(OSS_CONSTRAINT_DLL_NOT_LINKED) // 0x80093023L
			A(OSS_COMPARATOR_DLL_NOT_LINKED) // 0x80093024L
			A(OSS_COMPARATOR_CODE_NOT_LINKED) // 0x80093025L
			A(OSS_MEM_MGR_DLL_NOT_LINKED) // 0x80093026L
			A(OSS_PDV_DLL_NOT_LINKED) // 0x80093027L
			A(OSS_PDV_CODE_NOT_LINKED) // 0x80093028L
			A(OSS_API_DLL_NOT_LINKED) // 0x80093029L
			A(OSS_BERDER_DLL_NOT_LINKED) // 0x8009302AL
			A(OSS_PER_DLL_NOT_LINKED) // 0x8009302BL
			A(OSS_OPEN_TYPE_ERROR) // 0x8009302CL
			A(OSS_MUTEX_NOT_CREATED) // 0x8009302DL
			A(OSS_CANT_CLOSE_TRACE_FILE) // 0x8009302EL
			A(CRYPT_E_ASN1_ERROR) // 0x80093100L
			A(CRYPT_E_ASN1_INTERNAL) // 0x80093101L
			A(CRYPT_E_ASN1_EOD) // 0x80093102L
			A(CRYPT_E_ASN1_CORRUPT) // 0x80093103L
			A(CRYPT_E_ASN1_LARGE) // 0x80093104L
			A(CRYPT_E_ASN1_CONSTRAINT) // 0x80093105L
			A(CRYPT_E_ASN1_MEMORY) // 0x80093106L
			A(CRYPT_E_ASN1_OVERFLOW) // 0x80093107L
			A(CRYPT_E_ASN1_BADPDU) // 0x80093108L
			A(CRYPT_E_ASN1_BADARGS) // 0x80093109L
			A(CRYPT_E_ASN1_BADREAL) // 0x8009310AL
			A(CRYPT_E_ASN1_BADTAG) // 0x8009310BL
			A(CRYPT_E_ASN1_CHOICE) // 0x8009310CL
			A(CRYPT_E_ASN1_RULE) // 0x8009310DL
			A(CRYPT_E_ASN1_UTF8) // 0x8009310EL
			A(CRYPT_E_ASN1_PDU_TYPE) // 0x80093133L
			A(CRYPT_E_ASN1_NYI) // 0x80093134L
			A(CRYPT_E_ASN1_EXTENDED) // 0x80093201L
			A(CRYPT_E_ASN1_NOEOD) // 0x80093202L
			A(CERTSRV_E_BAD_REQUESTSUBJECT) // 0x80094001L
			A(CERTSRV_E_NO_REQUEST) // 0x80094002L
			A(CERTSRV_E_BAD_REQUESTSTATUS) // 0x80094003L
			A(CERTSRV_E_PROPERTY_EMPTY) // 0x80094004L
			A(CERTSRV_E_INVALID_CA_CERTIFICATE) // 0x80094005L
			A(CERTSRV_E_SERVER_SUSPENDED) // 0x80094006L
			A(CERTSRV_E_ENCODING_LENGTH) // 0x80094007L
			A(CERTSRV_E_ROLECONFLICT) // 0x80094008L
			A(CERTSRV_E_RESTRICTEDOFFICER) // 0x80094009L
			A(CERTSRV_E_KEY_ARCHIVAL_NOT_CONFIGURED) // 0x8009400AL
			A(CERTSRV_E_NO_VALID_KRA) // 0x8009400BL
			A(CERTSRV_E_BAD_REQUEST_KEY_ARCHIVAL) // 0x8009400CL
			A(CERTSRV_E_NO_CAADMIN_DEFINED) // 0x8009400DL
			A(CERTSRV_E_BAD_RENEWAL_CERT_ATTRIBUTE) // 0x8009400EL
			A(CERTSRV_E_NO_DB_SESSIONS) // 0x8009400FL
			A(CERTSRV_E_ALIGNMENT_FAULT) // 0x80094010L
			A(CERTSRV_E_ENROLL_DENIED) // 0x80094011L
			A(CERTSRV_E_TEMPLATE_DENIED) // 0x80094012L
			A(CERTSRV_E_DOWNLEVEL_DC_SSL_OR_UPGRADE) // 0x80094013L
			A(CERTSRV_E_ADMIN_DENIED_REQUEST) // 0x80094014L
			A(CERTSRV_E_NO_POLICY_SERVER) // 0x80094015L
			A(CERTSRV_E_UNSUPPORTED_CERT_TYPE) // 0x80094800L
			A(CERTSRV_E_NO_CERT_TYPE) // 0x80094801L
			A(CERTSRV_E_TEMPLATE_CONFLICT) // 0x80094802L
			A(CERTSRV_E_SUBJECT_ALT_NAME_REQUIRED) // 0x80094803L
			A(CERTSRV_E_ARCHIVED_KEY_REQUIRED) // 0x80094804L
			A(CERTSRV_E_SMIME_REQUIRED) // 0x80094805L
			A(CERTSRV_E_BAD_RENEWAL_SUBJECT) // 0x80094806L
			A(CERTSRV_E_BAD_TEMPLATE_VERSION) // 0x80094807L
			A(CERTSRV_E_TEMPLATE_POLICY_REQUIRED) // 0x80094808L
			A(CERTSRV_E_SIGNATURE_POLICY_REQUIRED) // 0x80094809L
			A(CERTSRV_E_SIGNATURE_COUNT) // 0x8009480AL
			A(CERTSRV_E_SIGNATURE_REJECTED) // 0x8009480BL
			A(CERTSRV_E_ISSUANCE_POLICY_REQUIRED) // 0x8009480CL
			A(CERTSRV_E_SUBJECT_UPN_REQUIRED) // 0x8009480DL
			A(CERTSRV_E_SUBJECT_DIRECTORY_GUID_REQUIRED) // 0x8009480EL
			A(CERTSRV_E_SUBJECT_DNS_REQUIRED) // 0x8009480FL
			A(CERTSRV_E_ARCHIVED_KEY_UNEXPECTED) // 0x80094810L
			A(CERTSRV_E_KEY_LENGTH) // 0x80094811L
			A(CERTSRV_E_SUBJECT_EMAIL_REQUIRED) // 0x80094812L
			A(CERTSRV_E_UNKNOWN_CERT_TYPE) // 0x80094813L
			A(CERTSRV_E_CERT_TYPE_OVERLAP) // 0x80094814L
			A(CERTSRV_E_TOO_MANY_SIGNATURES) // 0x80094815L
			A(XENROLL_E_KEY_NOT_EXPORTABLE) // 0x80095000L
			A(XENROLL_E_CANNOT_ADD_ROOT_CERT) // 0x80095001L
			A(XENROLL_E_RESPONSE_KA_HASH_NOT_FOUND) // 0x80095002L
			A(XENROLL_E_RESPONSE_UNEXPECTED_KA_HASH) // 0x80095003L
			A(XENROLL_E_RESPONSE_KA_HASH_MISMATCH) // 0x80095004L
			A(XENROLL_E_KEYSPEC_SMIME_MISMATCH) // 0x80095005L
			A(TRUST_E_SYSTEM_ERROR) // 0x80096001L
			A(TRUST_E_NO_SIGNER_CERT) // 0x80096002L
			A(TRUST_E_COUNTER_SIGNER) // 0x80096003L
			A(TRUST_E_CERT_SIGNATURE) // 0x80096004L
			A(TRUST_E_TIME_STAMP) // 0x80096005L
			A(TRUST_E_BAD_DIGEST) // 0x80096010L
			A(TRUST_E_BASIC_CONSTRAINTS) // 0x80096019L
			A(TRUST_E_FINANCIAL_CRITERIA) // 0x8009601EL
			A(MSSIPOTF_E_OUTOFMEMRANGE) // 0x80097001L
			A(MSSIPOTF_E_CANTGETOBJECT) // 0x80097002L
			A(MSSIPOTF_E_NOHEADTABLE) // 0x80097003L
			A(MSSIPOTF_E_BAD_MAGICNUMBER) // 0x80097004L
			A(MSSIPOTF_E_BAD_OFFSET_TABLE) // 0x80097005L
			A(MSSIPOTF_E_TABLE_TAGORDER) // 0x80097006L
			A(MSSIPOTF_E_TABLE_LONGWORD) // 0x80097007L
			A(MSSIPOTF_E_BAD_FIRST_TABLE_PLACEMENT) // 0x80097008L
			A(MSSIPOTF_E_TABLES_OVERLAP) // 0x80097009L
			A(MSSIPOTF_E_TABLE_PADBYTES) // 0x8009700AL
			A(MSSIPOTF_E_FILETOOSMALL) // 0x8009700BL
			A(MSSIPOTF_E_TABLE_CHECKSUM) // 0x8009700CL
			A(MSSIPOTF_E_FILE_CHECKSUM) // 0x8009700DL
			A(MSSIPOTF_E_FAILED_POLICY) // 0x80097010L
			A(MSSIPOTF_E_FAILED_HINTS_CHECK) // 0x80097011L
			A(MSSIPOTF_E_NOT_OPENTYPE) // 0x80097012L
			A(MSSIPOTF_E_FILE) // 0x80097013L
			A(MSSIPOTF_E_CRYPT) // 0x80097014L
			A(MSSIPOTF_E_BADVERSION) // 0x80097015L
			A(MSSIPOTF_E_DSIG_STRUCTURE) // 0x80097016L
			A(MSSIPOTF_E_PCONST_CHECK) // 0x80097017L
			A(MSSIPOTF_E_STRUCTURE) // 0x80097018L
			A(ERROR_CRED_REQUIRES_CONFIRMATION) // 0x80097019L
			A(TRUST_E_PROVIDER_UNKNOWN) // 0x800B0001L
			A(TRUST_E_ACTION_UNKNOWN) // 0x800B0002L
			A(TRUST_E_SUBJECT_FORM_UNKNOWN) // 0x800B0003L
			A(TRUST_E_SUBJECT_NOT_TRUSTED) // 0x800B0004L
			A(DIGSIG_E_ENCODE) // 0x800B0005L
			A(DIGSIG_E_DECODE) // 0x800B0006L
			A(DIGSIG_E_EXTENSIBILITY) // 0x800B0007L
			A(DIGSIG_E_CRYPTO) // 0x800B0008L
			A(PERSIST_E_SIZEDEFINITE) // 0x800B0009L
			A(PERSIST_E_SIZEINDEFINITE) // 0x800B000AL
			A(PERSIST_E_NOTSELFSIZING) // 0x800B000BL
			A(TRUST_E_NOSIGNATURE) // 0x800B0100L
			A(CERT_E_EXPIRED) // 0x800B0101L
			A(CERT_E_VALIDITYPERIODNESTING) // 0x800B0102L
			A(CERT_E_ROLE) // 0x800B0103L
			A(CERT_E_PATHLENCONST) // 0x800B0104L
			A(CERT_E_CRITICAL) // 0x800B0105L
			A(CERT_E_PURPOSE) // 0x800B0106L
			A(CERT_E_ISSUERCHAINING) // 0x800B0107L
			A(CERT_E_MALFORMED) // 0x800B0108L
			A(CERT_E_UNTRUSTEDROOT) // 0x800B0109L
			A(CERT_E_CHAINING) // 0x800B010AL
			A(TRUST_E_FAIL) // 0x800B010BL
			A(CERT_E_REVOKED) // 0x800B010CL
			A(CERT_E_UNTRUSTEDTESTROOT) // 0x800B010DL
			A(CERT_E_REVOCATION_FAILURE) // 0x800B010EL
			A(CERT_E_CN_NO_MATCH) // 0x800B010FL
			A(CERT_E_WRONG_USAGE) // 0x800B0110L
			A(TRUST_E_EXPLICIT_DISTRUST) // 0x800B0111L
			A(CERT_E_UNTRUSTEDCA) // 0x800B0112L
			A(CERT_E_INVALID_POLICY) // 0x800B0113L
			A(CERT_E_INVALID_NAME) // 0x800B0114L
			A(SPAPI_E_EXPECTED_SECTION_NAME) // 0x800F0000L
			A(SPAPI_E_BAD_SECTION_NAME_LINE) // 0x800F0001L
			A(SPAPI_E_SECTION_NAME_TOO_LONG) // 0x800F0002L
			A(SPAPI_E_GENERAL_SYNTAX) // 0x800F0003L
			A(SPAPI_E_WRONG_INF_STYLE) // 0x800F0100L
			A(SPAPI_E_SECTION_NOT_FOUND) // 0x800F0101L
			A(SPAPI_E_LINE_NOT_FOUND) // 0x800F0102L
			A(SPAPI_E_NO_BACKUP) // 0x800F0103L
			A(SPAPI_E_NO_ASSOCIATED_CLASS) // 0x800F0200L
			A(SPAPI_E_CLASS_MISMATCH) // 0x800F0201L
			A(SPAPI_E_DUPLICATE_FOUND) // 0x800F0202L
			A(SPAPI_E_NO_DRIVER_SELECTED) // 0x800F0203L
			A(SPAPI_E_KEY_DOES_NOT_EXIST) // 0x800F0204L
			A(SPAPI_E_INVALID_DEVINST_NAME) // 0x800F0205L
			A(SPAPI_E_INVALID_CLASS) // 0x800F0206L
			A(SPAPI_E_DEVINST_ALREADY_EXISTS) // 0x800F0207L
			A(SPAPI_E_DEVINFO_NOT_REGISTERED) // 0x800F0208L
			A(SPAPI_E_INVALID_REG_PROPERTY) // 0x800F0209L
			A(SPAPI_E_NO_INF) // 0x800F020AL
			A(SPAPI_E_NO_SUCH_DEVINST) // 0x800F020BL
			A(SPAPI_E_CANT_LOAD_CLASS_ICON) // 0x800F020CL
			A(SPAPI_E_INVALID_CLASS_INSTALLER) // 0x800F020DL
			A(SPAPI_E_DI_DO_DEFAULT) // 0x800F020EL
			A(SPAPI_E_DI_NOFILECOPY) // 0x800F020FL
			A(SPAPI_E_INVALID_HWPROFILE) // 0x800F0210L
			A(SPAPI_E_NO_DEVICE_SELECTED) // 0x800F0211L
			A(SPAPI_E_DEVINFO_LIST_LOCKED) // 0x800F0212L
			A(SPAPI_E_DEVINFO_DATA_LOCKED) // 0x800F0213L
			A(SPAPI_E_DI_BAD_PATH) // 0x800F0214L
			A(SPAPI_E_NO_CLASSINSTALL_PARAMS) // 0x800F0215L
			A(SPAPI_E_FILEQUEUE_LOCKED) // 0x800F0216L
			A(SPAPI_E_BAD_SERVICE_INSTALLSECT) // 0x800F0217L
			A(SPAPI_E_NO_CLASS_DRIVER_LIST) // 0x800F0218L
			A(SPAPI_E_NO_ASSOCIATED_SERVICE) // 0x800F0219L
			A(SPAPI_E_NO_DEFAULT_DEVICE_INTERFACE) // 0x800F021AL
			A(SPAPI_E_DEVICE_INTERFACE_ACTIVE) // 0x800F021BL
			A(SPAPI_E_DEVICE_INTERFACE_REMOVED) // 0x800F021CL
			A(SPAPI_E_BAD_INTERFACE_INSTALLSECT) // 0x800F021DL
			A(SPAPI_E_NO_SUCH_INTERFACE_CLASS) // 0x800F021EL
			A(SPAPI_E_INVALID_REFERENCE_STRING) // 0x800F021FL
			A(SPAPI_E_INVALID_MACHINENAME) // 0x800F0220L
			A(SPAPI_E_REMOTE_COMM_FAILURE) // 0x800F0221L
			A(SPAPI_E_MACHINE_UNAVAILABLE) // 0x800F0222L
			A(SPAPI_E_NO_CONFIGMGR_SERVICES) // 0x800F0223L
			A(SPAPI_E_INVALID_PROPPAGE_PROVIDER) // 0x800F0224L
			A(SPAPI_E_NO_SUCH_DEVICE_INTERFACE) // 0x800F0225L
			A(SPAPI_E_DI_POSTPROCESSING_REQUIRED) // 0x800F0226L
			A(SPAPI_E_INVALID_COINSTALLER) // 0x800F0227L
			A(SPAPI_E_NO_COMPAT_DRIVERS) // 0x800F0228L
			A(SPAPI_E_NO_DEVICE_ICON) // 0x800F0229L
			A(SPAPI_E_INVALID_INF_LOGCONFIG) // 0x800F022AL
			A(SPAPI_E_DI_DONT_INSTALL) // 0x800F022BL
			A(SPAPI_E_INVALID_FILTER_DRIVER) // 0x800F022CL
			A(SPAPI_E_NON_WINDOWS_NT_DRIVER) // 0x800F022DL
			A(SPAPI_E_NON_WINDOWS_DRIVER) // 0x800F022EL
			A(SPAPI_E_NO_CATALOG_FOR_OEM_INF) // 0x800F022FL
			A(SPAPI_E_DEVINSTALL_QUEUE_NONNATIVE) // 0x800F0230L
			A(SPAPI_E_NOT_DISABLEABLE) // 0x800F0231L
			A(SPAPI_E_CANT_REMOVE_DEVINST) // 0x800F0232L
			A(SPAPI_E_INVALID_TARGET) // 0x800F0233L
			A(SPAPI_E_DRIVER_NONNATIVE) // 0x800F0234L
			A(SPAPI_E_IN_WOW64) // 0x800F0235L
			A(SPAPI_E_SET_SYSTEM_RESTORE_POINT) // 0x800F0236L
			A(SPAPI_E_INCORRECTLY_COPIED_INF) // 0x800F0237L
			A(SPAPI_E_SCE_DISABLED) // 0x800F0238L
			A(SPAPI_E_UNKNOWN_EXCEPTION) // 0x800F0239L
			A(SPAPI_E_PNP_REGISTRY_ERROR) // 0x800F023AL
			A(SPAPI_E_REMOTE_REQUEST_UNSUPPORTED) // 0x800F023BL
			A(SPAPI_E_NOT_AN_INSTALLED_OEM_INF) // 0x800F023CL
			A(SPAPI_E_INF_IN_USE_BY_DEVICES) // 0x800F023DL
			A(SPAPI_E_DI_FUNCTION_OBSOLETE) // 0x800F023EL
			A(SPAPI_E_NO_AUTHENTICODE_CATALOG) // 0x800F023FL
			A(SPAPI_E_AUTHENTICODE_DISALLOWED) // 0x800F0240L
			A(SPAPI_E_AUTHENTICODE_TRUSTED_PUBLISHER) // 0x800F0241L
			A(SPAPI_E_AUTHENTICODE_TRUST_NOT_ESTABLISHED) // 0x800F0242L
			A(SPAPI_E_AUTHENTICODE_PUBLISHER_NOT_TRUSTED) // 0x800F0243L
			A(SPAPI_E_SIGNATURE_OSATTRIBUTE_MISMATCH) // 0x800F0244L
			A(SPAPI_E_ONLY_VALIDATE_VIA_AUTHENTICODE) // 0x800F0245L
			A(SPAPI_E_DEVICE_INSTALLER_NOT_READY) // 0x800F0246L
			A(SPAPI_E_DRIVER_STORE_ADD_FAILED) // 0x800F0247L
			A(SPAPI_E_DEVICE_INSTALL_BLOCKED) // 0x800F0248L
			A(SPAPI_E_DRIVER_INSTALL_BLOCKED) // 0x800F0249L
			A(SPAPI_E_WRONG_INF_TYPE) // 0x800F024AL
			A(SPAPI_E_FILE_HASH_NOT_IN_CATALOG) // 0x800F024BL
			A(SPAPI_E_DRIVER_STORE_DELETE_FAILED) // 0x800F024CL
			A(SPAPI_E_UNRECOVERABLE_STACK_OVERFLOW) // 0x800F0300L
			A(SPAPI_E_ERROR_NOT_INSTALLED) // 0x800F1000L
			A(SCARD_F_INTERNAL_ERROR) // 0x80100001L
			A(SCARD_E_CANCELLED) // 0x80100002L
			A(SCARD_E_INVALID_HANDLE) // 0x80100003L
			A(SCARD_E_INVALID_PARAMETER) // 0x80100004L
			A(SCARD_E_INVALID_TARGET) // 0x80100005L
			A(SCARD_E_NO_MEMORY) // 0x80100006L
			A(SCARD_F_WAITED_TOO_LONG) // 0x80100007L
			A(SCARD_E_INSUFFICIENT_BUFFER) // 0x80100008L
			A(SCARD_E_UNKNOWN_READER) // 0x80100009L
			A(SCARD_E_TIMEOUT) // 0x8010000AL
			A(SCARD_E_SHARING_VIOLATION) // 0x8010000BL
			A(SCARD_E_NO_SMARTCARD) // 0x8010000CL
			A(SCARD_E_UNKNOWN_CARD) // 0x8010000DL
			A(SCARD_E_CANT_DISPOSE) // 0x8010000EL
			A(SCARD_E_PROTO_MISMATCH) // 0x8010000FL
			A(SCARD_E_NOT_READY) // 0x80100010L
			A(SCARD_E_INVALID_VALUE) // 0x80100011L
			A(SCARD_E_SYSTEM_CANCELLED) // 0x80100012L
			A(SCARD_F_COMM_ERROR) // 0x80100013L
			A(SCARD_F_UNKNOWN_ERROR) // 0x80100014L
			A(SCARD_E_INVALID_ATR) // 0x80100015L
			A(SCARD_E_NOT_TRANSACTED) // 0x80100016L
			A(SCARD_E_READER_UNAVAILABLE) // 0x80100017L
			A(SCARD_P_SHUTDOWN) // 0x80100018L
			A(SCARD_E_PCI_TOO_SMALL) // 0x80100019L
			A(SCARD_E_READER_UNSUPPORTED) // 0x8010001AL
			A(SCARD_E_DUPLICATE_READER) // 0x8010001BL
			A(SCARD_E_CARD_UNSUPPORTED) // 0x8010001CL
			A(SCARD_E_NO_SERVICE) // 0x8010001DL
			A(SCARD_E_SERVICE_STOPPED) // 0x8010001EL
			A(SCARD_E_UNEXPECTED) // 0x8010001FL
			A(SCARD_E_ICC_INSTALLATION) // 0x80100020L
			A(SCARD_E_ICC_CREATEORDER) // 0x80100021L
			A(SCARD_E_UNSUPPORTED_FEATURE) // 0x80100022L
			A(SCARD_E_DIR_NOT_FOUND) // 0x80100023L
			A(SCARD_E_FILE_NOT_FOUND) // 0x80100024L
			A(SCARD_E_NO_DIR) // 0x80100025L
			A(SCARD_E_NO_FILE) // 0x80100026L
			A(SCARD_E_NO_ACCESS) // 0x80100027L
			A(SCARD_E_WRITE_TOO_MANY) // 0x80100028L
			A(SCARD_E_BAD_SEEK) // 0x80100029L
			A(SCARD_E_INVALID_CHV) // 0x8010002AL
			A(SCARD_E_UNKNOWN_RES_MNG) // 0x8010002BL
			A(SCARD_E_NO_SUCH_CERTIFICATE) // 0x8010002CL
			A(SCARD_E_CERTIFICATE_UNAVAILABLE) // 0x8010002DL
			A(SCARD_E_NO_READERS_AVAILABLE) // 0x8010002EL
			A(SCARD_E_COMM_DATA_LOST) // 0x8010002FL
			A(SCARD_E_NO_KEY_CONTAINER) // 0x80100030L
			A(SCARD_E_SERVER_TOO_BUSY) // 0x80100031L
			A(SCARD_E_PIN_CACHE_EXPIRED) // 0x80100032L
			A(SCARD_E_NO_PIN_CACHE) // 0x80100033L
			A(SCARD_E_READ_ONLY_CARD) // 0x80100034L
			A(SCARD_W_UNSUPPORTED_CARD) // 0x80100065L
			A(SCARD_W_UNRESPONSIVE_CARD) // 0x80100066L
			A(SCARD_W_UNPOWERED_CARD) // 0x80100067L
			A(SCARD_W_RESET_CARD) // 0x80100068L
			A(SCARD_W_REMOVED_CARD) // 0x80100069L
			A(SCARD_W_SECURITY_VIOLATION) // 0x8010006AL
			A(SCARD_W_WRONG_CHV) // 0x8010006BL
			A(SCARD_W_CHV_BLOCKED) // 0x8010006CL
			A(SCARD_W_EOF) // 0x8010006DL
			A(SCARD_W_CANCELLED_BY_USER) // 0x8010006EL
			A(SCARD_W_CARD_NOT_AUTHENTICATED) // 0x8010006FL
			A(SCARD_W_CACHE_ITEM_NOT_FOUND) // 0x80100070L
			A(SCARD_W_CACHE_ITEM_STALE) // 0x80100071L
			A(SCARD_W_CACHE_ITEM_TOO_BIG) // 0x80100072L
			A(COMADMIN_E_OBJECTERRORS) // 0x80110401L
			A(COMADMIN_E_OBJECTINVALID) // 0x80110402L
			A(COMADMIN_E_KEYMISSING) // 0x80110403L
			A(COMADMIN_E_ALREADYINSTALLED) // 0x80110404L
			A(COMADMIN_E_APP_FILE_WRITEFAIL) // 0x80110407L
			A(COMADMIN_E_APP_FILE_READFAIL) // 0x80110408L
			A(COMADMIN_E_APP_FILE_VERSION) // 0x80110409L
			A(COMADMIN_E_BADPATH) // 0x8011040AL
			A(COMADMIN_E_APPLICATIONEXISTS) // 0x8011040BL
			A(COMADMIN_E_ROLEEXISTS) // 0x8011040CL
			A(COMADMIN_E_CANTCOPYFILE) // 0x8011040DL
			A(COMADMIN_E_NOUSER) // 0x8011040FL
			A(COMADMIN_E_INVALIDUSERIDS) // 0x80110410L
			A(COMADMIN_E_NOREGISTRYCLSID) // 0x80110411L
			A(COMADMIN_E_BADREGISTRYPROGID) // 0x80110412L
			A(COMADMIN_E_AUTHENTICATIONLEVEL) // 0x80110413L
			A(COMADMIN_E_USERPASSWDNOTVALID) // 0x80110414L
			A(COMADMIN_E_CLSIDORIIDMISMATCH) // 0x80110418L
			A(COMADMIN_E_REMOTEINTERFACE) // 0x80110419L
			A(COMADMIN_E_DLLREGISTERSERVER) // 0x8011041AL
			A(COMADMIN_E_NOSERVERSHARE) // 0x8011041BL
			A(COMADMIN_E_DLLLOADFAILED) // 0x8011041DL
			A(COMADMIN_E_BADREGISTRYLIBID) // 0x8011041EL
			A(COMADMIN_E_APPDIRNOTFOUND) // 0x8011041FL
			A(COMADMIN_E_REGISTRARFAILED) // 0x80110423L
			A(COMADMIN_E_COMPFILE_DOESNOTEXIST) // 0x80110424L
			A(COMADMIN_E_COMPFILE_LOADDLLFAIL) // 0x80110425L
			A(COMADMIN_E_COMPFILE_GETCLASSOBJ) // 0x80110426L
			A(COMADMIN_E_COMPFILE_CLASSNOTAVAIL) // 0x80110427L
			A(COMADMIN_E_COMPFILE_BADTLB) // 0x80110428L
			A(COMADMIN_E_COMPFILE_NOTINSTALLABLE) // 0x80110429L
			A(COMADMIN_E_NOTCHANGEABLE) // 0x8011042AL
			A(COMADMIN_E_NOTDELETEABLE) // 0x8011042BL
			A(COMADMIN_E_SESSION) // 0x8011042CL
			A(COMADMIN_E_COMP_MOVE_LOCKED) // 0x8011042DL
			A(COMADMIN_E_COMP_MOVE_BAD_DEST) // 0x8011042EL
			A(COMADMIN_E_REGISTERTLB) // 0x80110430L
			A(COMADMIN_E_SYSTEMAPP) // 0x80110433L
			A(COMADMIN_E_COMPFILE_NOREGISTRAR) // 0x80110434L
			A(COMADMIN_E_COREQCOMPINSTALLED) // 0x80110435L
			A(COMADMIN_E_SERVICENOTINSTALLED) // 0x80110436L
			A(COMADMIN_E_PROPERTYSAVEFAILED) // 0x80110437L
			A(COMADMIN_E_OBJECTEXISTS) // 0x80110438L
			A(COMADMIN_E_COMPONENTEXISTS) // 0x80110439L
			A(COMADMIN_E_REGFILE_CORRUPT) // 0x8011043BL
			A(COMADMIN_E_PROPERTY_OVERFLOW) // 0x8011043CL
			A(COMADMIN_E_NOTINREGISTRY) // 0x8011043EL
			A(COMADMIN_E_OBJECTNOTPOOLABLE) // 0x8011043FL
			A(COMADMIN_E_APPLID_MATCHES_CLSID) // 0x80110446L
			A(COMADMIN_E_ROLE_DOES_NOT_EXIST) // 0x80110447L
			A(COMADMIN_E_START_APP_NEEDS_COMPONENTS) // 0x80110448L
			A(COMADMIN_E_REQUIRES_DIFFERENT_PLATFORM) // 0x80110449L
			A(COMADMIN_E_CAN_NOT_EXPORT_APP_PROXY) // 0x8011044AL
			A(COMADMIN_E_CAN_NOT_START_APP) // 0x8011044BL
			A(COMADMIN_E_CAN_NOT_EXPORT_SYS_APP) // 0x8011044CL
			A(COMADMIN_E_CANT_SUBSCRIBE_TO_COMPONENT) // 0x8011044DL
			A(COMADMIN_E_EVENTCLASS_CANT_BE_SUBSCRIBER) // 0x8011044EL
			A(COMADMIN_E_LIB_APP_PROXY_INCOMPATIBLE) // 0x8011044FL
			A(COMADMIN_E_BASE_PARTITION_ONLY) // 0x80110450L
			A(COMADMIN_E_START_APP_DISABLED) // 0x80110451L
			A(COMADMIN_E_CAT_DUPLICATE_PARTITION_NAME) // 0x80110457L
			A(COMADMIN_E_CAT_INVALID_PARTITION_NAME) // 0x80110458L
			A(COMADMIN_E_CAT_PARTITION_IN_USE) // 0x80110459L
			A(COMADMIN_E_FILE_PARTITION_DUPLICATE_FILES) // 0x8011045AL
			A(COMADMIN_E_CAT_IMPORTED_COMPONENTS_NOT_ALLOWED) // 0x8011045BL
			A(COMADMIN_E_AMBIGUOUS_APPLICATION_NAME) // 0x8011045CL
			A(COMADMIN_E_AMBIGUOUS_PARTITION_NAME) // 0x8011045DL
			A(COMADMIN_E_REGDB_NOTINITIALIZED) // 0x80110472L
			A(COMADMIN_E_REGDB_NOTOPEN) // 0x80110473L
			A(COMADMIN_E_REGDB_SYSTEMERR) // 0x80110474L
			A(COMADMIN_E_REGDB_ALREADYRUNNING) // 0x80110475L
			A(COMADMIN_E_MIG_VERSIONNOTSUPPORTED) // 0x80110480L
			A(COMADMIN_E_MIG_SCHEMANOTFOUND) // 0x80110481L
			A(COMADMIN_E_CAT_BITNESSMISMATCH) // 0x80110482L
			A(COMADMIN_E_CAT_UNACCEPTABLEBITNESS) // 0x80110483L
			A(COMADMIN_E_CAT_WRONGAPPBITNESS) // 0x80110484L
			A(COMADMIN_E_CAT_PAUSE_RESUME_NOT_SUPPORTED) // 0x80110485L
			A(COMADMIN_E_CAT_SERVERFAULT) // 0x80110486L
			A(COMQC_E_APPLICATION_NOT_QUEUED) // 0x80110600L
			A(COMQC_E_NO_QUEUEABLE_INTERFACES) // 0x80110601L
			A(COMQC_E_QUEUING_SERVICE_NOT_AVAILABLE) // 0x80110602L
			A(COMQC_E_NO_IPERSISTSTREAM) // 0x80110603L
			A(COMQC_E_BAD_MESSAGE) // 0x80110604L
			A(COMQC_E_UNAUTHENTICATED) // 0x80110605L
			A(COMQC_E_UNTRUSTED_ENQUEUER) // 0x80110606L
			A(MSDTC_E_DUPLICATE_RESOURCE) // 0x80110701L
			A(COMADMIN_E_OBJECT_PARENT_MISSING) // 0x80110808L
			A(COMADMIN_E_OBJECT_DOES_NOT_EXIST) // 0x80110809L
			A(COMADMIN_E_APP_NOT_RUNNING) // 0x8011080AL
			A(COMADMIN_E_INVALID_PARTITION) // 0x8011080BL
			A(COMADMIN_E_SVCAPP_NOT_POOLABLE_OR_RECYCLABLE) // 0x8011080DL
			A(COMADMIN_E_USER_IN_SET) // 0x8011080EL
			A(COMADMIN_E_CANTRECYCLELIBRARYAPPS) // 0x8011080FL
			A(COMADMIN_E_CANTRECYCLESERVICEAPPS) // 0x80110811L
			A(COMADMIN_E_PROCESSALREADYRECYCLED) // 0x80110812L
			A(COMADMIN_E_PAUSEDPROCESSMAYNOTBERECYCLED) // 0x80110813L
			A(COMADMIN_E_CANTMAKEINPROCSERVICE) // 0x80110814L
			A(COMADMIN_E_PROGIDINUSEBYCLSID) // 0x80110815L
			A(COMADMIN_E_DEFAULT_PARTITION_NOT_IN_SET) // 0x80110816L
			A(COMADMIN_E_RECYCLEDPROCESSMAYNOTBEPAUSED) // 0x80110817L
			A(COMADMIN_E_PARTITION_ACCESSDENIED) // 0x80110818L
			A(COMADMIN_E_PARTITION_MSI_ONLY) // 0x80110819L
			A(COMADMIN_E_LEGACYCOMPS_NOT_ALLOWED_IN_1_0_FORMAT) // 0x8011081AL
			A(COMADMIN_E_LEGACYCOMPS_NOT_ALLOWED_IN_NONBASE_PARTITIONS) // 0x8011081BL
			A(COMADMIN_E_COMP_MOVE_SOURCE) // 0x8011081CL
			A(COMADMIN_E_COMP_MOVE_DEST) // 0x8011081DL
			A(COMADMIN_E_COMP_MOVE_PRIVATE) // 0x8011081EL
			A(COMADMIN_E_BASEPARTITION_REQUIRED_IN_SET) // 0x8011081FL
			A(COMADMIN_E_CANNOT_ALIAS_EVENTCLASS) // 0x80110820L
			A(COMADMIN_E_PRIVATE_ACCESSDENIED) // 0x80110821L
			A(COMADMIN_E_SAFERINVALID) // 0x80110822L
			A(COMADMIN_E_REGISTRY_ACCESSDENIED) // 0x80110823L
			A(COMADMIN_E_PARTITIONS_DISABLED) // 0x80110824L
			A(ERROR_FLT_IO_COMPLETE) // 0x001F0001L
			A(ERROR_FLT_NO_HANDLER_DEFINED) // 0x801F0001L
			A(ERROR_FLT_CONTEXT_ALREADY_DEFINED) // 0x801F0002L
			A(ERROR_FLT_INVALID_ASYNCHRONOUS_REQUEST) // 0x801F0003L
			A(ERROR_FLT_DISALLOW_FAST_IO) // 0x801F0004L
			A(ERROR_FLT_INVALID_NAME_REQUEST) // 0x801F0005L
			A(ERROR_FLT_NOT_SAFE_TO_POST_OPERATION) // 0x801F0006L
			A(ERROR_FLT_NOT_INITIALIZED) // 0x801F0007L
			A(ERROR_FLT_FILTER_NOT_READY) // 0x801F0008L
			A(ERROR_FLT_POST_OPERATION_CLEANUP) // 0x801F0009L
			A(ERROR_FLT_INTERNAL_ERROR) // 0x801F000AL
			A(ERROR_FLT_DELETING_OBJECT) // 0x801F000BL
			A(ERROR_FLT_MUST_BE_NONPAGED_POOL) // 0x801F000CL
			A(ERROR_FLT_DUPLICATE_ENTRY) // 0x801F000DL
			A(ERROR_FLT_CBDQ_DISABLED) // 0x801F000EL
			A(ERROR_FLT_DO_NOT_ATTACH) // 0x801F000FL
			A(ERROR_FLT_DO_NOT_DETACH) // 0x801F0010L
			A(ERROR_FLT_INSTANCE_ALTITUDE_COLLISION) // 0x801F0011L
			A(ERROR_FLT_INSTANCE_NAME_COLLISION) // 0x801F0012L
			A(ERROR_FLT_FILTER_NOT_FOUND) // 0x801F0013L
			A(ERROR_FLT_VOLUME_NOT_FOUND) // 0x801F0014L
			A(ERROR_FLT_INSTANCE_NOT_FOUND) // 0x801F0015L
			A(ERROR_FLT_CONTEXT_ALLOCATION_NOT_FOUND) // 0x801F0016L
			A(ERROR_FLT_INVALID_CONTEXT_REGISTRATION) // 0x801F0017L
			A(ERROR_FLT_NAME_CACHE_MISS) // 0x801F0018L
			A(ERROR_FLT_NO_DEVICE_OBJECT) // 0x801F0019L
			A(ERROR_FLT_VOLUME_ALREADY_MOUNTED) // 0x801F001AL
			A(ERROR_FLT_ALREADY_ENLISTED) // 0x801F001BL
			A(ERROR_FLT_CONTEXT_ALREADY_LINKED) // 0x801F001CL
			A(ERROR_FLT_NO_WAITER_FOR_REPLY) // 0x801F0020L
			A(ERROR_HUNG_DISPLAY_DRIVER_THREAD) // 0x80260001L
			A(DWM_E_COMPOSITIONDISABLED) // 0x80263001L
			A(DWM_E_REMOTING_NOT_SUPPORTED) // 0x80263002L
			A(DWM_E_NO_REDIRECTION_SURFACE_AVAILABLE) // 0x80263003L
			A(DWM_E_NOT_QUEUING_PRESENTS) // 0x80263004L
			A(DWM_E_ADAPTER_NOT_FOUND) // 0x80263005L
			A(DWM_S_GDI_REDIRECTION_SURFACE) // 0x00263005L
			A(ERROR_MONITOR_NO_DESCRIPTOR) // 0x00261001L
			A(ERROR_MONITOR_UNKNOWN_DESCRIPTOR_FORMAT) // 0x00261002L
			A(ERROR_MONITOR_INVALID_DESCRIPTOR_CHECKSUM) // 0xC0261003L
			A(ERROR_MONITOR_INVALID_STANDARD_TIMING_BLOCK) // 0xC0261004L
			A(ERROR_MONITOR_WMI_DATABLOCK_REGISTRATION_FAILED) // 0xC0261005L
			A(ERROR_MONITOR_INVALID_SERIAL_NUMBER_MONDSC_BLOCK) // 0xC0261006L
			A(ERROR_MONITOR_INVALID_USER_FRIENDLY_MONDSC_BLOCK) // 0xC0261007L
			A(ERROR_MONITOR_NO_MORE_DESCRIPTOR_DATA) // 0xC0261008L
			A(ERROR_MONITOR_INVALID_DETAILED_TIMING_BLOCK) // 0xC0261009L
			A(ERROR_MONITOR_INVALID_MANUFACTURE_DATE) // 0xC026100AL
			A(ERROR_GRAPHICS_NOT_EXCLUSIVE_MODE_OWNER) // 0xC0262000L
			A(ERROR_GRAPHICS_INSUFFICIENT_DMA_BUFFER) // 0xC0262001L
			A(ERROR_GRAPHICS_INVALID_DISPLAY_ADAPTER) // 0xC0262002L
			A(ERROR_GRAPHICS_ADAPTER_WAS_RESET) // 0xC0262003L
			A(ERROR_GRAPHICS_INVALID_DRIVER_MODEL) // 0xC0262004L
			A(ERROR_GRAPHICS_PRESENT_MODE_CHANGED) // 0xC0262005L
			A(ERROR_GRAPHICS_PRESENT_OCCLUDED) // 0xC0262006L
			A(ERROR_GRAPHICS_PRESENT_DENIED) // 0xC0262007L
			A(ERROR_GRAPHICS_CANNOTCOLORCONVERT) // 0xC0262008L
			A(ERROR_GRAPHICS_DRIVER_MISMATCH) // 0xC0262009L
			A(ERROR_GRAPHICS_PARTIAL_DATA_POPULATED) // 0x4026200AL
			A(ERROR_GRAPHICS_PRESENT_REDIRECTION_DISABLED) // 0xC026200BL
			A(ERROR_GRAPHICS_PRESENT_UNOCCLUDED) // 0xC026200CL
			A(ERROR_GRAPHICS_NO_VIDEO_MEMORY) // 0xC0262100L
			A(ERROR_GRAPHICS_CANT_LOCK_MEMORY) // 0xC0262101L
			A(ERROR_GRAPHICS_ALLOCATION_BUSY) // 0xC0262102L
			A(ERROR_GRAPHICS_TOO_MANY_REFERENCES) // 0xC0262103L
			A(ERROR_GRAPHICS_TRY_AGAIN_LATER) // 0xC0262104L
			A(ERROR_GRAPHICS_TRY_AGAIN_NOW) // 0xC0262105L
			A(ERROR_GRAPHICS_ALLOCATION_INVALID) // 0xC0262106L
			A(ERROR_GRAPHICS_UNSWIZZLING_APERTURE_UNAVAILABLE) // 0xC0262107L
			A(ERROR_GRAPHICS_UNSWIZZLING_APERTURE_UNSUPPORTED) // 0xC0262108L
			A(ERROR_GRAPHICS_CANT_EVICT_PINNED_ALLOCATION) // 0xC0262109L
			A(ERROR_GRAPHICS_INVALID_ALLOCATION_USAGE) // 0xC0262110L
			A(ERROR_GRAPHICS_CANT_RENDER_LOCKED_ALLOCATION) // 0xC0262111L
			A(ERROR_GRAPHICS_ALLOCATION_CLOSED) // 0xC0262112L
			A(ERROR_GRAPHICS_INVALID_ALLOCATION_INSTANCE) // 0xC0262113L
			A(ERROR_GRAPHICS_INVALID_ALLOCATION_HANDLE) // 0xC0262114L
			A(ERROR_GRAPHICS_WRONG_ALLOCATION_DEVICE) // 0xC0262115L
			A(ERROR_GRAPHICS_ALLOCATION_CONTENT_LOST) // 0xC0262116L
			A(ERROR_GRAPHICS_GPU_EXCEPTION_ON_DEVICE) // 0xC0262200L
			A(ERROR_GRAPHICS_INVALID_VIDPN_TOPOLOGY) // 0xC0262300L
			A(ERROR_GRAPHICS_VIDPN_TOPOLOGY_NOT_SUPPORTED) // 0xC0262301L
			A(ERROR_GRAPHICS_VIDPN_TOPOLOGY_CURRENTLY_NOT_SUPPORTED) // 0xC0262302L
			A(ERROR_GRAPHICS_INVALID_VIDPN) // 0xC0262303L
			A(ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE) // 0xC0262304L
			A(ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET) // 0xC0262305L
			A(ERROR_GRAPHICS_VIDPN_MODALITY_NOT_SUPPORTED) // 0xC0262306L
			A(ERROR_GRAPHICS_MODE_NOT_PINNED) // 0x00262307L
			A(ERROR_GRAPHICS_INVALID_VIDPN_SOURCEMODESET) // 0xC0262308L
			A(ERROR_GRAPHICS_INVALID_VIDPN_TARGETMODESET) // 0xC0262309L
			A(ERROR_GRAPHICS_INVALID_FREQUENCY) // 0xC026230AL
			A(ERROR_GRAPHICS_INVALID_ACTIVE_REGION) // 0xC026230BL
			A(ERROR_GRAPHICS_INVALID_TOTAL_REGION) // 0xC026230CL
			A(ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_SOURCE_MODE) // 0xC0262310L
			A(ERROR_GRAPHICS_INVALID_VIDEO_PRESENT_TARGET_MODE) // 0xC0262311L
			A(ERROR_GRAPHICS_PINNED_MODE_MUST_REMAIN_IN_SET) // 0xC0262312L
			A(ERROR_GRAPHICS_PATH_ALREADY_IN_TOPOLOGY) // 0xC0262313L
			A(ERROR_GRAPHICS_MODE_ALREADY_IN_MODESET) // 0xC0262314L
			A(ERROR_GRAPHICS_INVALID_VIDEOPRESENTSOURCESET) // 0xC0262315L
			A(ERROR_GRAPHICS_INVALID_VIDEOPRESENTTARGETSET) // 0xC0262316L
			A(ERROR_GRAPHICS_SOURCE_ALREADY_IN_SET) // 0xC0262317L
			A(ERROR_GRAPHICS_TARGET_ALREADY_IN_SET) // 0xC0262318L
			A(ERROR_GRAPHICS_INVALID_VIDPN_PRESENT_PATH) // 0xC0262319L
			A(ERROR_GRAPHICS_NO_RECOMMENDED_VIDPN_TOPOLOGY) // 0xC026231AL
			A(ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGESET) // 0xC026231BL
			A(ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE) // 0xC026231CL
			A(ERROR_GRAPHICS_FREQUENCYRANGE_NOT_IN_SET) // 0xC026231DL
			A(ERROR_GRAPHICS_NO_PREFERRED_MODE) // 0x0026231EL
			A(ERROR_GRAPHICS_FREQUENCYRANGE_ALREADY_IN_SET) // 0xC026231FL
			A(ERROR_GRAPHICS_STALE_MODESET) // 0xC0262320L
			A(ERROR_GRAPHICS_INVALID_MONITOR_SOURCEMODESET) // 0xC0262321L
			A(ERROR_GRAPHICS_INVALID_MONITOR_SOURCE_MODE) // 0xC0262322L
			A(ERROR_GRAPHICS_NO_RECOMMENDED_FUNCTIONAL_VIDPN) // 0xC0262323L
			A(ERROR_GRAPHICS_MODE_ID_MUST_BE_UNIQUE) // 0xC0262324L
			A(ERROR_GRAPHICS_EMPTY_ADAPTER_MONITOR_MODE_SUPPORT_INTERSECTION) // 0xC0262325L
			A(ERROR_GRAPHICS_VIDEO_PRESENT_TARGETS_LESS_THAN_SOURCES) // 0xC0262326L
			A(ERROR_GRAPHICS_PATH_NOT_IN_TOPOLOGY) // 0xC0262327L
			A(ERROR_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_SOURCE) // 0xC0262328L
			A(ERROR_GRAPHICS_ADAPTER_MUST_HAVE_AT_LEAST_ONE_TARGET) // 0xC0262329L
			A(ERROR_GRAPHICS_INVALID_MONITORDESCRIPTORSET) // 0xC026232AL
			A(ERROR_GRAPHICS_INVALID_MONITORDESCRIPTOR) // 0xC026232BL
			A(ERROR_GRAPHICS_MONITORDESCRIPTOR_NOT_IN_SET) // 0xC026232CL
			A(ERROR_GRAPHICS_MONITORDESCRIPTOR_ALREADY_IN_SET) // 0xC026232DL
			A(ERROR_GRAPHICS_MONITORDESCRIPTOR_ID_MUST_BE_UNIQUE) // 0xC026232EL
			A(ERROR_GRAPHICS_INVALID_VIDPN_TARGET_SUBSET_TYPE) // 0xC026232FL
			A(ERROR_GRAPHICS_RESOURCES_NOT_RELATED) // 0xC0262330L
			A(ERROR_GRAPHICS_SOURCE_ID_MUST_BE_UNIQUE) // 0xC0262331L
			A(ERROR_GRAPHICS_TARGET_ID_MUST_BE_UNIQUE) // 0xC0262332L
			A(ERROR_GRAPHICS_NO_AVAILABLE_VIDPN_TARGET) // 0xC0262333L
			A(ERROR_GRAPHICS_MONITOR_COULD_NOT_BE_ASSOCIATED_WITH_ADAPTER) // 0xC0262334L
			A(ERROR_GRAPHICS_NO_VIDPNMGR) // 0xC0262335L
			A(ERROR_GRAPHICS_NO_ACTIVE_VIDPN) // 0xC0262336L
			A(ERROR_GRAPHICS_STALE_VIDPN_TOPOLOGY) // 0xC0262337L
			A(ERROR_GRAPHICS_MONITOR_NOT_CONNECTED) // 0xC0262338L
			A(ERROR_GRAPHICS_SOURCE_NOT_IN_TOPOLOGY) // 0xC0262339L
			A(ERROR_GRAPHICS_INVALID_PRIMARYSURFACE_SIZE) // 0xC026233AL
			A(ERROR_GRAPHICS_INVALID_VISIBLEREGION_SIZE) // 0xC026233BL
			A(ERROR_GRAPHICS_INVALID_STRIDE) // 0xC026233CL
			A(ERROR_GRAPHICS_INVALID_PIXELFORMAT) // 0xC026233DL
			A(ERROR_GRAPHICS_INVALID_COLORBASIS) // 0xC026233EL
			A(ERROR_GRAPHICS_INVALID_PIXELVALUEACCESSMODE) // 0xC026233FL
			A(ERROR_GRAPHICS_TARGET_NOT_IN_TOPOLOGY) // 0xC0262340L
			A(ERROR_GRAPHICS_NO_DISPLAY_MODE_MANAGEMENT_SUPPORT) // 0xC0262341L
			A(ERROR_GRAPHICS_VIDPN_SOURCE_IN_USE) // 0xC0262342L
			A(ERROR_GRAPHICS_CANT_ACCESS_ACTIVE_VIDPN) // 0xC0262343L
			A(ERROR_GRAPHICS_INVALID_PATH_IMPORTANCE_ORDINAL) // 0xC0262344L
			A(ERROR_GRAPHICS_INVALID_PATH_CONTENT_GEOMETRY_TRANSFORMATION) // 0xC0262345L
			A(ERROR_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_SUPPORTED) // 0xC0262346L
			A(ERROR_GRAPHICS_INVALID_GAMMA_RAMP) // 0xC0262347L
			A(ERROR_GRAPHICS_GAMMA_RAMP_NOT_SUPPORTED) // 0xC0262348L
			A(ERROR_GRAPHICS_MULTISAMPLING_NOT_SUPPORTED) // 0xC0262349L
			A(ERROR_GRAPHICS_MODE_NOT_IN_MODESET) // 0xC026234AL
			A(ERROR_GRAPHICS_DATASET_IS_EMPTY) // 0x0026234BL
			A(ERROR_GRAPHICS_NO_MORE_ELEMENTS_IN_DATASET) // 0x0026234CL
			A(ERROR_GRAPHICS_INVALID_VIDPN_TOPOLOGY_RECOMMENDATION_REASON) // 0xC026234DL
			A(ERROR_GRAPHICS_INVALID_PATH_CONTENT_TYPE) // 0xC026234EL
			A(ERROR_GRAPHICS_INVALID_COPYPROTECTION_TYPE) // 0xC026234FL
			A(ERROR_GRAPHICS_UNASSIGNED_MODESET_ALREADY_EXISTS) // 0xC0262350L
			A(ERROR_GRAPHICS_PATH_CONTENT_GEOMETRY_TRANSFORMATION_NOT_PINNED) // 0x00262351L
			A(ERROR_GRAPHICS_INVALID_SCANLINE_ORDERING) // 0xC0262352L
			A(ERROR_GRAPHICS_TOPOLOGY_CHANGES_NOT_ALLOWED) // 0xC0262353L
			A(ERROR_GRAPHICS_NO_AVAILABLE_IMPORTANCE_ORDINALS) // 0xC0262354L
			A(ERROR_GRAPHICS_INCOMPATIBLE_PRIVATE_FORMAT) // 0xC0262355L
			A(ERROR_GRAPHICS_INVALID_MODE_PRUNING_ALGORITHM) // 0xC0262356L
			A(ERROR_GRAPHICS_INVALID_MONITOR_CAPABILITY_ORIGIN) // 0xC0262357L
			A(ERROR_GRAPHICS_INVALID_MONITOR_FREQUENCYRANGE_CONSTRAINT) // 0xC0262358L
			A(ERROR_GRAPHICS_MAX_NUM_PATHS_REACHED) // 0xC0262359L
			A(ERROR_GRAPHICS_CANCEL_VIDPN_TOPOLOGY_AUGMENTATION) // 0xC026235AL
			A(ERROR_GRAPHICS_INVALID_CLIENT_TYPE) // 0xC026235BL
			A(ERROR_GRAPHICS_CLIENTVIDPN_NOT_SET) // 0xC026235CL
			A(ERROR_GRAPHICS_SPECIFIED_CHILD_ALREADY_CONNECTED) // 0xC0262400L    
			A(ERROR_GRAPHICS_CHILD_DESCRIPTOR_NOT_SUPPORTED) // 0xC0262401L    
			A(ERROR_GRAPHICS_UNKNOWN_CHILD_STATUS) // 0x4026242FL
			A(ERROR_GRAPHICS_NOT_A_LINKED_ADAPTER) // 0xC0262430L
			A(ERROR_GRAPHICS_LEADLINK_NOT_ENUMERATED) // 0xC0262431L
			A(ERROR_GRAPHICS_CHAINLINKS_NOT_ENUMERATED) // 0xC0262432L
			A(ERROR_GRAPHICS_ADAPTER_CHAIN_NOT_READY) // 0xC0262433L
			A(ERROR_GRAPHICS_CHAINLINKS_NOT_STARTED) // 0xC0262434L
			A(ERROR_GRAPHICS_CHAINLINKS_NOT_POWERED_ON) // 0xC0262435L
			A(ERROR_GRAPHICS_INCONSISTENT_DEVICE_LINK_STATE) // 0xC0262436L
			A(ERROR_GRAPHICS_LEADLINK_START_DEFERRED) // 0x40262437L
			A(ERROR_GRAPHICS_NOT_POST_DEVICE_DRIVER) // 0xC0262438L
			A(ERROR_GRAPHICS_POLLING_TOO_FREQUENTLY) // 0x40262439L
			A(ERROR_GRAPHICS_START_DEFERRED) // 0x4026243AL
			A(ERROR_GRAPHICS_ADAPTER_ACCESS_NOT_EXCLUDED) // 0xC026243BL
			A(ERROR_GRAPHICS_OPM_NOT_SUPPORTED) // 0xC0262500L    
			A(ERROR_GRAPHICS_COPP_NOT_SUPPORTED) // 0xC0262501L    
			A(ERROR_GRAPHICS_UAB_NOT_SUPPORTED) // 0xC0262502L    
			A(ERROR_GRAPHICS_OPM_INVALID_ENCRYPTED_PARAMETERS) // 0xC0262503L    
			A(ERROR_GRAPHICS_OPM_NO_VIDEO_OUTPUTS_EXIST) // 0xC0262505L
			A(ERROR_GRAPHICS_OPM_INTERNAL_ERROR) // 0xC026250BL
			A(ERROR_GRAPHICS_OPM_INVALID_HANDLE) // 0xC026250CL
			A(ERROR_GRAPHICS_PVP_INVALID_CERTIFICATE_LENGTH) // 0xC026250EL
			A(ERROR_GRAPHICS_OPM_SPANNING_MODE_ENABLED) // 0xC026250FL
			A(ERROR_GRAPHICS_OPM_THEATER_MODE_ENABLED) // 0xC0262510L
			A(ERROR_GRAPHICS_PVP_HFS_FAILED) // 0xC0262511L
			A(ERROR_GRAPHICS_OPM_INVALID_SRM) // 0xC0262512L
			A(ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_HDCP) // 0xC0262513L
			A(ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_ACP) // 0xC0262514L
			A(ERROR_GRAPHICS_OPM_OUTPUT_DOES_NOT_SUPPORT_CGMSA) // 0xC0262515L
			A(ERROR_GRAPHICS_OPM_HDCP_SRM_NEVER_SET) // 0xC0262516L
			A(ERROR_GRAPHICS_OPM_RESOLUTION_TOO_HIGH) // 0xC0262517L
			A(ERROR_GRAPHICS_OPM_ALL_HDCP_HARDWARE_ALREADY_IN_USE) // 0xC0262518L
			A(ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_NO_LONGER_EXISTS) // 0xC026251AL
			A(ERROR_GRAPHICS_OPM_SESSION_TYPE_CHANGE_IN_PROGRESS) // 0xC026251BL
			A(ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_DOES_NOT_HAVE_COPP_SEMANTICS) // 0xC026251CL
			A(ERROR_GRAPHICS_OPM_INVALID_INFORMATION_REQUEST) // 0xC026251DL
			A(ERROR_GRAPHICS_OPM_DRIVER_INTERNAL_ERROR) // 0xC026251EL
			A(ERROR_GRAPHICS_OPM_VIDEO_OUTPUT_DOES_NOT_HAVE_OPM_SEMANTICS) // 0xC026251FL
			A(ERROR_GRAPHICS_OPM_SIGNALING_NOT_SUPPORTED) // 0xC0262520L
			A(ERROR_GRAPHICS_OPM_INVALID_CONFIGURATION_REQUEST) // 0xC0262521L
			A(ERROR_GRAPHICS_I2C_NOT_SUPPORTED) // 0xC0262580L    
			A(ERROR_GRAPHICS_I2C_DEVICE_DOES_NOT_EXIST) // 0xC0262581L    
			A(ERROR_GRAPHICS_I2C_ERROR_TRANSMITTING_DATA) // 0xC0262582L    
			A(ERROR_GRAPHICS_I2C_ERROR_RECEIVING_DATA) // 0xC0262583L    
			A(ERROR_GRAPHICS_DDCCI_VCP_NOT_SUPPORTED) // 0xC0262584L    
			A(ERROR_GRAPHICS_DDCCI_INVALID_DATA) // 0xC0262585L    
			A(ERROR_GRAPHICS_DDCCI_MONITOR_RETURNED_INVALID_TIMING_STATUS_BYTE) // 0xC0262586L
			A(ERROR_GRAPHICS_MCA_INVALID_CAPABILITIES_STRING) // 0xC0262587L
			A(ERROR_GRAPHICS_MCA_INTERNAL_ERROR) // 0xC0262588L
			A(ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_COMMAND) // 0xC0262589L
			A(ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_LENGTH) // 0xC026258AL
			A(ERROR_GRAPHICS_DDCCI_INVALID_MESSAGE_CHECKSUM) // 0xC026258BL
			A(ERROR_GRAPHICS_INVALID_PHYSICAL_MONITOR_HANDLE) // 0xC026258CL
			A(ERROR_GRAPHICS_MONITOR_NO_LONGER_EXISTS) // 0xC026258DL
			A(ERROR_GRAPHICS_DDCCI_CURRENT_CURRENT_VALUE_GREATER_THAN_MAXIMUM_VALUE) // 0xC02625D8L
			A(ERROR_GRAPHICS_MCA_INVALID_VCP_VERSION) // 0xC02625D9L
			A(ERROR_GRAPHICS_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION) // 0xC02625DAL
			A(ERROR_GRAPHICS_MCA_MCCS_VERSION_MISMATCH) // 0xC02625DBL
			A(ERROR_GRAPHICS_MCA_UNSUPPORTED_MCCS_VERSION) // 0xC02625DCL
			A(ERROR_GRAPHICS_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED) // 0xC02625DEL
			A(ERROR_GRAPHICS_MCA_UNSUPPORTED_COLOR_TEMPERATURE) // 0xC02625DFL
			A(ERROR_GRAPHICS_ONLY_CONSOLE_SESSION_SUPPORTED) // 0xC02625E0L
			A(ERROR_GRAPHICS_NO_DISPLAY_DEVICE_CORRESPONDS_TO_NAME) // 0xC02625E1L
			A(ERROR_GRAPHICS_DISPLAY_DEVICE_NOT_ATTACHED_TO_DESKTOP) // 0xC02625E2L
			A(ERROR_GRAPHICS_MIRRORING_DEVICES_NOT_SUPPORTED) // 0xC02625E3L
			A(ERROR_GRAPHICS_INVALID_POINTER) // 0xC02625E4L
			A(ERROR_GRAPHICS_NO_MONITORS_CORRESPOND_TO_DISPLAY_DEVICE) // 0xC02625E5L
			A(ERROR_GRAPHICS_PARAMETER_ARRAY_TOO_SMALL) // 0xC02625E6L
			A(ERROR_GRAPHICS_INTERNAL_ERROR) // 0xC02625E7L
			A(ERROR_GRAPHICS_SESSION_TYPE_CHANGE_IN_PROGRESS) // 0xC02605E8L
			A(TPM_E_ERROR_MASK) // 0x80280000L
			A(TPM_E_AUTHFAIL) // 0x80280001L
			A(TPM_E_BADINDEX) // 0x80280002L
			A(TPM_E_BAD_PARAMETER) // 0x80280003L
			A(TPM_E_AUDITFAILURE) // 0x80280004L
			A(TPM_E_CLEAR_DISABLED) // 0x80280005L
			A(TPM_E_DEACTIVATED) // 0x80280006L
			A(TPM_E_DISABLED) // 0x80280007L
			A(TPM_E_DISABLED_CMD) // 0x80280008L
			A(TPM_E_FAIL) // 0x80280009L
			A(TPM_E_BAD_ORDINAL) // 0x8028000AL
			A(TPM_E_INSTALL_DISABLED) // 0x8028000BL
			A(TPM_E_INVALID_KEYHANDLE) // 0x8028000CL
			A(TPM_E_KEYNOTFOUND) // 0x8028000DL
			A(TPM_E_INAPPROPRIATE_ENC) // 0x8028000EL
			A(TPM_E_MIGRATEFAIL) // 0x8028000FL
			A(TPM_E_INVALID_PCR_INFO) // 0x80280010L
			A(TPM_E_NOSPACE) // 0x80280011L
			A(TPM_E_NOSRK) // 0x80280012L
			A(TPM_E_NOTSEALED_BLOB) // 0x80280013L
			A(TPM_E_OWNER_SET) // 0x80280014L
			A(TPM_E_RESOURCES) // 0x80280015L
			A(TPM_E_SHORTRANDOM) // 0x80280016L
			A(TPM_E_SIZE) // 0x80280017L
			A(TPM_E_WRONGPCRVAL) // 0x80280018L
			A(TPM_E_BAD_PARAM_SIZE) // 0x80280019L
			A(TPM_E_SHA_THREAD) // 0x8028001AL
			A(TPM_E_SHA_ERROR) // 0x8028001BL
			A(TPM_E_FAILEDSELFTEST) // 0x8028001CL
			A(TPM_E_AUTH2FAIL) // 0x8028001DL
			A(TPM_E_BADTAG) // 0x8028001EL
			A(TPM_E_IOERROR) // 0x8028001FL
			A(TPM_E_ENCRYPT_ERROR) // 0x80280020L
			A(TPM_E_DECRYPT_ERROR) // 0x80280021L
			A(TPM_E_INVALID_AUTHHANDLE) // 0x80280022L
			A(TPM_E_NO_ENDORSEMENT) // 0x80280023L
			A(TPM_E_INVALID_KEYUSAGE) // 0x80280024L
			A(TPM_E_WRONG_ENTITYTYPE) // 0x80280025L
			A(TPM_E_INVALID_POSTINIT) // 0x80280026L
			A(TPM_E_INAPPROPRIATE_SIG) // 0x80280027L
			A(TPM_E_BAD_KEY_PROPERTY) // 0x80280028L
			A(TPM_E_BAD_MIGRATION) // 0x80280029L
			A(TPM_E_BAD_SCHEME) // 0x8028002AL
			A(TPM_E_BAD_DATASIZE) // 0x8028002BL
			A(TPM_E_BAD_MODE) // 0x8028002CL
			A(TPM_E_BAD_PRESENCE) // 0x8028002DL
			A(TPM_E_BAD_VERSION) // 0x8028002EL
			A(TPM_E_NO_WRAP_TRANSPORT) // 0x8028002FL
			A(TPM_E_AUDITFAIL_UNSUCCESSFUL) // 0x80280030L
			A(TPM_E_AUDITFAIL_SUCCESSFUL) // 0x80280031L
			A(TPM_E_NOTRESETABLE) // 0x80280032L
			A(TPM_E_NOTLOCAL) // 0x80280033L
			A(TPM_E_BAD_TYPE) // 0x80280034L
			A(TPM_E_INVALID_RESOURCE) // 0x80280035L
			A(TPM_E_NOTFIPS) // 0x80280036L
			A(TPM_E_INVALID_FAMILY) // 0x80280037L
			A(TPM_E_NO_NV_PERMISSION) // 0x80280038L
			A(TPM_E_REQUIRES_SIGN) // 0x80280039L
			A(TPM_E_KEY_NOTSUPPORTED) // 0x8028003AL
			A(TPM_E_AUTH_CONFLICT) // 0x8028003BL
			A(TPM_E_AREA_LOCKED) // 0x8028003CL
			A(TPM_E_BAD_LOCALITY) // 0x8028003DL
			A(TPM_E_READ_ONLY) // 0x8028003EL
			A(TPM_E_PER_NOWRITE) // 0x8028003FL
			A(TPM_E_FAMILYCOUNT) // 0x80280040L
			A(TPM_E_WRITE_LOCKED) // 0x80280041L
			A(TPM_E_BAD_ATTRIBUTES) // 0x80280042L
			A(TPM_E_INVALID_STRUCTURE) // 0x80280043L
			A(TPM_E_KEY_OWNER_CONTROL) // 0x80280044L
			A(TPM_E_BAD_COUNTER) // 0x80280045L
			A(TPM_E_NOT_FULLWRITE) // 0x80280046L
			A(TPM_E_CONTEXT_GAP) // 0x80280047L
			A(TPM_E_MAXNVWRITES) // 0x80280048L
			A(TPM_E_NOOPERATOR) // 0x80280049L
			A(TPM_E_RESOURCEMISSING) // 0x8028004AL
			A(TPM_E_DELEGATE_LOCK) // 0x8028004BL
			A(TPM_E_DELEGATE_FAMILY) // 0x8028004CL
			A(TPM_E_DELEGATE_ADMIN) // 0x8028004DL
			A(TPM_E_TRANSPORT_NOTEXCLUSIVE) // 0x8028004EL
			A(TPM_E_OWNER_CONTROL) // 0x8028004FL
			A(TPM_E_DAA_RESOURCES) // 0x80280050L
			A(TPM_E_DAA_INPUT_DATA0) // 0x80280051L
			A(TPM_E_DAA_INPUT_DATA1) // 0x80280052L
			A(TPM_E_DAA_ISSUER_SETTINGS) // 0x80280053L
			A(TPM_E_DAA_TPM_SETTINGS) // 0x80280054L
			A(TPM_E_DAA_STAGE) // 0x80280055L
			A(TPM_E_DAA_ISSUER_VALIDITY) // 0x80280056L
			A(TPM_E_DAA_WRONG_W) // 0x80280057L
			A(TPM_E_BAD_HANDLE) // 0x80280058L
			A(TPM_E_BAD_DELEGATE) // 0x80280059L
			A(TPM_E_BADCONTEXT) // 0x8028005AL
			A(TPM_E_TOOMANYCONTEXTS) // 0x8028005BL
			A(TPM_E_MA_TICKET_SIGNATURE) // 0x8028005CL
			A(TPM_E_MA_DESTINATION) // 0x8028005DL
			A(TPM_E_MA_SOURCE) // 0x8028005EL
			A(TPM_E_MA_AUTHORITY) // 0x8028005FL
			A(TPM_E_PERMANENTEK) // 0x80280061L
			A(TPM_E_BAD_SIGNATURE) // 0x80280062L
			A(TPM_E_NOCONTEXTSPACE) // 0x80280063L
			A(TPM_E_COMMAND_BLOCKED) // 0x80280400L
			A(TPM_E_INVALID_HANDLE) // 0x80280401L
			A(TPM_E_DUPLICATE_VHANDLE) // 0x80280402L
			A(TPM_E_EMBEDDED_COMMAND_BLOCKED) // 0x80280403L
			A(TPM_E_EMBEDDED_COMMAND_UNSUPPORTED) // 0x80280404L
			A(TPM_E_RETRY) // 0x80280800L
			A(TPM_E_NEEDS_SELFTEST) // 0x80280801L
			A(TPM_E_DOING_SELFTEST) // 0x80280802L
			A(TPM_E_DEFEND_LOCK_RUNNING) // 0x80280803L
			A(TBS_E_INTERNAL_ERROR) // 0x80284001L
			A(TBS_E_BAD_PARAMETER) // 0x80284002L
			A(TBS_E_INVALID_OUTPUT_POINTER) // 0x80284003L
			A(TBS_E_INVALID_CONTEXT) // 0x80284004L
			A(TBS_E_INSUFFICIENT_BUFFER) // 0x80284005L
			A(TBS_E_IOERROR) // 0x80284006L
			A(TBS_E_INVALID_CONTEXT_PARAM) // 0x80284007L
			A(TBS_E_SERVICE_NOT_RUNNING) // 0x80284008L
			A(TBS_E_TOO_MANY_TBS_CONTEXTS) // 0x80284009L
			A(TBS_E_TOO_MANY_RESOURCES) // 0x8028400AL
			A(TBS_E_SERVICE_START_PENDING) // 0x8028400BL
			A(TBS_E_PPI_NOT_SUPPORTED) // 0x8028400CL
			A(TBS_E_COMMAND_CANCELED) // 0x8028400DL
			A(TBS_E_BUFFER_TOO_LARGE) // 0x8028400EL
			A(TBS_E_TPM_NOT_FOUND) // 0x8028400FL
			A(TBS_E_SERVICE_DISABLED) // 0x80284010L
			A(TBS_E_NO_EVENT_LOG) // 0x80284011L
			A(TPMAPI_E_INVALID_STATE) // 0x80290100L
			A(TPMAPI_E_NOT_ENOUGH_DATA) // 0x80290101L
			A(TPMAPI_E_TOO_MUCH_DATA) // 0x80290102L
			A(TPMAPI_E_INVALID_OUTPUT_POINTER) // 0x80290103L
			A(TPMAPI_E_INVALID_PARAMETER) // 0x80290104L
			A(TPMAPI_E_OUT_OF_MEMORY) // 0x80290105L
			A(TPMAPI_E_BUFFER_TOO_SMALL) // 0x80290106L
			A(TPMAPI_E_INTERNAL_ERROR) // 0x80290107L
			A(TPMAPI_E_ACCESS_DENIED) // 0x80290108L
			A(TPMAPI_E_AUTHORIZATION_FAILED) // 0x80290109L
			A(TPMAPI_E_INVALID_CONTEXT_HANDLE) // 0x8029010AL
			A(TPMAPI_E_TBS_COMMUNICATION_ERROR) // 0x8029010BL
			A(TPMAPI_E_TPM_COMMAND_ERROR) // 0x8029010CL
			A(TPMAPI_E_MESSAGE_TOO_LARGE) // 0x8029010DL
			A(TPMAPI_E_INVALID_ENCODING) // 0x8029010EL
			A(TPMAPI_E_INVALID_KEY_SIZE) // 0x8029010FL
			A(TPMAPI_E_ENCRYPTION_FAILED) // 0x80290110L
			A(TPMAPI_E_INVALID_KEY_PARAMS) // 0x80290111L
			A(TPMAPI_E_INVALID_MIGRATION_AUTHORIZATION_BLOB) // 0x80290112L
			A(TPMAPI_E_INVALID_PCR_INDEX) // 0x80290113L
			A(TPMAPI_E_INVALID_DELEGATE_BLOB) // 0x80290114L
			A(TPMAPI_E_INVALID_CONTEXT_PARAMS) // 0x80290115L
			A(TPMAPI_E_INVALID_KEY_BLOB) // 0x80290116L
			A(TPMAPI_E_INVALID_PCR_DATA) // 0x80290117L
			A(TPMAPI_E_INVALID_OWNER_AUTH) // 0x80290118L
			A(TPMAPI_E_FIPS_RNG_CHECK_FAILED) // 0x80290119L
			A(TPMAPI_E_EMPTY_TCG_LOG) // 0x8029011AL
			A(TPMAPI_E_INVALID_TCG_LOG_ENTRY) // 0x8029011BL
			A(TPMAPI_E_TCG_SEPARATOR_ABSENT) // 0x8029011CL
			A(TPMAPI_E_TCG_INVALID_DIGEST_ENTRY) // 0x8029011DL
			A(TBSIMP_E_BUFFER_TOO_SMALL) // 0x80290200L
			A(TBSIMP_E_CLEANUP_FAILED) // 0x80290201L
			A(TBSIMP_E_INVALID_CONTEXT_HANDLE) // 0x80290202L
			A(TBSIMP_E_INVALID_CONTEXT_PARAM) // 0x80290203L
			A(TBSIMP_E_TPM_ERROR) // 0x80290204L
			A(TBSIMP_E_HASH_BAD_KEY) // 0x80290205L
			A(TBSIMP_E_DUPLICATE_VHANDLE) // 0x80290206L
			A(TBSIMP_E_INVALID_OUTPUT_POINTER) // 0x80290207L
			A(TBSIMP_E_INVALID_PARAMETER) // 0x80290208L
			A(TBSIMP_E_RPC_INIT_FAILED) // 0x80290209L
			A(TBSIMP_E_SCHEDULER_NOT_RUNNING) // 0x8029020AL
			A(TBSIMP_E_COMMAND_CANCELED) // 0x8029020BL
			A(TBSIMP_E_OUT_OF_MEMORY) // 0x8029020CL
			A(TBSIMP_E_LIST_NO_MORE_ITEMS) // 0x8029020DL
			A(TBSIMP_E_LIST_NOT_FOUND) // 0x8029020EL
			A(TBSIMP_E_NOT_ENOUGH_SPACE) // 0x8029020FL
			A(TBSIMP_E_NOT_ENOUGH_TPM_CONTEXTS) // 0x80290210L
			A(TBSIMP_E_COMMAND_FAILED) // 0x80290211L
			A(TBSIMP_E_UNKNOWN_ORDINAL) // 0x80290212L
			A(TBSIMP_E_RESOURCE_EXPIRED) // 0x80290213L
			A(TBSIMP_E_INVALID_RESOURCE) // 0x80290214L
			A(TBSIMP_E_NOTHING_TO_UNLOAD) // 0x80290215L
			A(TBSIMP_E_HASH_TABLE_FULL) // 0x80290216L
			A(TBSIMP_E_TOO_MANY_TBS_CONTEXTS) // 0x80290217L
			A(TBSIMP_E_TOO_MANY_RESOURCES) // 0x80290218L
			A(TBSIMP_E_PPI_NOT_SUPPORTED) // 0x80290219L
			A(TBSIMP_E_TPM_INCOMPATIBLE) // 0x8029021AL
			A(TBSIMP_E_NO_EVENT_LOG) // 0x8029021BL
			A(TPM_E_PPI_ACPI_FAILURE) // 0x80290300L
			A(TPM_E_PPI_USER_ABORT) // 0x80290301L
			A(TPM_E_PPI_BIOS_FAILURE) // 0x80290302L
			A(TPM_E_PPI_NOT_SUPPORTED) // 0x80290303L
			A(PLA_E_DCS_NOT_FOUND) // 0x80300002L
			A(PLA_E_DCS_IN_USE) // 0x803000AAL
			A(PLA_E_TOO_MANY_FOLDERS) // 0x80300045L
			A(PLA_E_NO_MIN_DISK) // 0x80300070L
			A(PLA_E_DCS_ALREADY_EXISTS) // 0x803000B7L
			A(PLA_S_PROPERTY_IGNORED) // 0x00300100L
			A(PLA_E_PROPERTY_CONFLICT) // 0x80300101L
			A(PLA_E_DCS_SINGLETON_REQUIRED) // 0x80300102L
			A(PLA_E_CREDENTIALS_REQUIRED) // 0x80300103L
			A(PLA_E_DCS_NOT_RUNNING) // 0x80300104L
			A(PLA_E_CONFLICT_INCL_EXCL_API) // 0x80300105L
			A(PLA_E_NETWORK_EXE_NOT_VALID) // 0x80300106L
			A(PLA_E_EXE_ALREADY_CONFIGURED) // 0x80300107L
			A(PLA_E_EXE_PATH_NOT_VALID) // 0x80300108L
			A(PLA_E_DC_ALREADY_EXISTS) // 0x80300109L
			A(PLA_E_DCS_START_WAIT_TIMEOUT) // 0x8030010AL
			A(PLA_E_DC_START_WAIT_TIMEOUT) // 0x8030010BL
			A(PLA_E_REPORT_WAIT_TIMEOUT) // 0x8030010CL
			A(PLA_E_NO_DUPLICATES) // 0x8030010DL
			A(PLA_E_EXE_FULL_PATH_REQUIRED) // 0x8030010EL
			A(PLA_E_INVALID_SESSION_NAME) // 0x8030010FL
			A(PLA_E_PLA_CHANNEL_NOT_ENABLED) // 0x80300110L
			A(PLA_E_TASKSCHED_CHANNEL_NOT_ENABLED) // 0x80300111L
			A(PLA_E_RULES_MANAGER_FAILED) // 0x80300112L
			A(PLA_E_CABAPI_FAILURE) // 0x80300113L
			A(FVE_E_LOCKED_VOLUME) // 0x80310000L
			A(FVE_E_NOT_ENCRYPTED) // 0x80310001L
			A(FVE_E_NO_TPM_BIOS) // 0x80310002L
			A(FVE_E_NO_MBR_METRIC) // 0x80310003L
			A(FVE_E_NO_BOOTSECTOR_METRIC) // 0x80310004L
			A(FVE_E_NO_BOOTMGR_METRIC) // 0x80310005L
			A(FVE_E_WRONG_BOOTMGR) // 0x80310006L
			A(FVE_E_SECURE_KEY_REQUIRED) // 0x80310007L
			A(FVE_E_NOT_ACTIVATED) // 0x80310008L
			A(FVE_E_ACTION_NOT_ALLOWED) // 0x80310009L
			A(FVE_E_AD_SCHEMA_NOT_INSTALLED) // 0x8031000AL
			A(FVE_E_AD_INVALID_DATATYPE) // 0x8031000BL
			A(FVE_E_AD_INVALID_DATASIZE) // 0x8031000CL
			A(FVE_E_AD_NO_VALUES) // 0x8031000DL
			A(FVE_E_AD_ATTR_NOT_SET) // 0x8031000EL
			A(FVE_E_AD_GUID_NOT_FOUND) // 0x8031000FL
			A(FVE_E_BAD_INFORMATION) // 0x80310010L
			A(FVE_E_TOO_SMALL) // 0x80310011L
			A(FVE_E_SYSTEM_VOLUME) // 0x80310012L
			A(FVE_E_FAILED_WRONG_FS) // 0x80310013L
			A(FVE_E_BAD_PARTITION_SIZE) // 0x80310014L
			A(FVE_E_NOT_SUPPORTED) // 0x80310015L
			A(FVE_E_BAD_DATA) // 0x80310016L
			A(FVE_E_VOLUME_NOT_BOUND) // 0x80310017L
			A(FVE_E_TPM_NOT_OWNED) // 0x80310018L
			A(FVE_E_NOT_DATA_VOLUME) // 0x80310019L
			A(FVE_E_AD_INSUFFICIENT_BUFFER) // 0x8031001AL
			A(FVE_E_CONV_READ) // 0x8031001BL
			A(FVE_E_CONV_WRITE) // 0x8031001CL
			A(FVE_E_KEY_REQUIRED) // 0x8031001DL
			A(FVE_E_CLUSTERING_NOT_SUPPORTED) // 0x8031001EL
			A(FVE_E_VOLUME_BOUND_ALREADY) // 0x8031001FL
			A(FVE_E_OS_NOT_PROTECTED) // 0x80310020L
			A(FVE_E_PROTECTION_DISABLED) // 0x80310021L
			A(FVE_E_RECOVERY_KEY_REQUIRED) // 0x80310022L
			A(FVE_E_FOREIGN_VOLUME) // 0x80310023L
			A(FVE_E_OVERLAPPED_UPDATE) // 0x80310024L
			A(FVE_E_TPM_SRK_AUTH_NOT_ZERO) // 0x80310025L
			A(FVE_E_FAILED_SECTOR_SIZE) // 0x80310026L
			A(FVE_E_FAILED_AUTHENTICATION) // 0x80310027L
			A(FVE_E_NOT_OS_VOLUME) // 0x80310028L
			A(FVE_E_AUTOUNLOCK_ENABLED) // 0x80310029L
			A(FVE_E_WRONG_BOOTSECTOR) // 0x8031002AL
			A(FVE_E_WRONG_SYSTEM_FS) // 0x8031002BL
			A(FVE_E_POLICY_PASSWORD_REQUIRED) // 0x8031002CL
			A(FVE_E_CANNOT_SET_FVEK_ENCRYPTED) // 0x8031002DL
			A(FVE_E_CANNOT_ENCRYPT_NO_KEY) // 0x8031002EL
			A(FVE_E_BOOTABLE_CDDVD) // 0x80310030L
			A(FVE_E_PROTECTOR_EXISTS) // 0x80310031L
			A(FVE_E_RELATIVE_PATH) // 0x80310032L
			A(FVE_E_PROTECTOR_NOT_FOUND) // 0x80310033L
			A(FVE_E_INVALID_KEY_FORMAT) // 0x80310034L
			A(FVE_E_INVALID_PASSWORD_FORMAT) // 0x80310035L
			A(FVE_E_FIPS_RNG_CHECK_FAILED) // 0x80310036L
			A(FVE_E_FIPS_PREVENTS_RECOVERY_PASSWORD) // 0x80310037L
			A(FVE_E_FIPS_PREVENTS_EXTERNAL_KEY_EXPORT) // 0x80310038L
			A(FVE_E_NOT_DECRYPTED) // 0x80310039L
			A(FVE_E_INVALID_PROTECTOR_TYPE) // 0x8031003AL
			A(FVE_E_NO_PROTECTORS_TO_TEST) // 0x8031003BL
			A(FVE_E_KEYFILE_NOT_FOUND) // 0x8031003CL
			A(FVE_E_KEYFILE_INVALID) // 0x8031003DL
			A(FVE_E_KEYFILE_NO_VMK) // 0x8031003EL
			A(FVE_E_TPM_DISABLED) // 0x8031003FL
			A(FVE_E_NOT_ALLOWED_IN_SAFE_MODE) // 0x80310040L
			A(FVE_E_TPM_INVALID_PCR) // 0x80310041L
			A(FVE_E_TPM_NO_VMK) // 0x80310042L
			A(FVE_E_PIN_INVALID) // 0x80310043L
			A(FVE_E_AUTH_INVALID_APPLICATION) // 0x80310044L
			A(FVE_E_AUTH_INVALID_CONFIG) // 0x80310045L
			A(FVE_E_FIPS_DISABLE_PROTECTION_NOT_ALLOWED) // 0x80310046L
			A(FVE_E_FS_NOT_EXTENDED) // 0x80310047L
			A(FVE_E_FIRMWARE_TYPE_NOT_SUPPORTED) // 0x80310048L
			A(FVE_E_NO_LICENSE) // 0x80310049L
			A(FVE_E_NOT_ON_STACK) // 0x8031004AL
			A(FVE_E_FS_MOUNTED) // 0x8031004BL
			A(FVE_E_TOKEN_NOT_IMPERSONATED) // 0x8031004CL
			A(FVE_E_DRY_RUN_FAILED) // 0x8031004DL
			A(FVE_E_REBOOT_REQUIRED) // 0x8031004EL
			A(FVE_E_DEBUGGER_ENABLED) // 0x8031004FL
			A(FVE_E_RAW_ACCESS) // 0x80310050L
			A(FVE_E_RAW_BLOCKED) // 0x80310051L
			A(FVE_E_BCD_APPLICATIONS_PATH_INCORRECT) // 0x80310052L
			A(FVE_E_NOT_ALLOWED_IN_VERSION) // 0x80310053L
			A(FVE_E_NO_AUTOUNLOCK_MASTER_KEY) // 0x80310054L
			A(FVE_E_MOR_FAILED) // 0x80310055L
			A(FVE_E_HIDDEN_VOLUME) // 0x80310056L
			A(FVE_E_TRANSIENT_STATE) // 0x80310057L
			A(FVE_E_PUBKEY_NOT_ALLOWED) // 0x80310058L
			A(FVE_E_VOLUME_HANDLE_OPEN) // 0x80310059L
			A(FVE_E_NO_FEATURE_LICENSE) // 0x8031005AL
			A(FVE_E_INVALID_STARTUP_OPTIONS) // 0x8031005BL
			A(FVE_E_POLICY_RECOVERY_PASSWORD_NOT_ALLOWED) // 0x8031005CL
			A(FVE_E_POLICY_RECOVERY_PASSWORD_REQUIRED) // 0x8031005DL
			A(FVE_E_POLICY_RECOVERY_KEY_NOT_ALLOWED) // 0x8031005EL
			A(FVE_E_POLICY_RECOVERY_KEY_REQUIRED) // 0x8031005FL
			A(FVE_E_POLICY_STARTUP_PIN_NOT_ALLOWED) // 0x80310060L
			A(FVE_E_POLICY_STARTUP_PIN_REQUIRED) // 0x80310061L
			A(FVE_E_POLICY_STARTUP_KEY_NOT_ALLOWED) // 0x80310062L
			A(FVE_E_POLICY_STARTUP_KEY_REQUIRED) // 0x80310063L
			A(FVE_E_POLICY_STARTUP_PIN_KEY_NOT_ALLOWED) // 0x80310064L
			A(FVE_E_POLICY_STARTUP_PIN_KEY_REQUIRED) // 0x80310065L
			A(FVE_E_POLICY_STARTUP_TPM_NOT_ALLOWED) // 0x80310066L
			A(FVE_E_POLICY_STARTUP_TPM_REQUIRED) // 0x80310067L
			A(FVE_E_POLICY_INVALID_PIN_LENGTH) // 0x80310068L
			A(FVE_E_KEY_PROTECTOR_NOT_SUPPORTED) // 0x80310069L
			A(FVE_E_POLICY_PASSPHRASE_NOT_ALLOWED) // 0x8031006AL
			A(FVE_E_POLICY_PASSPHRASE_REQUIRED) // 0x8031006BL
			A(FVE_E_FIPS_PREVENTS_PASSPHRASE) // 0x8031006CL
			A(FVE_E_OS_VOLUME_PASSPHRASE_NOT_ALLOWED) // 0x8031006DL
			A(FVE_E_INVALID_BITLOCKER_OID) // 0x8031006EL
			A(FVE_E_VOLUME_TOO_SMALL) // 0x8031006FL
			A(FVE_E_DV_NOT_SUPPORTED_ON_FS) // 0x80310070L
			A(FVE_E_DV_NOT_ALLOWED_BY_GP) // 0x80310071L
			A(FVE_E_POLICY_USER_CERTIFICATE_NOT_ALLOWED) // 0x80310072L
			A(FVE_E_POLICY_USER_CERTIFICATE_REQUIRED) // 0x80310073L
			A(FVE_E_POLICY_USER_CERT_MUST_BE_HW) // 0x80310074L
			A(FVE_E_POLICY_USER_CONFIGURE_FDV_AUTOUNLOCK_NOT_ALLOWED) // 0x80310075L
			A(FVE_E_POLICY_USER_CONFIGURE_RDV_AUTOUNLOCK_NOT_ALLOWED) // 0x80310076L
			A(FVE_E_POLICY_USER_CONFIGURE_RDV_NOT_ALLOWED) // 0x80310077L
			A(FVE_E_POLICY_USER_ENABLE_RDV_NOT_ALLOWED) // 0x80310078L
			A(FVE_E_POLICY_USER_DISABLE_RDV_NOT_ALLOWED) // 0x80310079L
			A(FVE_E_POLICY_INVALID_PASSPHRASE_LENGTH) // 0x80310080L
			A(FVE_E_POLICY_PASSPHRASE_TOO_SIMPLE) // 0x80310081L
			A(FVE_E_RECOVERY_PARTITION) // 0x80310082L
			A(FVE_E_POLICY_CONFLICT_FDV_RK_OFF_AUK_ON) // 0x80310083L
			A(FVE_E_POLICY_CONFLICT_RDV_RK_OFF_AUK_ON) // 0x80310084L
			A(FVE_E_NON_BITLOCKER_OID) // 0x80310085L
			A(FVE_E_POLICY_PROHIBITS_SELFSIGNED) // 0x80310086L
			A(FVE_E_POLICY_CONFLICT_RO_AND_STARTUP_KEY_REQUIRED) // 0x80310087L
			A(FVE_E_CONV_RECOVERY_FAILED) // 0x80310088L
			A(FVE_E_VIRTUALIZED_SPACE_TOO_BIG) // 0x80310089L
			A(FVE_E_POLICY_CONFLICT_OSV_RP_OFF_ADB_ON) // 0x80310090L
			A(FVE_E_POLICY_CONFLICT_FDV_RP_OFF_ADB_ON) // 0x80310091L
			A(FVE_E_POLICY_CONFLICT_RDV_RP_OFF_ADB_ON) // 0x80310092L
			A(FVE_E_NON_BITLOCKER_KU) // 0x80310093L
			A(FVE_E_PRIVATEKEY_AUTH_FAILED) // 0x80310094L
			A(FVE_E_REMOVAL_OF_DRA_FAILED) // 0x80310095L
			A(FVE_E_OPERATION_NOT_SUPPORTED_ON_VISTA_VOLUME) // 0x80310096L
			A(FVE_E_CANT_LOCK_AUTOUNLOCK_ENABLED_VOLUME) // 0x80310097L
			A(FVE_E_FIPS_HASH_KDF_NOT_ALLOWED) // 0x80310098L
			A(FVE_E_ENH_PIN_INVALID) // 0x80310099L
			A(FVE_E_INVALID_PIN_CHARS) // 0x8031009AL
			A(FVE_E_INVALID_DATUM_TYPE) // 0x8031009BL
			A(FWP_E_CALLOUT_NOT_FOUND) // 0x80320001L
			A(FWP_E_CONDITION_NOT_FOUND) // 0x80320002L
			A(FWP_E_FILTER_NOT_FOUND) // 0x80320003L
			A(FWP_E_LAYER_NOT_FOUND) // 0x80320004L
			A(FWP_E_PROVIDER_NOT_FOUND) // 0x80320005L
			A(FWP_E_PROVIDER_CONTEXT_NOT_FOUND) // 0x80320006L
			A(FWP_E_SUBLAYER_NOT_FOUND) // 0x80320007L
			A(FWP_E_NOT_FOUND) // 0x80320008L
			A(FWP_E_ALREADY_EXISTS) // 0x80320009L
			A(FWP_E_IN_USE) // 0x8032000AL
			A(FWP_E_DYNAMIC_SESSION_IN_PROGRESS) // 0x8032000BL
			A(FWP_E_WRONG_SESSION) // 0x8032000CL
			A(FWP_E_NO_TXN_IN_PROGRESS) // 0x8032000DL
			A(FWP_E_TXN_IN_PROGRESS) // 0x8032000EL
			A(FWP_E_TXN_ABORTED) // 0x8032000FL
			A(FWP_E_SESSION_ABORTED) // 0x80320010L
			A(FWP_E_INCOMPATIBLE_TXN) // 0x80320011L
			A(FWP_E_TIMEOUT) // 0x80320012L
			A(FWP_E_NET_EVENTS_DISABLED) // 0x80320013L
			A(FWP_E_INCOMPATIBLE_LAYER) // 0x80320014L
			A(FWP_E_KM_CLIENTS_ONLY) // 0x80320015L
			A(FWP_E_LIFETIME_MISMATCH) // 0x80320016L
			A(FWP_E_BUILTIN_OBJECT) // 0x80320017L
			A(FWP_E_TOO_MANY_CALLOUTS) // 0x80320018L
			A(FWP_E_NOTIFICATION_DROPPED) // 0x80320019L
			A(FWP_E_TRAFFIC_MISMATCH) // 0x8032001AL
			A(FWP_E_INCOMPATIBLE_SA_STATE) // 0x8032001BL
			A(FWP_E_NULL_POINTER) // 0x8032001CL
			A(FWP_E_INVALID_ENUMERATOR) // 0x8032001DL
			A(FWP_E_INVALID_FLAGS) // 0x8032001EL
			A(FWP_E_INVALID_NET_MASK) // 0x8032001FL
			A(FWP_E_INVALID_RANGE) // 0x80320020L
			A(FWP_E_INVALID_INTERVAL) // 0x80320021L
			A(FWP_E_ZERO_LENGTH_ARRAY) // 0x80320022L
			A(FWP_E_NULL_DISPLAY_NAME) // 0x80320023L
			A(FWP_E_INVALID_ACTION_TYPE) // 0x80320024L
			A(FWP_E_INVALID_WEIGHT) // 0x80320025L
			A(FWP_E_MATCH_TYPE_MISMATCH) // 0x80320026L
			A(FWP_E_TYPE_MISMATCH) // 0x80320027L
			A(FWP_E_OUT_OF_BOUNDS) // 0x80320028L
			A(FWP_E_RESERVED) // 0x80320029L
			A(FWP_E_DUPLICATE_CONDITION) // 0x8032002AL
			A(FWP_E_DUPLICATE_KEYMOD) // 0x8032002BL
			A(FWP_E_ACTION_INCOMPATIBLE_WITH_LAYER) // 0x8032002CL
			A(FWP_E_ACTION_INCOMPATIBLE_WITH_SUBLAYER) // 0x8032002DL
			A(FWP_E_CONTEXT_INCOMPATIBLE_WITH_LAYER) // 0x8032002EL
			A(FWP_E_CONTEXT_INCOMPATIBLE_WITH_CALLOUT) // 0x8032002FL
			A(FWP_E_INCOMPATIBLE_AUTH_METHOD) // 0x80320030L
			A(FWP_E_INCOMPATIBLE_DH_GROUP) // 0x80320031L
			A(FWP_E_EM_NOT_SUPPORTED) // 0x80320032L
			A(FWP_E_NEVER_MATCH) // 0x80320033L
			A(FWP_E_PROVIDER_CONTEXT_MISMATCH) // 0x80320034L
			A(FWP_E_INVALID_PARAMETER) // 0x80320035L
			A(FWP_E_TOO_MANY_SUBLAYERS) // 0x80320036L
			A(FWP_E_CALLOUT_NOTIFICATION_FAILED) // 0x80320037L
			A(FWP_E_INVALID_AUTH_TRANSFORM) // 0x80320038L
			A(FWP_E_INVALID_CIPHER_TRANSFORM) // 0x80320039L
			A(FWP_E_DROP_NOICMP) // 0x80320104L
			A(FWP_E_INCOMPATIBLE_CIPHER_TRANSFORM) // 0x8032003AL
			A(FWP_E_INVALID_TRANSFORM_COMBINATION) // 0x8032003BL
			A(FWP_E_DUPLICATE_AUTH_METHOD) // 0x8032003CL
			A(WS_S_ASYNC) // 0x003D0000L
			A(WS_S_END) // 0x003D0001L
			A(WS_E_INVALID_FORMAT) // 0x803D0000L
			A(WS_E_OBJECT_FAULTED) // 0x803D0001L
			A(WS_E_NUMERIC_OVERFLOW) // 0x803D0002L
			A(WS_E_INVALID_OPERATION) // 0x803D0003L
			A(WS_E_OPERATION_ABORTED) // 0x803D0004L
			A(WS_E_ENDPOINT_ACCESS_DENIED) // 0x803D0005L
			A(WS_E_OPERATION_TIMED_OUT) // 0x803D0006L
			A(WS_E_OPERATION_ABANDONED) // 0x803D0007L
			A(WS_E_QUOTA_EXCEEDED) // 0x803D0008L
			A(WS_E_NO_TRANSLATION_AVAILABLE) // 0x803D0009L
			A(WS_E_SECURITY_VERIFICATION_FAILURE) // 0x803D000AL
			A(WS_E_ADDRESS_IN_USE) // 0x803D000BL
			A(WS_E_ADDRESS_NOT_AVAILABLE) // 0x803D000CL
			A(WS_E_ENDPOINT_NOT_FOUND) // 0x803D000DL
			A(WS_E_ENDPOINT_NOT_AVAILABLE) // 0x803D000EL
			A(WS_E_ENDPOINT_FAILURE) // 0x803D000FL
			A(WS_E_ENDPOINT_UNREACHABLE) // 0x803D0010L
			A(WS_E_ENDPOINT_ACTION_NOT_SUPPORTED) // 0x803D0011L
			A(WS_E_ENDPOINT_TOO_BUSY) // 0x803D0012L
			A(WS_E_ENDPOINT_FAULT_RECEIVED) // 0x803D0013L
			A(WS_E_ENDPOINT_DISCONNECTED) // 0x803D0014L
			A(WS_E_PROXY_FAILURE) // 0x803D0015L
			A(WS_E_PROXY_ACCESS_DENIED) // 0x803D0016L
			A(WS_E_NOT_SUPPORTED) // 0x803D0017L
			A(WS_E_PROXY_REQUIRES_BASIC_AUTH) // 0x803D0018L
			A(WS_E_PROXY_REQUIRES_DIGEST_AUTH) // 0x803D0019L
			A(WS_E_PROXY_REQUIRES_NTLM_AUTH) // 0x803D001AL
			A(WS_E_PROXY_REQUIRES_NEGOTIATE_AUTH) // 0x803D001BL
			A(WS_E_SERVER_REQUIRES_BASIC_AUTH) // 0x803D001CL
			A(WS_E_SERVER_REQUIRES_DIGEST_AUTH) // 0x803D001DL
			A(WS_E_SERVER_REQUIRES_NTLM_AUTH) // 0x803D001EL
			A(WS_E_SERVER_REQUIRES_NEGOTIATE_AUTH) // 0x803D001FL
			A(WS_E_INVALID_ENDPOINT_URL) // 0x803D0020L
			A(WS_E_OTHER) // 0x803D0021L
			A(WS_E_SECURITY_TOKEN_EXPIRED) // 0x803D0022L
			A(WS_E_SECURITY_SYSTEM_FAILURE) // 0x803D0023L
			A(E_MBN_CONTEXT_NOT_ACTIVATED) // 0x80548201L
			A(E_MBN_BAD_SIM) // 0x80548202L
			A(E_MBN_DATA_CLASS_NOT_AVAILABLE) // 0x80548203L
			A(E_MBN_INVALID_ACCESS_STRING) // 0x80548204L
			A(E_MBN_MAX_ACTIVATED_CONTEXTS) // 0x80548205L
			A(E_MBN_PACKET_SVC_DETACHED) // 0x80548206L
			A(E_MBN_PROVIDER_NOT_VISIBLE) // 0x80548207L
			A(E_MBN_RADIO_POWER_OFF) // 0x80548208L
			A(E_MBN_SERVICE_NOT_ACTIVATED) // 0x80548209L
			A(E_MBN_SIM_NOT_INSERTED) // 0x8054820AL
			A(E_MBN_VOICE_CALL_IN_PROGRESS) // 0x8054820BL
			A(E_MBN_INVALID_CACHE) // 0x8054820CL
			A(E_MBN_NOT_REGISTERED) // 0x8054820DL
			A(E_MBN_PROVIDERS_NOT_FOUND) // 0x8054820EL
			A(E_MBN_PIN_NOT_SUPPORTED) // 0x8054820FL
			A(E_MBN_PIN_REQUIRED) // 0x80548210L
			A(E_MBN_PIN_DISABLED) // 0x80548211L
			A(E_MBN_FAILURE) // 0x80548212L
			A(E_MBN_INVALID_PROFILE) // 0x80548218L
			A(E_MBN_DEFAULT_PROFILE_EXIST) // 0x80548219L
			A(E_MBN_SMS_ENCODING_NOT_SUPPORTED) // 0x80548220L
			A(E_MBN_SMS_FILTER_NOT_SUPPORTED) // 0x80548221L
			A(E_MBN_SMS_INVALID_MEMORY_INDEX) // 0x80548222L
			A(E_MBN_SMS_LANG_NOT_SUPPORTED) // 0x80548223L
			A(E_MBN_SMS_MEMORY_FAILURE) // 0x80548224L
			A(E_MBN_SMS_NETWORK_TIMEOUT) // 0x80548225L
			A(E_MBN_SMS_UNKNOWN_SMSC_ADDRESS) // 0x80548226L
			A(E_MBN_SMS_FORMAT_NOT_SUPPORTED) // 0x80548227L
			A(E_MBN_SMS_OPERATION_NOT_ALLOWED) // 0x80548228L
			A(E_MBN_SMS_MEMORY_FULL) // 0x80548229L
			A(UI_E_CREATE_FAILED) // 0x802A0001L
			A(UI_E_SHUTDOWN_CALLED) // 0x802A0002L
			A(UI_E_ILLEGAL_REENTRANCY) // 0x802A0003L
			A(UI_E_OBJECT_SEALED) // 0x802A0004L
			A(UI_E_VALUE_NOT_SET) // 0x802A0005L
			A(UI_E_VALUE_NOT_DETERMINED) // 0x802A0006L
			A(UI_E_INVALID_OUTPUT) // 0x802A0007L
			A(UI_E_BOOLEAN_EXPECTED) // 0x802A0008L
			A(UI_E_DIFFERENT_OWNER) // 0x802A0009L
			A(UI_E_AMBIGUOUS_MATCH) // 0x802A000AL
			A(UI_E_FP_OVERFLOW) // 0x802A000BL
			A(UI_E_WRONG_THREAD) // 0x802A000CL
			A(UI_E_STORYBOARD_ACTIVE) // 0x802A0101L
			A(UI_E_STORYBOARD_NOT_PLAYING) // 0x802A0102L
			A(UI_E_START_KEYFRAME_AFTER_END) // 0x802A0103L
			A(UI_E_END_KEYFRAME_NOT_DETERMINED) // 0x802A0104L
			A(UI_E_LOOPS_OVERLAP) // 0x802A0105L
			A(UI_E_TRANSITION_ALREADY_USED) // 0x802A0106L
			A(UI_E_TRANSITION_NOT_IN_STORYBOARD) // 0x802A0107L
			A(UI_E_TRANSITION_ECLIPSED) // 0x802A0108L
			A(UI_E_TIME_BEFORE_LAST_UPDATE) // 0x802A0109L
			A(UI_E_TIMER_CLIENT_ALREADY_CONNECTED) // 0x802A010AL
			#undef A
		};
		for(SIZE_T nIndex = 0; nIndex < DIM(g_pMap); nIndex++)
			if(g_pMap[nIndex].nValue == nValue)
				return CString(g_pMap[nIndex].pszName);
		return _T("");
	}

public:
// CNotifyIconWindow
	static ULONGLONG GetCommonControlsVersion() throw()
	{
		DWORD nMajorVersion, nMinorVersion;
		_W(SUCCEEDED(AtlGetCommCtrlVersion(&nMajorVersion, &nMinorVersion)));
		return (ULONGLONG) ((nMajorVersion << 16) + nMinorVersion) << 32;
	}
	static ULONGLONG GetShellVersion() throw()
	{
		DWORD nMajorVersion, nMinorVersion;
		_W(SUCCEEDED(AtlGetShellVersion(&nMajorVersion, &nMinorVersion)));
		return (ULONGLONG) ((nMajorVersion << 16) + nMinorVersion) << 32;
	}
	CNotifyIconWindow()
	{
	}
	HWND Create()
	{
		return __super::Create(NULL, rcDefault, _T("AlaxInfo.ShowHresult.NotifyIconWindow"));
	}
	static CString GetDefaultInfoTitle()
	{
		return AtlLoadString(IDS_NOTIFYICON_DEFAULTTIPTITLE);
	}
	static DWORD GetDefaultInfoFlags() throw()
	{
		return NIIF_NONE;
	}
	VOID SetBalloonToolTip(DWORD nFlags, LPCTSTR pszTitle, LPCTSTR pszText, UINT nTimeout = 30 * 1000)
	{
		if(IsWindow() && m_NotifyIconData.uFlags & NIF_INFO)
		{
			m_NotifyIconData.uTimeout = nTimeout;
			m_NotifyIconData.dwInfoFlags = nFlags;
			_tcsncpy_s(m_NotifyIconData.szInfoTitle, pszTitle, _TRUNCATE);
			_tcsncpy_s(m_NotifyIconData.szInfo, pszText, _TRUNCATE);
			_W(Shell_NotifyIcon(NIM_MODIFY, &m_NotifyIconData));
		}
	}
	BOOL Process(LPCTSTR pszText)
	{
		#pragma region Parse
		if(_tcslen(pszText) > 24)
			return FALSE;
		CString sText = pszText;
		sText.Trim();
		if(!sText.IsEmpty() && _tcschr(_T("Ll"), sText[sText.GetLength() - 1]))
			sText.Delete(sText.GetLength() - 1);
		if(sText.IsEmpty())
			return FALSE;
		LONGLONG nLongLongResult;
		if(_tcsnicmp(sText, _T("0x"), 2) == 0)
		{
			SIZE_T nIndex = 2;
			for(; nIndex < (SIZE_T) sText.GetLength(); nIndex++)
				if(!_tcschr(_T("0123456789ABCDEFabcdef"), sText[nIndex]))
					return FALSE;
			if(!StrToInt64Ex(sText, STIF_SUPPORT_HEX, &nLongLongResult))
				return FALSE; 
		} else
		{
			SIZE_T nIndex = 0;
			if(sText[0] == _T('-'))
				nIndex++;
			for(; nIndex < (SIZE_T) sText.GetLength(); nIndex++)
				if(!_tcschr(_T("0123456789"), sText[nIndex]))
					return FALSE;
			nLongLongResult = _ttoi64(sText);
		}
		const LONG nHighLongLongResult = (LONG) (nLongLongResult >> 32);
		if(!nLongLongResult || nHighLongLongResult > 0 || nHighLongLongResult < -1)
			return FALSE;
		HRESULT nResult = (LONG) nLongLongResult;
		#pragma endregion
		#pragma region Lookup
		CString sTitle = _T("System"), sMessage, sIdentifier;
		if(IsQuartzResult(nResult, &sMessage))
		{
			sTitle = _T("DirectShow");
			sIdentifier = LookupQuartzIdentifier(nResult);
		} else if(IsWmResult(nResult, &sMessage))
			sTitle = _T("Windows Media");
		else if(IsMfResult(nResult, &sMessage))
		{
			sTitle = _T("Media Foundation");
			sIdentifier = LookupMfIdentifier(nResult);
		} else if(IsWs2Result(nResult, &sMessage))
			sTitle = _T("Sockets");
		else if(IsWinHttpResult(nResult, &sMessage))
			sTitle = _T("WinHTTP");
		else if(IsWinInetResult(nResult, &sMessage))
			sTitle = _T("WinInet");
		else 
		{
			sMessage = AtlFormatSystemMessage(nResult);
			sIdentifier = LookupSystemIdentifier(nResult);
			if(sIdentifier.IsEmpty())
				sIdentifier = LookupHresultSystemIdentifier(nResult);
		}
		if(sMessage.IsEmpty())
			return FALSE;
		#pragma endregion
		#pragma region Present
		_A(!sTitle.IsEmpty());
		m_sQuery = AtlFormatString(_T("0x%08X"), nResult) + _T(" ") + sIdentifier + _T(" ") + sMessage;
		CRoArrayT<CString> TitleArray;
		_W(TitleArray.Add(AtlFormatString(_T("0x%08X"), nResult)) >= 0);
		if(!sIdentifier.IsEmpty())
			_W(TitleArray.Add(sIdentifier) >= 0);
		if(!sTitle.IsEmpty())
			_W(TitleArray.Add(sTitle) >= 0);
		SetBalloonToolTip(NIIF_INFO, _StringHelper::Join(TitleArray, _T(" - ")), sMessage);
		#pragma endregion
		return TRUE;
	}

// Window message handelrs
	LRESULT OnCreate(CREATESTRUCT*)
	{
		m_nCommonControlsVersion = GetCommonControlsVersion();
		m_nShellVersion = GetShellVersion();
		SetIcon(AtlLoadIconImage(IDI_MODULE, LR_DEFAULTCOLOR, GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON)), TRUE);
		SetIcon(AtlLoadIconImage(IDI_MODULE, LR_DEFAULTCOLOR, GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON)), FALSE);
		CIcon Icon = AtlLoadIconImage(IDI_MODULE, LR_DEFAULTCOLOR, GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON));
		CString sTip = AtlLoadString(IDS_PROJNAME);
		ZeroMemory(&m_NotifyIconData, sizeof m_NotifyIconData);
		m_NotifyIconData.hWnd = m_hWnd;
		m_NotifyIconData.uID = 1;
		m_NotifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP | NIF_STATE;
		m_NotifyIconData.uCallbackMessage = WM_NOTIFYICON;
		m_NotifyIconData.hIcon = Icon;
		_tcsncpy_s(m_NotifyIconData.szTip, sTip,  _TRUNCATE);
		if(m_nShellVersion >= 0x0005000000000000) // 5.0
		{
			m_NotifyIconData.cbSize = sizeof m_NotifyIconData; 
			m_NotifyIconData.uFlags |= NIF_INFO;
			_tcsncpy_s(m_NotifyIconData.szInfoTitle, GetDefaultInfoTitle(), _TRUNCATE);
			m_NotifyIconData.dwInfoFlags = GetDefaultInfoFlags();
		} else
			m_NotifyIconData.cbSize = NOTIFYICONDATA_V1_SIZE; 
		_W(Shell_NotifyIcon(NIM_ADD, &m_NotifyIconData));
		m_NotifyIconData.uFlags &= ~(NIF_ICON | NIF_STATE);
		m_NextClipboardViewerWindow = SetClipboardViewer();
		_pAtlModule->Lock();
		#if defined(_DEBUG)
		//Process(AtlFormatString(_T("%d"), 0x80040227)); // VFW_E_WRONG_STATE
		Process(_T("0xC00D36B9")); // MF_E_NO_MORE_TYPES)
		//Process(AtlFormatString(_T("0x%x"), HRESULT_FROM_WIN32(WSAEADDRINUSE))); // WSAEADDRINUSE
		//Process(AtlFormatString(_T("0x%x"), HRESULT_FROM_WIN32(ERROR_WINHTTP_AUTODETECTION_FAILED))); // ERROR_WINHTTP_AUTODETECTION_FAILED
		Process(_T("0x80290208L"));
		#endif // defined(_DEBUG)
		return TRUE;
	}
	LRESULT OnDestroy() throw()
	{
		_pAtlModule->Unlock();
		_W(ChangeClipboardChain(m_NextClipboardViewerWindow));
		_W(Shell_NotifyIcon(NIM_DELETE, &m_NotifyIconData));
		m_NotifyIconData.hWnd = NULL;
		return 0;
	}
	LRESULT OnChangeCbChain(CWindow RemovedWindow, CWindow NextWindow)
	{
		if(m_NextClipboardViewerWindow == RemovedWindow)
			m_NextClipboardViewerWindow = NextWindow;
		else if(m_NextClipboardViewerWindow)
			m_NextClipboardViewerWindow.SendMessage(WM_CHANGECBCHAIN, (WPARAM) (HWND) RemovedWindow, (LPARAM) (HWND) NextWindow);
		return 0;
	}
	LRESULT OnDrawClipboard()
	{
		if(OpenClipboard()) 
		{
			_ATLTRY
			{
				if(IsClipboardFormatAvailable(CF_UNICODETEXT))
				{
					CGlobalMemoryWeakHandle DataHandle = (HGLOBAL) GetClipboardData(CF_UNICODETEXT);
					__E(DataHandle);
					CGlobalMemoryHandle::CLockT<WCHAR> pszData(DataHandle);
					Process(CW2CT(pszData));
				}
			}
			_ATLCATCHALL()
			{
				_W(CloseClipboard());
				_ATLRETHROW;
			}
			_W(CloseClipboard());
		}
		return 0;
	}
	LRESULT OnMouseMove(UINT, CPoint)
	{
		return 0;
	}
	LRESULT OnLButtonDblClk(UINT, CPoint)
	{
		if(!IsWindowEnabled())
			return 0;
		INT nDefaultCommandIdentifier = ID_APP_ABOUT;
		//CWaitCursor WaitCursor;
		if(nDefaultCommandIdentifier)
			_W(PostMessage(WM_COMMAND, nDefaultCommandIdentifier));
		return 0;
	}
	LRESULT OnRButtonUp(UINT, CPoint)
	{
		CMenu ContainerMenu;
		_W(ContainerMenu.LoadMenu(IDR_NOTIFYICON));
		CMenuHandle Menu;
		INT nDefaultCommandIdentifier = ID_APP_ABOUT;
		Menu = ContainerMenu.GetSubMenu(0);
		//CWaitCursor WaitCursor;
		_A(Menu);
		if(nDefaultCommandIdentifier)
			_W(Menu.SetMenuDefaultItem(nDefaultCommandIdentifier));
		CPoint Position;
		GetCursorPos(&Position);
		_W(SetForegroundWindow(m_hWnd));
		_W(Menu.TrackPopupMenu(TPM_LEFTALIGN | TPM_TOPALIGN | TPM_VERTICAL, Position.x, Position.y, m_hWnd));
		return 0;
	}
	LRESULT OnNotifyIconBalloonUserClick(UINT, WPARAM, LPARAM)
	{
		if(m_sQuery.IsEmpty())
			return 0;
		CWinHttpRequestIdentifier Identifier(_T("http://google.com/search"));
		Identifier.AddSearchValue(_T("q"), m_sQuery);
		CWaitCursor WaitCursor;
		ShellExecute(m_hWnd, NULL, CW2CT(Identifier.GetValue()), NULL, NULL, SW_SHOWDEFAULT);
		return 0;
	}
	LRESULT OnNotifyIcon(UINT uMsg, WPARAM wParam, LPARAM lParam) throw()
	{
		_A(wParam == m_NotifyIconData.uID);
		return SendMessage((UINT) lParam);
	}
	LRESULT OnApplicationAbout(UINT, INT, HWND) throw()
	{
		CAboutDialog Dialog;
		EnableWindow(FALSE);
		Dialog.DoModal();
		EnableWindow(TRUE);
		return 0;
	}
	LRESULT OnApplicationExit(UINT = 0, INT = ID_APP_EXIT, HWND = NULL) throw()
	{
		PostQuitMessage(0);
		return 0;
	}
};
